# Copyright 2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import os

from fabric.api import env, local, settings
from fabric.context_managers import hide

from .django import get_django_settings, syncdb


def createdb():
    """Create the database."""
    _set_postgres_environment()
    local("%(ENV)s %(BIN)s/createdb %(DATABASE)s" % env.postgres)

def dropdb(warn_only=False):
    """Remove the database."""
    _set_postgres_environment()
    if isinstance(warn_only, basestring):
        warn_only = warn_only.lower() == 'yes'
    with settings(warn_only=warn_only):
        local("%(ENV)s %(BIN)s/dropdb %(DATABASE)s" % env.postgres)

def grantuser():
    pass

def setup_postgresql_server():
    """Setup the PostgreSQL server."""
    _set_postgres_environment()
    local("mkdir -p %(DATA)s" % env.postgres)
    local("%(ENV)s %(BIN)s/initdb -A trust -D %(DATA)s" % env.postgres)
    local("cat <<EOF\nfsync = off\nEOF > %(DATA)s/postgresql.conf" %
        env.postgres)
    start_database()
    setup_database()

def start_database():
    """Start the PostgreSQL server."""
    _set_postgres_environment()
    cmd = ['%(ENV)s', '%(BIN)s/pg_ctl', 'start', '-w', '-l',
        '%(HOST)s/postgresql.log', '-o', '"-F -k %(HOST)s -h \'\'"']
    local(' '.join(cmd) % env.postgres)

def stop_database():
    """Stop the PostgreSQL server."""
    _set_postgres_environment()
    cmd = ['%(ENV)s', '%(BIN)s/pg_ctl', 'stop', '-w', '-m', 'fast']
    local(' '.join(cmd) % env.postgres)

def shutdown_postgresql_server():
    """Shutdown the PostgreSQL server."""
    _set_postgres_environment()
    dropdb()
    stop_database()
    local("rm -rf %s" % env.postgres['HOST'])

def setup_database():
    """Setup the database."""
    success = _check_database()
    if not success:
        createdb()
        _createrole()
    syncdb()


# helpers
# =======

def _set_database_environment():
    """Update the environment with the database settings."""
    if 'database' in env:
        # environment already set up
        return

    settings = get_django_settings('DATABASE_HOST', 'DATABASE_PORT',
        'DATABASE_USER', 'DATABASE_PASSWORD', 'DATABASE_NAME')
    env.database = {
        'HOST': settings['DATABASE_HOST'],
        'PORT': settings['DATABASE_PORT'],
        'NAME': settings['DATABASE_NAME'],
        'USER': settings['DATABASE_USER'],
        'PASSWORD': settings['DATABASE_PASSWORD'],
    }

def _set_postgres_environment():
    """Update the environment with the PostgreSQL settings."""
    if 'postgres' in env:
        # environment already set up
        return

    _set_database_environment()

    pg_env = []
    env.postgres = {
        'BIN': '/usr/lib/postgresql/8.4/bin',
        'DATABASE': env.database['NAME'],
    }

    host = os.environ.get('PGHOST', env.database['HOST'])
    if host:
        data = "%s/data" % host
        env.postgres['HOST'] = host
        env.postgres['DATA'] = data
        pg_env.append("PGHOST=%s" % host)
        pg_env.append("PGDATA=%s" % data)

    port = env.database['PORT']
    if port:
        pg_env.append("PGPORT=%s" % port)

    env.postgres['ENV'] = ' '.join(pg_env)

def _check_database():
    """Check the database is accessible."""
    with hide('aborts', 'running'):
        try:
            _set_postgres_environment()
            local("%(ENV)s psql -U postgres -c '\q'"
                % env.postgres, capture=False)
            success = True
        except SystemExit:
            # there was an error connecting to the db,
            # presumably the db didn't exist
            success = False
    return success

def _createrole():
    """Create required users/roles."""
    _set_postgres_environment()
    local("%(ENV)s %(BIN)s/createuser --superuser --createdb postgres" % env.postgres)
