# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import logging

from django.conf import settings
from django.core.serializers.json import DateTimeAwareJSONEncoder
from django.http import (
    HttpResponseBadRequest,
    HttpResponseForbidden,
    )
from django.shortcuts import render_to_response
from django.utils import simplejson
from piston.emitters import Emitter
from piston.handler import BaseHandler

from identityprovider.api10.decorators import (
    api_user_required,
    plain_user_required,
    named_operation,
)
from identityprovider.api10.forms import (
    PASSWORD_POLICY_ERROR,
    WebserviceCreateAccountForm,
)
from identityprovider.models import (
    EmailAddress,
    Account,
    AuthToken,
    AuthTokenFactory,
)
from identityprovider.models.const import (
    EmailStatus,
    LoginTokenType,
)
from identityprovider.models.captcha import (
    Captcha,
    NewCaptchaError,
    VerifyCaptchaError,
)
from identityprovider.models.team import TeamMembership
from identityprovider.signals import (
    account_created,
    account_email_validated,
    application_token_created,
    application_token_invalidated,
)
from identityprovider.utils import (
    CannotResetPasswordException,
    encrypt_launchpad_password,
    get_person_and_account_by_email,
    password_policy_compliant,
    PersonAndAccountNotFoundException,
)
from oauth_backend.models import Token


class LazrRestfulEmitter(Emitter):
    """JSON emitter, lazr.restful flavoured.

    Reports content-type 'application/json', without specifying a charset
    as that seems to confuse lazr.restfulclient.
    """
    def render(self, request):
        # ensure_ascii=False was causing bug #820216
        # it looks like this is because simplejson's implementation is
        # broken
        # http://code.google.com/p/simplejson/issues/detail?id=48
        # https://code.djangoproject.com/ticket/6234
        seria = simplejson.dumps(self.construct(),
            cls=DateTimeAwareJSONEncoder, indent=4)

        return seria
Emitter.register('lazr.restful', LazrRestfulEmitter, 'application/json')


class LazrRestfulHandler(BaseHandler):
    allowed_methods = ('GET', 'POST')

    response = {}

    baseurl = settings.SSO_ROOT_URL.strip('/')

    def __init__(self):
        self.response = self.response.copy()
        for key, value in self.response.items():
            if key.endswith('_link'):
                self.response[key] = value % self.baseurl

    def read(self, request):
        if not 'ws.op' in request.GET:
            return self.response
        return self.named_operation(request, request.GET)

    def create(self, request):
        if not 'ws.op' in request.POST:
            return HttpResponseBadRequest('No operation name given.')
        return self.named_operation(request, request.POST)

    def named_operation(self, request, serialized):
        method_name = serialized['ws.op']
        method = getattr(self, method_name, None)
        is_named_operation = getattr(method, 'is_named_operation', False)
        if not is_named_operation:
            return HttpResponseBadRequest('No such operation: %s' %
                method_name)
        request.data = self.lazr_restful_deserialize(serialized)
        return method(request)

    def lazr_restful_deserialize(self, serialized):
        data = {}
        for key in serialized:
            if key == 'ws.op':
                continue
            try:
                value = simplejson.loads(serialized[key])
            except (ValueError, TypeError):
                value = serialized[key]
            data[key] = value
        return data


class RootHandler(LazrRestfulHandler):
    allowed_methods = ('GET',)

    response = {
        "registrations_collection_link": "%s/api/1.0/registration",
        "captchas_collection_link": "%s/api/1.0/captchas",
        "validations_collection_link": "%s/api/1.0/validation",
        "authentications_collection_link": "%s/api/1.0/authentications",
        "resource_type_link": "%s/api/1.0/#service-root",
        "accounts_collection_link": "%s/api/1.0/accounts",
    }

    def read(self, request):
        if ('application/vd.sun.wadl+xml' in request.META['HTTP_ACCEPT'] or
            'application/vnd.sun.wadl+xml' in request.META['HTTP_ACCEPT']):
            context = {'baseurl': self.baseurl}
            return render_to_response('wadl1.0.xml', context)
        else:
            return self.response


class CaptchaHandler(LazrRestfulHandler):
    response = {
        "total_size": 0,
        "start": None,
        "resource_type_link": "%s/api/1.0/#captchas",
        "entries": []
    }

    @named_operation
    def new(self, request):
        try:
            return Captcha.new().serialize()
        except NewCaptchaError, e:
            logging.exception("Failed to connect to reCaptcha server")
            # TODO: Some better return here?
            return e.dummy


class RegistrationHandler(LazrRestfulHandler):
    response = {
        "total_size": 0,
        "start": None,
        "resource_type_link": "%s/api/1.0/#registrations",
        "entries": []
    }

    @named_operation
    def register(self, request):
        data = request.data
        data['remote_ip'] = request.environ['REMOTE_ADDR']
        form = WebserviceCreateAccountForm(data)
        try:
            if not form.is_valid():
                errors = dict((k, map(unicode, v))
                              for (k, v) in form.errors.items())
                result = {'status': 'error', 'errors': errors}
                return result
        except VerifyCaptchaError:
            logging.exception("reCaptcha connection error")
            return {'status': 'error', 'errors':
                {'captcha_solution': [
                    'Unable to verify captcha.  Please try again shortly.']}}

        cleaned_data = form.cleaned_data
        requested_email = cleaned_data['email']
        emails = EmailAddress.objects.filter(email__iexact=requested_email)
        if len(emails) > 0:
            return {'status': 'error', 'errors':
                    {'email': ['Email already registered']}}

        platform = cleaned_data['platform']
        redirection_url = cleaned_data['validate_redirect_to']
        encrypted_password = encrypt_launchpad_password(
            cleaned_data['password'])
        displayname = cleaned_data['displayname']
        email = cleaned_data['email']
        if platform in ['desktop', 'mobile']:
            account = Account.objects.create_account(
                displayname, email,
                encrypted_password, password_encrypted=True,
                email_validated=False)

        if platform == 'desktop':
            token = AuthTokenFactory().new_api_email_validation_token(
                account, email, redirection_url)
        else:
            token = AuthTokenFactory().new(None, None, email,
                LoginTokenType.NEWPERSONLESSACCOUNT, redirection_url,
                displayname=displayname,
                password=encrypted_password)

        template = '{0}-newuser.txt'.format(platform)
        if template is None:
            msg = ('Invalid platform requested during registration: {0}. '
                'Using default platform.'.format(platform))
            logging.error(msg)
            template = 'desktop-newuser.txt'
        token.sendNewUserEmail(template=template)

        if platform in ['desktop', 'mobile']:
            account_created.send(sender=self,
                                 openid_identifier=account.openid_identifier)

        return {
            'status': 'ok',
            'message': "Email verification required."
        }

    @named_operation
    def request_password_reset_token(self, request):
        data = request.data
        email = data['email']
        error = False
        try:
            person, account = get_person_and_account_by_email(email)
        except (CannotResetPasswordException,
                PersonAndAccountNotFoundException):
            error = True

        if error or (account is not None and not account.can_reset_password):
            return HttpResponseForbidden(
                "CanNotResetPasswordError: Can't reset password for this "
                "account.")

        token = AuthTokenFactory().new(account, email, email,
                                       LoginTokenType.PASSWORDRECOVERY, None)

        token.sendPasswordResetEmail()

        return {
            'status': 'ok',
            'message': "Password reset token sent."
        }

    @named_operation
    def set_new_password(self, request):
        data = request.data
        new_password = data['new_password']
        try:
            token = AuthToken.objects.get(
                email=data['email'], token=data['token'],
                token_type=LoginTokenType.PASSWORDRECOVERY)
        except AuthToken.DoesNotExist:
            return HttpResponseForbidden(
                "CanNotSetPasswordError: Invalid token.")
        if not token.requester:
            token.delete()
            return {
                'status': 'error',
                'message': "Wrong token, request new one."
            }
        if not password_policy_compliant(new_password):
            return {
                'status': 'error',
                'errors': [PASSWORD_POLICY_ERROR]
            }
        password_obj = token.requester.accountpassword
        password_obj.password = encrypt_launchpad_password(new_password)
        password_obj.save()

        token.consume()

        return {
            'status': 'ok',
            'message': "Password changed"
        }


def _serialize_account(user):
    emails = EmailAddress.objects.filter(account=user,
                            status=EmailStatus.VALIDATED)
    preferred_email = user.preferredemail
    if preferred_email is not None:
        preferred_email = preferred_email.email

    if user.person:
        username = user.person.name
    else:
        username = user.openid_identifier

    return {
        'username': username,
        'displayname': user.displayname,
        'openid_identifier': user.openid_identifier,
        'preferred_email': preferred_email,
        'verified_emails': [e.email for e in emails],
        'unverified_emails': [e.email for e in user.unverified_emails()],
    }


class AuthenticationHandler(LazrRestfulHandler):
    """ All these methods assume that they're run behind Basic Auth """
    response = {
        "total_size": 0,
        "start": None,
        "resource_type_link": "%s/api/1.0/#authentications",
        "entries": []
    }

    @plain_user_required
    @named_operation
    def authenticate(self, request):
        data = request.data
        account = request.user
        token = account.create_oauth_token(data['token_name'])
        application_token_created.send(
            sender=self, openid_identifier=account.openid_identifier)
        return token.serialize()

    @api_user_required
    @named_operation
    def list_tokens(self, request):
        data = request.data
        tokens = Token.objects.filter(
            consumer__user__username=data['consumer_key'])
        result = [{'token': t.token, 'name': t.name} for t in tokens]
        return result

    @api_user_required
    @named_operation
    def validate_token(self, request):
        data = request.data
        try:
            token = Token.objects.get(
                consumer__user__username=data['consumer_key'],
                token=data['token'])
            return token.serialize()
        except Token.DoesNotExist:
            return False

    @api_user_required
    @named_operation
    def invalidate_token(self, request):
        data = request.data
        tokens = Token.objects.filter(token=data['token'],
            consumer__user__username=data['consumer_key'])
        tokens.delete()
        application_token_invalidated.send(
            sender=self, openid_identifier=data['consumer_key'])

    @api_user_required
    @named_operation
    def team_memberships(self, request):
        data = request.data
        accounts = Account.objects.filter(
            openid_identifier=data['openid_identifier'])
        accounts = list(accounts)

        if len(accounts) == 1:
            account = accounts[0]
            memberships = (
                TeamMembership.get_team_memberships_for_user(
                    data['team_names'], account, include_private=True))
            return memberships
        else:
            return []

    @api_user_required
    @named_operation
    def account_by_email(self, request):
        data = request.data
        account = Account.objects.get_by_email(data['email'])
        if account:
            return _serialize_account(account)
        else:
            return None

    @api_user_required
    @named_operation
    def account_by_openid(self, request):
        data = request.data
        try:
            account = Account.objects.get(openid_identifier=data['openid'])
        except Account.DoesNotExist:
            return None
        else:
            return _serialize_account(account)


class AccountsHandler(LazrRestfulHandler):
    @named_operation
    def me(self, request):
        account = request.user
        return _serialize_account(account)

    @named_operation
    def team_memberships(self, request):
        team_names = request.data['team_names']

        memberships = (
            TeamMembership.get_team_memberships_for_user(
                team_names, request.user, include_private=True))
        return memberships

    @named_operation
    def validate_email(self, request):
        email_token = request.data['email_token']
        try:
            token = request.user.authtoken_set.get(
                token=email_token, token_type=LoginTokenType.VALIDATEEMAIL)

            email = EmailAddress.objects.get(email__iexact=token.email)
            email.status = EmailStatus.VALIDATED
            email.save()

            token.consume()

            account_email_validated.send(
                openid_identifier=request.user.openid_identifier,
                sender=self)
            return {'email': email.email}
        except AuthToken.DoesNotExist:
            return {'errors': {'email_token': ["Bad email token!"]}}
