# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from datetime import datetime
from django.http import HttpResponse

from django.conf import settings
from oauth_backend.models import Consumer, Token
from piston.authentication import OAuthAuthentication
from piston.oauth import OAuthError as PistonOAuthError

from identityprovider.models import Account, AccountPassword
from identityprovider.models.api import APIUser
from identityprovider.models.const import EmailStatus
from identityprovider.utils import (
    validate_launchpad_password, get_object_or_none)


class LaunchpadBackend(object):

    valid_email_statuses = (
        EmailStatus.PREFERRED, EmailStatus.VALIDATED, EmailStatus.NEW)

    def authenticate(self, username=None, password=None,
                     encrypted_password=None):
        try:
            account = Account.objects.select_related('accountpassword').get(
                         emailaddress__email__iexact=username,
                         emailaddress__status__in=self.valid_email_statuses)
        except Account.DoesNotExist:
            return None

        if not account.is_active:
            return None

        try:
            ap = account.accountpassword
            if ap is None:
                return None
        except AccountPassword.DoesNotExist:
            # password lost somehow, user should reset it
            return None

        try:
            if password is not None:
                if not validate_launchpad_password(password, ap.password):
                    return None
            elif encrypted_password is not None:
                if encrypted_password != ap.password:
                    return None
            else:
                return None
        except (UnicodeDecodeError, UnicodeEncodeError):
            # Password not ASCII, we must fail
            return None

        # we use now() here because last_login maps to django's auth model
        # which uses localtime.
        account.last_login = datetime.now()
        return account

    def get_user(self, user_id):
        """ Returns a Launchpad Account object instead of Django's built-in
        User object. """
        return get_object_or_none(Account.objects.select_related(), pk=user_id)

    def supports_anonymous_user(self):
        return False

    def supports_object_permissions(self):
        return False


def basic_authenticate(username, password):
    """Function used by lazr.authentication middleware"""
    api_user = APIUser.authenticate(username, password)
    if api_user is not None:
        return api_user
    user = LaunchpadBackend().authenticate(username, password)
    return user


class SSOOAuthAuthentication(OAuthAuthentication):
    def is_authenticated(self, request):
        if not self.is_valid_request(request):
            return False
        try:
            consumer, token, parameters = self.validate_token(request)
        except PistonOAuthError:
            return False
        if not (consumer and token):
            return False
        # only allow authentication if account is active
        account = Account.objects.get(openid_identifier=consumer.key)
        if not account.is_active:
            return False
        request.user = account
        request.throttle_extra = token.consumer.id
        return True

    def challenge(self):
        resp = HttpResponse("Authorization Required")
        resp['WWW-Authenticate'] = 'OAuth realm="%s"' % self.realm
        resp.status_code = 401
        return resp


def oauth_authenticate(oauth_consumer, oauth_token, parameters):
    """Currently only checks that given consumer and token are in database"""
    try:
        consumer = Consumer.objects.get(
            user__username=oauth_consumer.key)
        token = Token.objects.get(token=oauth_token.key)

    except (Token.DoesNotExist, Consumer.DoesNotExist):
        return None
    else:
        # only allow authentication if account is active
        account = Account.objects.get(openid_identifier=consumer.key)
        if not account.is_active:
            return None

        if token.consumer.key == oauth_consumer.key:
            return account
        else:
            return None


def internal_authorize(user, teams_setting=None):
    """
    Returns true if user is a logged-in member of teams,
    specified in Django teams_setting.
    """
    if teams_setting is None:
        teams_setting = 'SSO_VERSIONS_TEAMS'
    if user.is_authenticated() and user.person is not None:
        for team in getattr(settings, teams_setting, []):
            if user.person.in_team(team):
                return True
    return False
