# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import logging

from django.template import RequestContext
from django.http import Http404, HttpResponseRedirect, HttpResponseForbidden
from django.core.cache import cache
from django.conf import settings
from django.template.loader import render_to_string
from django.utils.http import urlencode

from datetime import datetime, timedelta
import functools
from hashlib import sha1

from identityprovider import auth
from identityprovider.cookies import set_test_cookie, test_cookie_worked

def guest_required(func):
    @functools.wraps(func)
    def _guest_required_decorator(request, *args, **kwargs):
        if request.user.is_authenticated():
            return HttpResponseRedirect('/')
        else:
            return func(request, *args, **kwargs)
    return _guest_required_decorator


def dont_cache(func):
    def _dont_cache_decorator(request, *args, **kwargs):
        response = func(request, *args, **kwargs)
        response["Expires"] = "Tue, 03 Jul 2001 06:00:00 GMT"
        response["Cache-Control"] = ("no-store, no-cache, "
                                     "must-revalidate, max-age=0")
        response['Pragma'] = 'no-cache'
        return response
    return _dont_cache_decorator


def check_readonly(func):
    """ A readonly aware decorator.

    The decorated view will not be accessible at all during readonly mode.
    Instead, a static warning page will be displayed.
    """
    def wrapper(request, *args, **kwargs):
        if settings.READ_ONLY_MODE:
            html = render_to_string('readonly.html',
                {'readonly': False},
                context_instance=RequestContext(request))
            return HttpResponseForbidden(html)
        return func(request, *args, **kwargs)
    functools.update_wrapper(wrapper, func)
    return wrapper


disable_cookie_check = False

def requires_cookies(func):
    @functools.wraps(func)
    def wrapper(request, *args, **kwargs):
        if disable_cookie_check or test_cookie_worked(request):
            return func(request, *args, **kwargs)
        quoted = urlencode({'next': request.get_full_path()})
        return set_test_cookie(HttpResponseRedirect('/+cookie?' + quoted))
    return wrapper


class ratelimit(object):
    """ A rate-limiting decorator.
        Strongly based on Simon Willison's code,
        http://github.com/simonw/ratelimitcache/blob/master/ratelimitcache.py
    """
    # This class is designed to be sub-classed
    minutes = 2  # The time period
    requests = 20  # Number of allowed requests in that time period

    prefix = 'rl-'  # Prefix for memcache key

    def __init__(self, **options):
        for key, value in options.items():
            setattr(self, key, value)

    def __call__(self, fn):
        def wrapper(request, *args, **kwargs):
            return self.view_wrapper(request, fn, *args, **kwargs)
        functools.update_wrapper(wrapper, fn)
        return wrapper

    def view_wrapper(self, request, fn, *args, **kwargs):
        if not self.should_ratelimit(request):
            return fn(request, *args, **kwargs)

        counts = self.get_counters(request).values()
        # Increment rate limiting counter
        self.cache_incr(self.current_key(request))

        # Have they failed?
        if sum(int(x) for x in counts) >= self.requests:
            return self.disallowed(request)

        return fn(request, *args, **kwargs)

    def cache_get_many(self, keys):
        return cache.get_many(keys)

    def cache_incr(self, key):
        # memcache is only backend that can increment atomically
        try:
            # add first, to ensure the key exists
            cache.add(key, 0, timeout=self.expire_after())
            cache.incr(key)
        except AttributeError:
            cache.set(key, cache.get(key, 0) + 1, self.expire_after())

    def should_ratelimit(self, request):
        return True

    def get_counters(self, request):
        return self.cache_get_many(self.keys_to_check(request))

    def keys_to_check(self, request):
        extra = self.key_extra(request)
        now = datetime.utcnow()
        return [
            '%s%s-%s' % (
                self.prefix,
                extra,
                (now - timedelta(minutes=minute)).strftime('%Y%m%d%H%M')
            ) for minute in range(self.minutes + 1)
        ]

    def current_key(self, request):
        return '%s%s-%s' % (
            self.prefix,
            self.key_extra(request),
            datetime.utcnow().strftime('%Y%m%d%H%M')
        )

    def remote_ip(self, request):
        if 'HTTP_X_FORWARDED_FOR' in request.META:
            remote_ip = request.META['HTTP_X_FORWARDED_FOR']
            # Because X-Forwarded-For can be a list of IP's we need only one
            remote_ip = remote_ip.split(',')[0].strip()
        else:
            remote_ip = request.META.get('REMOTE_ADDR')
        return remote_ip

    def key_extra(self, request):
        # By default, their IP address is used
        return self.remote_ip(request)

    def disallowed(self, request):
        remote_ip = self.remote_ip(request)
        logger = logging.getLogger('ratelimit.disallowed')
        logger.warn("%s (%s) exceeded rate limit for %s",
                     request.user.username, remote_ip, request.user.id)
        return HttpResponseForbidden(render_to_string('limitexceeded.html'))

    def expire_after(self):
        "Used for setting the memcached cache expiry"
        return (self.minutes + 1) * 60

    def reset_count(self, request):
        "Reset the rate limiting limit count for a request"
        for key in self.keys_to_check(request):
            cache.delete(key)


class ratelimit_post(ratelimit):
    "Rate limit POSTs - can be used to protect a login form"
    key_field = None  # If provided, this POST var will affect the rate limit

    def should_ratelimit(self, request):
        return request.method == 'POST'

    def key_extra(self, request):
        # IP address and key_field (if it is set)
        extra = super(ratelimit_post, self).key_extra(request)
        if self.key_field:
            value = sha1(request.POST.get(self.key_field, '').encode('utf-8')).hexdigest()
            extra += '-' + value
        return extra


class limitlogin(ratelimit_post):
    """ Limit login POSTs, per username.
        Also, take default values from settings.
    """
    key_field = 'email'

    def __init__(self, **options):
        args = {
            'minutes': getattr(settings, 'LOGIN_LIMIT_MINUTES', 2),
            'requests': getattr(settings, 'LOGIN_LIMIT_REQUESTS', 20),
            }
        args.update(options)
        super(limitlogin, self).__init__(**args)


def internal_only(view):
    """
    Restricts access to internal team members only, and masks the
    existence of the view from others.
    """
    @functools.wraps(view)
    def wrapper(request, *args, **kwargs):
        if auth.internal_authorize(request.user):
            return view(request, *args, **kwargs)
        raise Http404
    return wrapper
