# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import hashlib
import urllib2
from datetime import datetime

from django import forms
from django.conf import settings
from django.utils.translation import ugettext as _

from identityprovider.widgets import CommaSeparatedWidget


class CommaSeparatedField(forms.MultipleChoiceField):
    widget = CommaSeparatedWidget

    def clean(self, value):
        return ','.join(super(CommaSeparatedField, self).clean(value))


class OneTimePasswordField(forms.CharField):
    """A string of chars between 34 and 48 chars in length.

    Note: we don't check the actual characters as different keyboard
    layouts can create different otps, which should be handled by
    the validation server.
    """
    def __init__(self, *args, **kwargs):
        error_msg = _('The one-time password is invalid.')
        kwargs['min_length'] = 34
        kwargs['max_length'] = 48
        kwargs['error_messages'] = {
            'invalid': error_msg,
            'max_length': error_msg,
            'min_length': error_msg,
            }
        super(OneTimePasswordField, self).__init__(*args, **kwargs)

    def request_otp_validation(self, otp):
        # XXX Currently yubico's api verification replies with
        # BACKEND_ERROR if the nonce is 40 chars long.
        nonce = hashlib.sha1(str(datetime.utcnow())).hexdigest()[:30]
        url = settings.YUBICO_API_URL.format(
                client_id=settings.YUBICO_CLIENT_ID, otp=otp, nonce=nonce)

        try:
            response = urllib2.urlopen(url)
        except urllib2.URLError:
            raise forms.ValidationError(
                _('There was an error checking your one-time password.'))

        return self.convert_response_to_dict(response.read())

    def convert_response_to_dict(self, response):
        response_lines = response.split()
        response_dict = {}
        for line in response_lines:
            key, delim, val = line.partition('=')
            response_dict[key] = val
        return response_dict

    def clean(self, otp):
        otp = super(OneTimePasswordField, self).clean(otp)

        # If the otp has the correct length, then test it against the
        # server.
        result = self.request_otp_validation(otp)

        if result.get('status', None) == 'OK':
            return otp

        raise forms.ValidationError(_('The one-time password is invalid.'))
