# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import logging

from django import forms
from django.contrib import auth
from django.forms import Form, fields, widgets
from django.utils import translation
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext as _

from identityprovider.const import (
    SREG_DATA_FIELDS_ORDER,
    SREG_LABELS,
)
from identityprovider.models import (
    Account,
    EmailAddress,
    verify_token_string,
    TeamMembership,
)
from identityprovider.models.const import EmailStatus
from identityprovider.utils import (
    get_person_and_account_by_email,
    CannotResetPasswordException,
    PersonAndAccountNotFoundException,
    encrypt_launchpad_password,
    password_policy_compliant,
)
from identityprovider.widgets import (
    ROAwareTextInput,
    ROAwareSelect,
)

logger = logging.getLogger('sso')

default_errors = {
    'required': _(u'Required field.'),
    'invalid': _(u'Invalid value.'),
}

email_errors = {
    'required': _(u'Required field.'),
    # Translators: This is used on all email-related input forms.
    'invalid': _(u'Invalid email.'),
}


class GenericEmailForm(forms.Form):
    email = fields.EmailField(error_messages=email_errors,
        widget=widgets.TextInput(attrs={'class': 'textType', 'size': '26'}))


class LoginForm(GenericEmailForm):
    password = fields.CharField(error_messages=default_errors,
        widget=widgets.PasswordInput(attrs={'class': 'textType',
                                            'size': ' 26'}))

    def clean(self):
        if 'email' in self.cleaned_data and 'password' in self.cleaned_data:
            email = self.cleaned_data['email']
            password = self.cleaned_data['password']

            password_error = forms.ValidationError(
                mark_safe(_("Password didn't match.")))

            account = auth.authenticate(username=email, password=password)
            if account is None:
                account = Account.objects.get_by_email(email)
                if account and not account.is_active:
                    raise forms.ValidationError(_("Your account has been "
                                                  "deactivated"))
                else:
                    raise password_error

            # Check that the email used for logging in is not 'NEW'. It's
            # possible to use it to log in to the SSO with API, but not through
            # web.
            # WARNING: even if the RFC says that emails should be
            # treated case-sensitive, from a usability perspective we
            # need to treat the case-insensitive.
            email_obj = account.emailaddress_set.get(email__iexact=email)
            if email_obj.status == EmailStatus.NEW:
                raise password_error

            # Pass account to the view, as it's already authenticated and
            # there's no good reason to do exactly the same call there.
            self.cleaned_data['account'] = account
            return self.cleaned_data


class ForgotPasswordForm(GenericEmailForm):

    def clean_email(self):
        if 'email' in self.cleaned_data:
            email = self.cleaned_data['email']
            try:
                person, account = get_person_and_account_by_email(email)
            except (CannotResetPasswordException,
                    PersonAndAccountNotFoundException):
                # Return the email. We do not raise a ValidationError,
                # because we'd like to show the 'Email sent' page anyway
                pass
            return email


class ResetPasswordForm(forms.Form):
    password = fields.CharField(error_messages=default_errors,
        widget=widgets.PasswordInput(attrs={
            'class': 'textType',
            'size': '20',
        }))
    passwordconfirm = fields.CharField(error_messages=default_errors,
        widget=widgets.PasswordInput(attrs={
            'class': 'textType',
            'size': '20',
        }))

    def clean_password(self):
        if 'password' in self.cleaned_data:
            password = self.cleaned_data['password']
            try:
                str(password)
            except UnicodeEncodeError:
                raise forms.ValidationError(
                    _("Invalid characters in password"))
            if not password_policy_compliant(password):
                raise forms.ValidationError(_("Password must be at least "
                    "8 characters long, and must contain at least one "
                    "number and an upper case letter."))
            return password

    def clean(self):
        cleaned_data = self.cleaned_data
        password = cleaned_data.get('password')
        passwordconfirm = cleaned_data.get('passwordconfirm')
        if (password != passwordconfirm
            and not self['password'].errors):
            raise forms.ValidationError(_("Passwords didn't match"))
        return cleaned_data


# {{workflow}}
class OldNewAccountForm(GenericEmailForm):
    pass


class NewAccountForm(GenericEmailForm, ResetPasswordForm):
    displayname = fields.CharField(error_messages=default_errors,
        widget=widgets.TextInput(attrs={'class': 'textType', 'size': '20'}))


class ConfirmNewAccountForm(ResetPasswordForm):
    displayname = fields.CharField(error_messages=default_errors,
        widget=widgets.TextInput(attrs={'class': 'textType', 'size': '20'}))


class TokenForm(GenericEmailForm):
    """If the confirmation-code and e-mail are entered correctly, and
    if the specified code exists and can be used, the confirmation
    object is placed in cleaned_data['confirmation']."""

    confirmation_code = fields.CharField()

    def clean(self):
        data = self.cleaned_data
        confirmation_code = data.get('confirmation_code')
        email = data.get('email')
        if confirmation_code is not None and email is not None:
            confirmation = verify_token_string(confirmation_code, email)
            if confirmation is None:
                msg = _("Unknown confirmation code.")
                self.errors['confirmation_code'] = [msg]
            data['confirmation'] = confirmation
        return data


class PreferredEmailField(forms.ModelChoiceField):
    def label_from_instance(self, obj):
        return obj.email


class EditAccountForm(forms.Form):
    displayname = fields.CharField(error_messages=default_errors,
        widget=ROAwareTextInput(attrs={'class': 'textType', 'size': '20'}))
    preferred_email = PreferredEmailField(
        error_messages=email_errors,
        queryset=EmailAddress.objects.none(),
        widget=ROAwareSelect,
        empty_label=None)
    password = fields.CharField(required=False,
        widget=widgets.PasswordInput(attrs={
            'class': 'disableAutoComplete textType',
            'size': '20'
            }))
    passwordconfirm = fields.CharField(required=False,
        widget=widgets.PasswordInput(attrs={
            'class': 'disableAutoComplete textType',
            'size': '20'
        }))

    def __init__(self, data=None, **kwargs):
        # These keyword arguments are required to render the form successfully
        assert 'account' in kwargs, True
        self.account = kwargs.pop('account')

        if data is None:
            kwargs['initial'] = {'displayname': self.account.displayname}

        super(EditAccountForm, self).__init__(data, **kwargs)

        if self.account.preferredemail is None:
            preferredemail_id = None
        else:
            preferredemail_id = self.account.preferredemail.id

        # Override the field to set initial value (current preferred email)
        self.fields['preferred_email'] = PreferredEmailField(
            queryset=EmailAddress.objects.filter(account=self.account).\
                exclude(status=EmailStatus.NEW).order_by('-status', 'email'),
            initial=preferredemail_id,
            widget=ROAwareSelect,
            error_messages=email_errors,
            empty_label=None)

    def clean_displayname(self):
        name = self.cleaned_data['displayname'].strip()
        if len(name) == 0:
            raise forms.ValidationError(_("Required field"))
        return name

    def clean_preferred_email(self):
        if 'preferred_email' in self.cleaned_data:
            email = self.cleaned_data['preferred_email']
            if email.status == EmailStatus.NEW:
                logger.debug("status is NEW.")
                raise forms.ValidationError(_("Please select an "
                                              "already validated address."))
            return email

    def clean_password(self):
        password = self.cleaned_data.get('password')
        if not password and (self.data.get('currentpassword') or
            self.data.get('passwordconfirm')):
            raise forms.ValidationError(_("Required field"))
        if password:
            try:
                str(password)
            except UnicodeEncodeError:
                raise forms.ValidationError(
                    _("Invalid characters in password"))
            if not password_policy_compliant(password):
                raise forms.ValidationError(_("Password must be at least "
                    "8 characters long, and must contain at least one "
                    "number and an upper case letter."))
        return password

    def clean_passwordconfirm(self):
        password = self.cleaned_data.get('passwordconfirm')
        if not password and (self.data.get('currentpassword') or
            self.data.get('password')):
            raise forms.ValidationError(_("Required field."))
        return password

    def clean(self):
        cleaned_data = self.cleaned_data
        password = cleaned_data.get('password')
        passwordconfirm = cleaned_data.get('passwordconfirm')
        if password or passwordconfirm:
            if (password != passwordconfirm and
                not self['password'].errors and
                not self['passwordconfirm'].errors):
                raise forms.ValidationError(_("Passwords didn't match"))
        return cleaned_data

    def save_account(self):
        if self.is_valid():
            self.account.displayname = self.cleaned_data['displayname']
            self.account.preferredemail = self.cleaned_data['preferred_email']
            self.account.save()
            password = self.cleaned_data['password']
            if password:
                password_obj = self.account.accountpassword
                password_obj.password = encrypt_launchpad_password(password)
                password_obj.save()
            return True
        return False


class NewEmailForm(forms.Form):
    newemail = fields.EmailField(error_messages=email_errors,
        widget=widgets.TextInput(attrs={'class': 'textType', 'size': '26'}))

    def clean_newemail(self):
        data = self.cleaned_data['newemail']
        try:
            email = EmailAddress.objects.get(email__iexact=data)
            if email.status != EmailStatus.NEW:
                raise forms.ValidationError(_("Email already registered."))
        except EmailAddress.DoesNotExist:
            return data
        return data


class PreAuthorizeForm(forms.Form):
    trust_root = forms.CharField(error_messages=default_errors)
    callback = forms.CharField(error_messages=default_errors)


class SRegRequestForm(Form):
    """A form object for user control over OpenID sreg data.
    """
    fields = {}

    @property
    def data_approved_for_request(self):
        """Return the list of sreg data approved for the request."""
        return dict([(f, self.data[f]) for f in self.data
            if self.field_approved(f)])

    def __init__(self, request, sreg_request, rpconfig, approved_data=None):
        self.request = request
        self.request_method = request.META.get('REQUEST_METHOD')
        self.sreg_request = sreg_request
        self.rpconfig = rpconfig
        self.approved_data = approved_data
        # generate initial form data
        sreg_fields = [f for f in SREG_DATA_FIELDS_ORDER if f in set(
            self.sreg_request.required + self.sreg_request.optional)]
        if rpconfig is not None:
            if rpconfig.allowed_sreg:
                fields = set(sreg_fields).intersection(
                    set(rpconfig.allowed_sreg.split(',')))
            else:
                fields = set()
        else:
            fields = sreg_fields
        self.data = self._get_sreg_data_for_user(request.user, fields)

        super(SRegRequestForm, self).__init__(self.data)
        self._init_fields(self.data)
        self.should_display = len(self.data) > 0

    def _get_sreg_data_for_user(self, user, sreg_fields):
        """Get the sreg data to ask about in the form based on the user's
        account record.
        """
        values = {}
        values['fullname'] = user.displayname
        if user.preferredemail is not None:
            values['email'] = user.preferredemail.email
        if user.person is not None:
            values['nickname'] = user.person.name
            if user.person.time_zone is not None:
                values['timezone'] = user.person.time_zone
        if user.preferredlanguage is not None:
            values['language'] = user.preferredlanguage
        else:
            values['language'] = translation.get_language_from_request(
                self.request)
        logger.debug("values (sreg_fields) = " + str(values))

        return dict([(f, values[f]) for f in sreg_fields if f in values])

    def _init_fields(self, data):
        """Initialises form fields for the user's sreg data.
        """
        for key, val in data.items():
            label = "%s: %s" % (SREG_LABELS.get(key, key), val)
            attrs = {}
            if key in self.sreg_request.required:
                attrs['class'] = 'required'
                if self.rpconfig is not None:
                    attrs['disabled'] = 'disabled'
            self.fields[key] = fields.BooleanField(
                label=label, widget=forms.CheckboxInput(attrs=attrs,
                    check_test=self.check_test))

    def check_test(self, value):
        """Determines if a checkbox should be pre-checked based on previously
        approved user data, openid request and relying party type.
        """
        for k, v in self.data.items():
            if value == v:
                value = k
                break

        if self.rpconfig and value in self.sreg_request.required:
            return True
        elif (self.approved_data and
                value in self.approved_data.get('requested', [])):
            return value in self.approved_data.get('approved', [])
        elif self.rpconfig:
            return True
        else:
            return value in self.sreg_request.required

    def field_approved(self, field):
        """Check if the field should be returned in the response based on user
        preferences and overridden for trusted relying parties.
        """
        approved = set(self.request.POST.keys())
        if self.rpconfig is not None:
            sreg_fields = set(self.sreg_request.required)
            if self.rpconfig.auto_authorize:
                # also include optional fields
                sreg_fields.update(set(self.sreg_request.optional))
            approved.update(sreg_fields)
        return field in approved


class TeamsRequestForm(Form):
    """A form object for user control over OpenID teams data.
    """

    fields = {}

    @property
    def teams_approved_by_user(self):
        """Get the list of teams actually approved by the user for the request.
        """
        if self.request_method == 'POST':
            return [t for t in self.memberships if t in self.request.POST]
        else:
            return []

    def __init__(self, request, teams_request, rpconfig, approved_data=None):
        self.request = request
        self.request_method = request.META.get('REQUEST_METHOD')
        self.teams_request = teams_request
        self.rpconfig = rpconfig
        self.approved_data = approved_data
        self.memberships = self._get_teams_for_user(
            request.user, rpconfig and rpconfig.can_query_any_team)

        super(TeamsRequestForm, self).__init__(self.memberships)

        self._init_fields(self.memberships)
        self.should_display = len(self.memberships) > 0

    def _get_teams_for_user(self, user, include_private=False):
        """Get the list of teams to ask about in the form based on the user's
        team membership.
        """
        all_teams = self.teams_request.allRequestedTeams()
        return dict([(t, t) for t in
            TeamMembership.get_team_memberships_for_user(all_teams, user,
                include_private)])

    def _init_fields(self, form_data):
        """Initialises form fields for the user's team memberships.
        """
        for team in form_data:
            self.fields[team] = fields.BooleanField(
                label=team, widget=forms.CheckboxInput(
                    check_test=self.check_test))

    def check_test(self, value):
        """Determines if a checkbox should be pre-checked based on previously
        approved user data and relying party type.
        """
        if (self.approved_data and
            value in self.approved_data.get('requested', [])):
            return value in self.approved_data.get('approved', [])
        else:
            return self.rpconfig != None
