# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import textwrap
import time
from optparse import make_option

import openid.store.nonce
from django.core.management.base import BaseCommand
from django.db import connection, transaction
from django.utils.translation import ugettext as _


class Command(BaseCommand):

    option_list = BaseCommand.option_list + (
        make_option('-s', '--sessions', dest='sessions', default=False,
            action='store_true', help='Cleanup sessions.'),
        make_option('-n', '--nonces', dest='nonces', default=False,
            action='store_true', help='Cleanup nonces.'),
        make_option('-l', '--limit', dest='limit', default=10000,
            action='store', help='Number of rows to process per batch.'),
    )
    help = _("""Clean unnecessary/stalled data from database.""")

    def handle(self, *args, **options):
        limit = int(options['limit'])
        nonce_expire_stamp = int(time.time()) - openid.store.nonce.SKEW
        queries = {'sessions': """DELETE FROM django_session
            WHERE session_key = ANY(SELECT session_key
                                    FROM django_session
                                    WHERE expire_date < CURRENT_TIMESTAMP
                                    LIMIT %s)""" % limit,
            'nonces': """DELETE FROM openidnonce
            WHERE timestamp = ANY(SELECT timestamp
                                  FROM openidnonce
                                  WHERE timestamp < %s
                                  LIMIT %s)""" % (nonce_expire_stamp, limit),
        }
        verbosity = int(options['verbosity'])

        selected_queries = [query for query in queries if options.get(query)]
        if not selected_queries:
            self.stdout.write(textwrap.dedent("""
                No items selected to clean up. Please select at least one of

                    --sessions
                    --nonces
            """))

        for item in selected_queries:
            if verbosity >= 1:
                self.stdout.write("\nCleaning %s..." % item)
            cursor = connection.cursor()
            cursor.execute(queries[item])
            while cursor.rowcount > 0:
                if verbosity >= 2:
                    self.stdout.write(".")
                cursor.execute(queries[item])
            transaction.commit_unless_managed()
