# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import logging
from datetime import date, datetime, timedelta
from django.conf import settings
from django.contrib.auth import get_backends
from django.contrib.auth.models import User
from django.db import models
from django.utils.translation import ugettext_lazy as _

from oauth_backend.models import Consumer

from identityprovider.models.const import (AccountCreationRationale,
                                           AccountStatus, EmailStatus)
from identityprovider.utils import (
    encrypt_launchpad_password,
    generate_openid_identifier,
    get_object_or_none,
)

__all__ = [
    'Account',
    'AccountPassword',
    'DisplaynameField',
    'PasswordField',
    'LPOpenIdIdentifier',
]


class AccountManager(models.Manager):
    """ Extend the default manager with a method that takes care of creating
    an Account, and the necessary AccountPassword, EmailAddress objects. """

    def create_account(self, displayname, username, password,
        creation_rationale=None, salt=None, password_encrypted=False,
            email_validated=True):
        from identityprovider.models import EmailAddress
        if creation_rationale is None:
            creation_rationale = (
                AccountCreationRationale.OWNER_CREATED_LAUNCHPAD)
        account = Account.objects.create(
            displayname=displayname,
            creation_rationale=creation_rationale,
            status=AccountStatus.ACTIVE)
        email_status = (
            EmailStatus.PREFERRED if email_validated else EmailStatus.NEW)
        EmailAddress.objects.create(
            email=username,
            account=account,
            status=email_status)
        if not password_encrypted:
            password = encrypt_launchpad_password(password, salt=salt)
        password = AccountPassword.objects.create(
            account=account,
            password=password)
        return account

    def get_by_email(self, email):
        return get_object_or_none(Account.objects.select_related(),
                                  emailaddress__email__iexact=email)


class DisplaynameField(models.TextField):
    def __init__(self, null=False):
        super(DisplaynameField, self).__init__(null=null)


class PasswordField(models.TextField):
    def __init__(self, null=False):
        super(PasswordField, self).__init__(null=null)


class Account(models.Model):
    date_created = models.DateTimeField(default=datetime.utcnow,
                                        editable=False)
    creation_rationale = \
            models.IntegerField(
                choices=AccountCreationRationale._get_choices())

    status = models.IntegerField(choices=AccountStatus._get_choices())
    date_status_set = models.DateTimeField(default=datetime.utcnow)
    displayname = DisplaynameField()
    openid_identifier = models.TextField(default=generate_openid_identifier)
    status_comment = models.TextField(blank=True, null=True)
    preferredlanguage = models.TextField(blank=True, null=True)
    old_openid_identifier = models.TextField(blank=True, null=True)

    objects = AccountManager()

    class Meta:
        app_label = 'identityprovider'
        db_table = u'account'
        verbose_name = _('account')
        verbose_name_plural = _('accounts')

    def __unicode__(self):
        return self.displayname

    @property
    def openid_identity_url(self):
        return '%s+id/%s' % (settings.SSO_ROOT_URL, self.openid_identifier)

    def verified_emails(self):
        """Returns all verified e-mail addresses (including
        preferred), with preferred sort first."""
        email = self.emailaddress_set.filter(
            status__in=[EmailStatus.PREFERRED, EmailStatus.VALIDATED])
        return email.order_by('-status', 'email')

    def unverified_emails(self):
        email = self.emailaddress_set.filter(status=EmailStatus.NEW)
        return email.order_by('email')

    def last_authenticated_sites(self, limit=10):
        sites = self.openidrpsummary_set.order_by('-date_last_used')
        return sites if limit is None else sites[:limit]

    def sites_with_active_sessions(self, age_hours=24):
        """
        Return list of sites on which user *may* have active sessions

        Current rule of thumb is to return all sites accessed within previous
        24 hours.
        """
        max_date_last_used = (datetime.utcnow() -
                              timedelta(hours=age_hours))
        sites = self.openidrpsummary_set.filter(
            date_last_used__gte=max_date_last_used)
        return sites.order_by('-date_last_used')

    def _get_preferredemail(self):
        # import here to avoid circular dependancy issue
        from identityprovider.models.emailaddress import EmailAddress
        if not hasattr(self, '_preferredemail'):
            try:
                email = None
                account_emails = self.emailaddress_set.filter(
                    status=EmailStatus.PREFERRED)
                if account_emails.count() > 0:
                    email = account_emails[0]
                else:
                    # b/c of launchpad data issues, a user can have a preferred
                    # email for the lp_person but not this account, so check
                    # them too
                    from .person import Person
                    persons = set(Person.objects.filter(
                        emailaddress__account=self.id)) 
                    # could in theory not include person linked to be openid
                    # so include that just in case
                    if self.person:
                        persons.add(self.person)
                    emails = EmailAddress.objects.filter(
                            lp_person__in=persons,
                            status=EmailStatus.PREFERRED)
                    if emails.count() > 0:
                        email = emails[0]
                        logging.warn("Old LP email used as preferred email, " 
                                     "needs migrating. Account id: %s" % self.id)

                if not email:
                    # Try to determine a suitable address, and mark it as preferred
                    # note: we don't try to make a person-based email preferred
                    emails = self.emailaddress_set.filter(
                        status=EmailStatus.VALIDATED)
                    email = emails.order_by('date_created')[0]
                    email.status = EmailStatus.PREFERRED
                    email.save()
                    logging.info(
                        "Updating preferred email for account %s" % self.id)

                self._preferredemail = email
            except:
                # no error in preffered email should raise, as this breaks
                # login completely
                self._preferredemail = None
        return self._preferredemail

    def _set_preferredemail(self, email):
        current = self.preferredemail
        if current is not None:
            current.status = EmailStatus.VALIDATED
            current.save()
        email.status = EmailStatus.PREFERRED
        email.save()
        self._preferredemail = email

    preferredemail = property(_get_preferredemail, _set_preferredemail)

    @property
    def person(self):
        if not hasattr(self, '_person'):
            open_ids = LPOpenIdIdentifier.objects.filter(
                identifier=self.openid_identifier)
            if len(open_ids) > 0:
                # Look up Person object with the same lp_account id as it's in
                # lp_OpenIdIdentifier table
                lp_account = open_ids[0].lp_account
                # Importing it here to prevent cyclic import issue
                from .person import Person
                try:
                    self._person = Person.objects.select_related(
                        'personlocation').get(lp_account=lp_account)
                except Person.DoesNotExist:
                    pass
        return getattr(self, '_person', None)

    @property
    def user(self):
        if not hasattr(self, '_user'):
            # When in read only mode we don't want to create users so we'll
            # return what is or None if not there
            if settings.READ_ONLY_MODE:
                self._user = get_object_or_none(
                    User, username=self.openid_identifier)
            else:
                if self.preferredemail:
                    self._user, _ = User.objects.get_or_create(
                        username=self.openid_identifier,
                        defaults={'email': self.preferredemail})
                else:
                    return None

        return self._user

    def _get_last_login(self):
        if self.user is not None:
            return self.user.last_login

    def _set_last_login(self, login_datetime):
        # we can live without last_login
        if settings.READ_ONLY_MODE:
            return

        if self.user is not None:
            self.user.last_login = login_datetime
            self.user.save()

    last_login = property(_get_last_login, _set_last_login)

    @property
    def can_reactivate(self):
        return (self.status == AccountStatus.DEACTIVATED
            or self.status == AccountStatus.NOACCOUNT)

    @property
    def can_reset_password(self):
        return (self.is_active or self.can_reactivate)

    @property
    def is_active(self):
        return (self.status == AccountStatus.ACTIVE)

    def is_authenticated(self):
        return True

    def person_in_team(self, team):
        return self.person is not None and self.person.in_team(team)

    @property
    def is_staff(self):
        # Logging into the admin system with this breaks things
        return False

    @property
    def is_superuser(self):
        return False

    @property
    def first_name(self):
        sname = self.displayname.split(' ')
        return sname[0]

    def get_and_delete_messages(self):
        return []

    def get_full_name(self):
        return self.displayname

    def has_module_perms(self, app_label):
        if not self.is_active:
            return False
        for backend in get_backends():
            if hasattr(backend, "has_module_perms"):
                if backend.has_module_perms(self, app_label):
                    return True
        return False

    def has_perm(self, perm):
        if not self.is_active:
            return False
        for backend in get_backends():
            if hasattr(backend, "has_perm"):
                if backend.has_perm(self, perm):
                    return True
        return False

    def create_oauth_token(self, token_name):
        user, _ = User.objects.get_or_create(username=self.openid_identifier)
        try:
            consumer = user.oauth_consumer
        except Consumer.DoesNotExist:
            consumer = Consumer.objects.create(user=user)
        token = consumer.token_set.create(name=token_name)
        return token

    def oauth_tokens(self):
        user, _ = User.objects.get_or_create(username=self.openid_identifier)
        try:
            consumer = user.oauth_consumer
            return consumer.token_set.all()
        except Consumer.DoesNotExist:
            return []

    def save(self, force_insert=False, force_update=False, **kwargs):
        if settings.READ_ONLY_MODE:
            return

        # if we suspend the account, reset the password too, to force the user
        # to reset it after the account is re-enabled
        if self.status == AccountStatus.SUSPENDED:
            try:
                # by setting a plain text value we make sure its not going to
                # match against anything
                self.accountpassword.password = 'invalid'
                self.accountpassword.save()
            except AccountPassword.DoesNotExist:
                # no password, so nothing to reset
                pass

        # now go and effectively save it
        super(Account, self).save(force_insert, force_update, **kwargs)


class AccountPassword(models.Model):
    account = models.OneToOneField(Account, db_column='account')
    password = PasswordField()

    class Meta:
        app_label = 'identityprovider'
        db_table = u'accountpassword'
        verbose_name = _('account password')
        verbose_name_plural = _('account passwords')

    def __unicode__(self):
        return _("Password for %s") % unicode(self.account)


class LPOpenIdIdentifier(models.Model):
    """
    Mapping between Launchpad accounts and OpenID identifiers
    """
    identifier = models.TextField(unique=True, primary_key=True)
    lp_account = models.IntegerField(null=False, db_column='account')
    date_created = models.DateTimeField(null=False, default=date.today)

    class Meta:
        app_label = 'identityprovider'
        db_table = u'lp_openididentifier'
        verbose_name = _("LP OpenID Identifier")
        verbose_name_plural = _("LP OpenID Identifiers")

    def __unicode__(self):
        return _("LP OpenID Identifier for %s") % unicode(self.lp_account)
