# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import datetime
import random
import re
from urlparse import urljoin

from django.conf import settings
from django.core.mail import send_mail
from django.core.urlresolvers import reverse
from django.db import models
from django.http import urlencode
from django.template.loader import render_to_string
from django.utils.translation import ugettext_lazy as _

from identityprovider.branding import current_brand
from identityprovider.models.const import LoginTokenType
from identityprovider.models import Account, DisplaynameField, PasswordField
from identityprovider.utils import format_address

__all__ = (
    'AUTHTOKEN_LENGTH',
    'AUTHTOKEN_PATTERN',
    'AuthToken',
    'AuthTokenFactory',
    'get_type_of_token',
    'send_validation_email_request',
    'verify_token_string'
)


AUTHTOKEN_LENGTH = 6

# For now, we have some older tokens (20 chars) still active, while we
# introduce the shorter tokens.  After a while, we could probably
# bound this pattern's repetition more tightly.
#AUTHTOKEN_PATTERN = '[A-Za-z0-9]{%s}' % AUTHTOKEN_LENGTH
AUTHTOKEN_PATTERN = '[A-Za-z0-9]+'  # % AUTHTOKEN_LENGTH


class AuthToken(models.Model):
    date_created = models.DateTimeField(default=datetime.datetime.utcnow,
        blank=True, editable=False)
    date_consumed = models.DateTimeField(blank=True, null=True)
    token_type = models.IntegerField(choices=LoginTokenType._get_choices())
    token = models.TextField(unique=True)
    requester = models.ForeignKey(Account, db_column='requester',
        null=True, blank=True)
    requester_email = models.TextField(null=True, blank=True)
    email = models.TextField(db_index=True)
    redirection_url = models.TextField(null=True, blank=True)

    displayname = DisplaynameField(null=True)
    password = PasswordField(null=True)

    title = 'Launchpad Email Verification'

    class Meta:
        app_label = 'identityprovider'
        db_table = u'authtoken'

    def __unicode__(self):
        return self.token

    def get_absolute_url(self):
        path = reverse('identityprovider.views.ui.claim_token',
            kwargs={'authtoken': self.token})
        return urljoin(settings.SSO_ROOT_URL, path) + '?' + \
               urlencode({'email': self.email})

    def get_direct_absolute_url(self):
        path = reverse('identityprovider.views.ui.confirm_account',
                       kwargs={'authtoken': self.token,
                               'email_address': self.email})
        return urljoin(settings.SSO_ROOT_URL, path)

    def consume(self):
        self.date_consumed = datetime.datetime.utcnow()
        self.save()

        result = AuthToken.objects.filter(
            email=self.email, token_type=self.token_type,
            requester=self.requester,
            date_consumed=None)
        for token in result:
            token.date_consumed = datetime.datetime.utcnow()
            token.save()

    def _send_email(self, from_name, subject, message):
        """Send an email to this token's email address."""
        from_address = format_address(
            settings.NOREPLY_FROM_ADDRESS, from_name)
        to_address = format_address(self.email)
        send_mail(subject, message, from_address, [to_address])

    def sendEmailValidationRequest(self):
        replacements = {
            'requester': self.requester.displayname,
            'requester_email': self.requester_email,
            'toaddress': self.email,
            'token': self.token,
            'token_url': self.get_absolute_url(),
            'support_form_url': settings.SUPPORT_FORM_URL,
        }
        message = render_to_string('%s/email/validate-email.txt' %
            current_brand.template_dir, replacements)
        subject = u'%s: %s' % (current_brand.name,
                               _(": Validate your email address"))
        self._send_email(u"%s Email Validator" % current_brand.name, subject,
                         message)

    def sendEmailValidationToken(self, template='email-validation-token.txt'):
        replacements = {
            'token': self.token,
            'toaddress': self.email,
        }
        template_path = '%s/email/%s' % (current_brand.template_dir, template)
        message = render_to_string(template_path, replacements)
        subject = u"%s: %s" % (current_brand.name,
                               _("Validate your email address"))
        self._send_email(u"%s Email Validator" % current_brand.name, subject,
                         message)

    def sendPasswordResetEmail(self):
        replacements = {
            'token': self.token,
            'token_url': self.get_absolute_url(),
        }
        from_name = unicode(current_brand.name)
        message = render_to_string('%s/email/forgottenpassword.txt' %
            current_brand.template_dir, replacements)
        subject = u"%s: %s" % (current_brand.name, _("Forgotten Password"))
        self._send_email(from_name, subject, message)

    # {{workflow}} remove the `old` parameter
    def sendNewUserEmail(self, template='web-newuser.txt', old=False):
        url = self.get_absolute_url() if old else self.get_direct_absolute_url()
        replacements = {
            'token': self.token,
            'token_url': url,
            'toaddress': self.email,
        }
        from_name = unicode(current_brand.name)
        template_path = '%s/email/%s' % (current_brand.template_dir, template)
        message = render_to_string(template_path, replacements)
        subject = u"%s: %s" % (current_brand.name,
                               _("Finish your registration"))
        self._send_email(from_name, subject, message)


class InvalidEmailException(Exception):
    pass


class AuthTokenFactory:
    title = 'Launchpad email address confirmation'

    def new_api_email_validation_token(self, requester, email,
            redirection_url=None):
        return self.new(requester, None, email,
                        LoginTokenType.VALIDATEEMAIL, redirection_url)

    def new(self, requester, requester_email, email, token_type,
            redirection_url, **kwargs):
        if not valid_email(email):
            raise InvalidEmailException()

        if token_type not in [LoginTokenType.PASSWORDRECOVERY,
                               LoginTokenType.NEWACCOUNT,
                               LoginTokenType.VALIDATEEMAIL,
                               LoginTokenType.NEWPERSONLESSACCOUNT]:
            raise ValueError(
                "token_type is not a valid AuthToken type: %s" %
                token_type)

        if isinstance(redirection_url, str):
            redirection_url = unicode(redirection_url)

        token = AuthToken(requester=requester,
                          requester_email=requester_email,
                          email=email, token_type=token_type,
                          redirection_url=redirection_url,
                          **kwargs)
        token.token = create_unique_token_for_table(
            AUTHTOKEN_LENGTH, AuthToken, 'token')
        token.save()
        return token


def create_token(token_length):
    """Create a random token string.

    :param token_length: Specifies how long you want the token.
    """
    # Since tokens are, in general, user-visible, vowels are not included
    # below to prevent them from having curse/offensive words.
    characters = '0123456789bcdfghjklmnpqrstvwxzBCDFGHJKLMNPQRSTVWXZ'
    token = ''.join(
        random.choice(characters) for count in range(token_length))
    return unicode(token)


def create_unique_token_for_table(token_length, obj, column):
    """Create a new unique token in a table.

    Generates a token and makes sure it does not already exist in
    the table and column specified.

    :param token_length: The length for the token string
    :param column: Database column where the token will be stored.

    :return: A new token string
    """
    token = create_token(token_length)
    try:
        params = {column: token}
        while obj.objects.get(**params) is not None:
            token = create_token(token_length)
            params = {column: token}
    except Exception:
        pass
    return token


def valid_email(emailaddr):
    """Validate an email address.

    >>> valid_email('kiko.async@hotmail.com')
    True
    >>> valid_email('kiko+async@hotmail.com')
    True
    >>> valid_email('kiko-async@hotmail.com')
    True
    >>> valid_email('kiko_async@hotmail.com')
    True
    >>> valid_email('kiko@async.com.br')
    True
    >>> valid_email('kiko@canonical.com')
    True
    >>> valid_email('kiko@UBUNTU.COM')
    True
    >>> valid_email('i@tv')
    True
    >>> valid_email('kiko@gnu.info')
    True
    >>> valid_email('user@z.de')
    True
    >>> valid_email('bob=dobbs@example.com')
    True

    >>> valid_email('user@z..de')
    False
    >>> valid_email('user@.z.de')
    False

    As per OOPS-256D762:

    >>> valid_email('keith@risby-family.co.uk')
    True
    >>> valid_email('keith@risby-family-.co.uk')
    False
    >>> valid_email('keith@-risby-family.co.uk')
    False
    """
    email_re = r"^[_\.0-9a-zA-Z-+=]+@(([0-9a-zA-Z-]{1,}\.)*)[a-zA-Z]{2,}$"
    email_match = re.match(email_re, emailaddr)
    if not email_match:
        return False
    host_minus_tld = email_match.group(1)
    if not host_minus_tld:
        return True
    for part in host_minus_tld.split("."):
        if part.startswith("-") or part.endswith("-"):
            return False
    return True


def get_type_of_token(token_string):
    try:
        token = AuthToken.objects.get(token=token_string)
        return token.token_type
    except AuthToken.DoesNotExist:
        return None


def send_validation_email_request(account, email, redirection_url):
    if account.preferredemail is None:
        preferredemail_email = None
    else:
        preferredemail_email = account.preferredemail.email

    token = AuthTokenFactory().new(
        requester=account,
        requester_email=preferredemail_email,
        email=email,
        token_type=LoginTokenType.VALIDATEEMAIL,
        redirection_url=redirection_url)
    token.sendEmailValidationRequest()

    return token


def verify_token_string(token_string, email):
    try:
        return AuthToken.objects.get(token=token_string, email=email,
                                     date_consumed=None)
    except AuthToken.DoesNotExist:
        return None
