# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from django.db import models

from identityprovider.models import (
    Account, Country, Language, LibraryFileAlias, LPOpenIdIdentifier,
)
from identityprovider.models.const import AccountCreationRationale

__all__ = (
    'Person',
)


class Person(models.Model):
    displayname = models.TextField()
    teamowner = models.ForeignKey('self', db_column='teamowner',
                                  related_name='owned_teams',
                                  null=True, blank=True)
    teamdescription = models.TextField(null=True, blank=True)
    name = models.TextField(unique=True)
    language = models.ForeignKey(Language, db_column='language',
        null=True, blank=True)
    fti = models.TextField()  # This field type is a guess.
    defaultmembershipperiod = models.IntegerField(null=True, blank=True)
    defaultrenewalperiod = models.IntegerField(null=True, blank=True)
    subscriptionpolicy = models.IntegerField(default=1)
    merged = models.ForeignKey('self', db_column='merged',
                               related_name='merged_personas',
                               null=True, blank=True)
    datecreated = models.DateTimeField(auto_now_add=True)
    addressline1 = models.TextField(null=True, blank=True)
    addressline2 = models.TextField(null=True, blank=True)
    organization = models.TextField(null=True, blank=True)
    city = models.TextField(null=True, blank=True)
    province = models.TextField(null=True, blank=True)
    country = models.ForeignKey(Country, db_column='country',
        null=True, blank=True)
    postcode = models.TextField(null=True, blank=True)
    phone = models.TextField(null=True, blank=True)
    homepage_content = models.TextField(null=True, blank=True)
    icon = models.ForeignKey(LibraryFileAlias, db_column='icon',
                             related_name='person_icons',
                             null=True, blank=True)
    mugshot = models.ForeignKey(LibraryFileAlias, db_column='mugshot',
                                related_name='person_mugshots',
                                null=True, blank=True)
    hide_email_addresses = models.BooleanField(default=False)
    creation_rationale = models.IntegerField(null=True, blank=True,
        choices=AccountCreationRationale._get_choices())
    creation_comment = models.TextField(null=True, blank=True)
    registrant = models.ForeignKey('self', db_column='registrant',
                                   related_name='registered_teams',
                                   null=True, blank=True)
    logo = models.ForeignKey(LibraryFileAlias, db_column='logo',
                             related_name='person_logos',
                             null=True, blank=True)
    renewal_policy = models.IntegerField(default=10)
    personal_standing = models.IntegerField(default=0)
    personal_standing_reason = models.TextField(null=True, blank=True)
    mail_resumption_date = models.DateField(null=True, blank=True)
    mailing_list_auto_subscribe_policy = models.IntegerField(default=1)
    mailing_list_receive_duplicates = models.BooleanField(default=True)
    visibility = models.IntegerField(default=1)
    verbose_bugnotifications = models.BooleanField(default=False)
    lp_account = models.IntegerField(null=True, db_column='account')

    class Meta:
        app_label = 'identityprovider'
        db_table = u'lp_person'

    def __unicode__(self):
        return self.displayname

    def in_team(self, team):
        from identityprovider.models import TeamParticipation
        # Just to be fully compatible with
        # lp:lib/registry/model/person.py:Person.inTeam()
        if isinstance(team, (str, unicode)):
            try:
                team = Person.objects.get(name=team)
            except Person.DoesNotExist:
                return False
        try:
            TeamParticipation.objects.get(team=team, person=self)
            return True
        except TeamParticipation.DoesNotExist:
            if self.id == team.teamowner.id:
                return True
        return False

    @property
    def account(self):
        if self.lp_account is not None:
            open_ids = LPOpenIdIdentifier.objects.filter(
                lp_account=self.lp_account)
            if len(open_ids) > 0:
                openid_identifier = open_ids[0].identifier
                accounts = Account.objects.filter(
                    openid_identifier=openid_identifier)
                if len(accounts) > 0:
                    return accounts[0]

    def is_team(self):
        return self.teamowner is not None

    @property
    def time_zone(self):
        try:
            return self.personlocation.time_zone
        except:
            return None


class PersonLocation(models.Model):
    date_created = models.DateTimeField()
    person = models.OneToOneField(Person, db_column='person')
    latitude = models.FloatField(null=True, blank=True)
    longitude = models.FloatField(null=True, blank=True)
    time_zone = models.TextField(null=True, blank=True)
    last_modified_by = models.ForeignKey(Person, db_column='last_modified_by',
        related_name='last_modified_locations')
    date_last_modified = models.DateTimeField()
    visible = models.BooleanField(default=True)
    locked = models.BooleanField(default=False)

    class Meta:
        app_label = 'identityprovider'
        db_table = u'lp_personlocation'
