# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from django.db import models

from identityprovider.const import PERSON_VISIBILITY_PUBLIC
from identityprovider.models import Person

__all__ = (
    'TeamMembership',
    'TeamParticipation',
)


class TeamMembership(models.Model):
    person = models.ForeignKey(Person, db_column='person',
                               related_name='team_memberships')
    team = models.ForeignKey(Person, db_column='team',
                             related_name='team_members')
    status = models.IntegerField()
    date_joined = models.DateTimeField(auto_now_add=True)
    date_expires = models.DateTimeField()
    last_changed_by = models.ForeignKey(Person, db_column='last_changed_by',
                                        related_name='changed_memberships',
                                        null=True, blank=True)
    last_change_comment = models.TextField(null=True, blank=True)
    proposed_by = models.ForeignKey(Person, db_column='proposed_by',
                                    related_name='proposed_memberships',
                                    null=True, blank=True)
    acknowledged_by = models.ForeignKey(
        Person, db_column='acknowledged_by',
        related_name='acknowledged_memberships',
        null=True, blank=True)
    reviewed_by = models.ForeignKey(Person, db_column='reviewed_by',
                                    related_name='reviewed_memberships',
                                    null=True, blank=True)
    date_proposed = models.DateTimeField(null=True, blank=True)
    date_last_changed = models.DateTimeField(null=True, blank=True)
    date_acknowledged = models.DateTimeField(null=True, blank=True)
    date_reviewed = models.DateTimeField(null=True, blank=True)
    proponent_comment = models.TextField(null=True, blank=True)
    acknowledger_comment = models.TextField(null=True, blank=True)
    reviewer_comment = models.TextField(null=True, blank=True)
    date_created = models.DateTimeField(auto_now_add=True)

    class Meta:
        app_label = 'identityprovider'
        db_table = u'teammembership'
        unique_together = ('person', 'team')

    @staticmethod
    def get_team_memberships_for_user(team_names, user, include_private=False):
        teams = Person.objects.filter(
            name__in=team_names,      # All team names provided
            teamowner__isnull=False,  # Only teams
        )
        if not include_private:
            teams = teams.filter(
                visibility=PERSON_VISIBILITY_PUBLIC)
        # Only teams in which the user is a member, which can be either through
        # direct membership or by being a team owner.
        teams = teams.filter(
            models.Q(team_participations__person=user.person) |
            models.Q(teamowner__id=models.F('id')),
        )
        # By default it's a query set result, not a true list
        return list(teams.values_list('name', flat=True))


class TeamParticipation(models.Model):
    team = models.ForeignKey(Person, db_column='team',
                             related_name='team_participations')
    person = models.ForeignKey(Person, db_column='person',
                               related_name='person_participations')

    class Meta:
        app_label = 'identityprovider'
        db_table = u'lp_teamparticipation'
        unique_together = ('team', 'person')

    def __unicode__(self):
        return "%s in %s" % (unicode(self.person), unicode(self.team))
