from configglue.schema import (
    BoolOption,
    Section,
    DictOption,
    IntOption,
    ListOption,
    StringOption,
    TupleOption,
)
from django_configglue.schema import schemas


DjangoSchema = schemas.get('1.3')


class UpperCaseDictOption(DictOption):
    """ A DictOption with all upper-case keys. """
    def parse(self, section, parser=None, raw=False):
        parsed = super(UpperCaseDictOption, self).parse(
            section, parser, raw)
        result = {}
        for k, v in parsed.items():
            result[k.upper()] = v
        return result


class SSOSchema(DjangoSchema):
    # default
    branded_template_dir = StringOption()
    db_connections = ListOption(item=UpperCaseDictOption(
        spec={'database_id': StringOption(),
              'database_host': StringOption(),
              'database_name': StringOption(),
              'database_user': StringOption(),
              'database_password': StringOption(),
              'database_port': StringOption()},
        strict=True))
    db_statement_timeout_millis = IntOption(
        default=0, help='0 disables timeout')
    extra_pythonpath = ListOption(item=StringOption())
    handler_timeout_millis = IntOption(default=10000)
    oops_dir = StringOption(help='Absolute path to the directory oops'
        ' reports will be stored in')
    pgconnect_timeout = IntOption(default=10)
    serve_static_media = BoolOption()
    support_form_url = StringOption()
    test_dsn = StringOption()
    google_analytics_id = StringOption()

    #i18n
    class i18n(Section):
        supported_languages = ListOption(item=StringOption(),
            help="List of currently supported languages")
        language_names = DictOption(
            help="Mapping between language-code and name in that language")

    # api
    class api(Section):
        api_enabled = BoolOption()
        api_host = StringOption()
        oauth_data_store = StringOption()

    # branding
    class branding(Section):
        brand = StringOption()

    # captcha
    class captcha(Section):
        captcha_use_proxy = BoolOption()
        captcha_proxies = DictOption()
        captcha_api_url = StringOption()
        captcha_api_url_secure = StringOption()
        captcha_image_url_pattern = StringOption(raw=True)
        captcha_verify_url = StringOption()
        captcha_public_key = StringOption()
        captcha_private_key = StringOption()
        disable_captcha_verification = BoolOption()
        email_whitelist_regexp_list = ListOption(
            item=StringOption(raw=True)
        )

    # debug
    class debug(Section):
        openid_debug = BoolOption()
        testing = BoolOption()

    # email
    class email(Section):
        noreply_from_address = StringOption()
        feedback_to_address = StringOption()
        enable_token_debug = BoolOption()

    # openid
    class openid(Section):
        pre_authorization_validity = IntOption()
        openid_preauthorization_acl = ListOption(
            item=TupleOption(length=2))
        openid_create_users = BoolOption()
        openid_launchpad_teams_mapping = DictOption()
        openid_sso_server_url = StringOption()
        openid_set_language_from_sreg = BoolOption()
        openid_sreg_extra_fields = ListOption(item=StringOption())
        openid_launchpad_staff_teams = ListOption(item=StringOption())

    # preflight
    class preflight(Section):
        preflight_base_template = StringOption()

    # readonly
    class readonly(Section):
        app_servers = ListOption(item=UpperCaseDictOption(
            spec={'server_id': StringOption(),
                  'host': StringOption(),
                  'port': StringOption(),
                  'scheme': StringOption(default='http'),
                  'virtual_host': StringOption(),
                  },
            strict=True))
        readonly_secret = StringOption(raw=True)
        dbfailover_attempts = IntOption()
        dbfailover_flag_dir = StringOption()
        dbrecover_interval = IntOption()
        dbrecover_attempts = IntOption()
        dbrecover_multiplier = IntOption()
        read_only_mode = BoolOption(default=False)

    # saml
    # Settings from upstream:
    class saml2(Section):
        saml2idp_autosubmit = BoolOption(default=True)
        saml2idp_issuer = StringOption()
        saml2idp_certificate_file = StringOption()
        saml2idp_private_key_file = StringOption()
        saml2idp_signing = BoolOption(default=True)
        saml2idp_valid_acs = ListOption(
            item=StringOption(),
            help="List of ACS URLs accepted by /+saml login")
        saml2idp_processor_classes = ListOption(
            item=StringOption(),
            help="List of SAML 2.0 AuthnRequest processors")
        # Settings specific to canonical-identity-provider:
        saml2_teams = ListOption(
            item=StringOption(),
            help="List of teams able to login using SAML")

    # sso
    class sso(Section):
        sso_media_root = StringOption()
        sso_restrict_rp = BoolOption()
        sso_root_url = StringOption()
        sso_provider_url = StringOption()
        sso_versions_teams = ListOption(
            item=StringOption(),
            help="List of teams able to access /+versions page")
        sso_account_update_url = StringOption(
            null=True,
            help="URL to POST when account data has changed")
        embedded_trust_root = StringOption()

    # yubico
    class yubico(Section):
        yubico_client_id = StringOption(fatal=True)
        yubico_api_url = StringOption(
            default=('http://api.yubico.com/wsapi/2.0/verify?id={client_id}&'
                     'otp={otp}&nonce={nonce}'))

    class profile(Section):
        profile_filename = StringOption(
            default="sso.profile",
            help="Filename of the file created by the profile middleware")
        profile_pattern = StringOption(
            default="identityprovider\..*",
            help="Regex pattern for limiting functions to be considered for profiling")

    class devserver(Section):
        devserver_ignored_prefixes = ListOption(
            item=StringOption(),
            help="List of url prefixes which will be ignored")
        devserver_modules = ListOption(
            item=StringOption(),
            help="List of devserver modules to enable")
        devserver_default_addr = StringOption(default="127.0.0.1")
        devserver_default_port = StringOption(default="8000")
        devserver_truncate_sql = BoolOption(default=True)
        devserver_ajax_content_length = IntOption(default=300)
        devserver_auto_profile = BoolOption(default=False)

    class statsd(Section):
        statsd_host = StringOption(default='localhost')
        statsd_port = IntOption(default=8125)

    # settings for testing (acceptance, api)
    class testing(Section):
        test_account_email = StringOption()
        test_account_password = StringOption()
        qa_account_email = StringOption()
        qa_account_password = StringOption()
        email_address_pattern = StringOption(raw=True)
        imap_server = StringOption()
        imap_username = StringOption()
        imap_password = StringOption()
        test_discovery_root = StringOption()

    class django_jenkins(Section):
        project_apps = ListOption()
        jenkins_tasks = TupleOption(
            default=(
                'django_jenkins.tasks.run_pylint',
                'django_jenkins.tasks.with_coverage',
                'django_jenkins.tasks.django_tests',
            ))
        jenkins_test_runner = StringOption()

    # override django section
    class django(DjangoSchema.django):
        caches = DictOption(item=UpperCaseDictOption())
