#!/usr/bin/env python

import quopri
import re

from django.conf import settings
from canonical.isd.tests.sst import mail
from sst.actions import *


def register_account(email_address=None, displayname="My Name",
                     password="Admin007", fetch_email=True, verify=True):
    """Register an account and verify the token by default.
    If `register_account` creates the email address for you then it returns
    the email address.

    If you pass in an `email_address` then the verification token is returned.
    """
    return_email = False
    if email_address is None:
        return_email = True
        email_address = mail.make_unique_test_email_address()

    vcode = None
    go_to('/+new_account')
    wait_for(assert_title, 'Create account')
    write_textfield('id_displayname', displayname)
    write_textfield('id_email', email_address)
    write_textfield('id_password', password)
    write_textfield('id_passwordconfirm', password)
    # Even though the recaptcha field is ignored for our tests, we do
    # want to verify that it is on the page.
    write_textfield('recaptcha_response_field', 'ignored')
    click_button(get_element(name='continue'))
    wait_for(assert_title_contains, 'Registration mail sent')
    if fetch_email:
        vcode = get_verification_code_for_address(email_address)
        if verify:
            write_textfield(get_element(name='confirmation_code'), vcode)
            click_button(get_element(css_class='btn'))
    if return_email:
        return email_address
    return vcode


def add_email(address, verify=False):
    go_to('/+emails')
    wait_for(assert_title_contains, "'s email addresses")
    write_textfield('id_newemail', address)
    click_button(get_element(name='continue'))
    code = get_verification_code_for_address(address)
    if verify:
        go_to('/+enter_token')
        wait_for(write_textfield, 'id_confirmation_code', code)
        write_textfield('id_email', address)
        click_button(get_element(name='continue'))
        assert_title('Complete email address validation')
        click_button(get_element(name='continue'))
    return code


def skip_local():
    """Skip a test if running locally"""
    base = get_base_url()
    if 'localhost' in base:
        print "This test can't run locally."
        skip()


def skip_production():
    if get_base_url() == 'https://login.ubuntu.com/':
        print 'This test should not run against production.'
        skip()


def production_only():
    if get_base_url() != 'https://login.ubuntu.com/':  #Production
        print 'This is a test only for Production.'
        skip()


def delete_email():
    go_to('/+emails')
    remove_link = get_element(tag='a', text='Remove')
    click_link(remove_link)
    confirm_button = get_element(tag='button', text='Yes, remove')
    click_button(confirm_button)


def try_to_validate_email(address, code):
    go_to('/+enter_token')
    wait_for(write_textfield, 'id_confirmation_code', code)
    write_textfield('id_email', address)
    click_button(get_element(name='continue'))


def login(email, password):
    go_to('/')
    write_textfield('id_email', email)
    write_textfield('id_password', password)
    click_button(get_element(name='continue'))


def logout():
    go_to('/+logout')


def _get_verification_data_for_address(email_address):
    """A private helper for public helpers below.

    Note: We have two different public helpers here for verification
    code and link so that functional tests don't need to deal with
    idioms like:
        vcode, ignored = get_verification_for_address(email_address).
    """
    email_msg = mail.get_latest_email_sent_to(email_address)
    vcode = link = None
    if email_msg:
        # The body is encoded as quoted-printable.  This affects any line longer
        # than a certain length.  Decode now to not have to worry about it in
        # the regexen.
        body = quopri.decodestring(email_msg.get_payload())
        match = re.search(
            'Here is your confirmation code:(.*)(Enter|If you made)',
            body, re.S)
        if match:
            vcode = match.group(1).strip()
        else:
            raise AssertionError(
                "No verification code found in email.")
        match = re.search(
            'confirm your (?:account|email address|reset):(.*)If',
            body, re.S)
        if match:
            link = match.group(1).strip()
        else:
            raise AssertionError(
                "No verification link found in email.")
    return vcode, link


def get_verification_code_for_address(email_address):
    print("Retrieving verification code for %s." % email_address)
    vcode, link = _get_verification_data_for_address(email_address)
    print("Verification code retrieved: %s." % vcode)
    return vcode


def get_verification_link_for_address(email_address):
    print("Retrieving verification link for %s." % email_address)
    vcode, link = _get_verification_data_for_address(email_address)
    print("Verification link retrieved: %s." % link)
    return link


def request_password_reset(email_address):
    logout()
    go_to('/+forgot_password')
    write_textfield('id_email', email_address)
    # Even though the recaptcha field is ignored for our tests, we do
    # want to verify that it is on the page.
    write_textfield('recaptcha_response_field', 'ignored')
    click_button(get_element(name='continue'))


def login_to_test_account():
    from sst import actions

    login(settings.TEST_ACCOUNT_EMAIL, settings.TEST_ACCOUNT_PASSWORD)
    # Wait for the reload
    wait_for(exists_element, id='ubuntu-header')

    if exists_element(tag='span', css_class='error', text="Password didn't match."):
        if get_base_url() == 'https://login.ubuntu.com/':
            actions._raise("This test requires a test account to be present")

        register_account(settings.TEST_ACCOUNT_EMAIL, "Test Account",
            settings.TEST_ACCOUNT_PASSWORD)


def login_from_redirect(email=settings.QA_ACCOUNT_EMAIL,
        password=settings.QA_ACCOUNT_PASSWORD):
    wait_for(assert_title, 'Log in')
    write_textfield('id_email', email)
    write_textfield('id_password', password)
    click_button(get_element(name='continue'))
    wait_for(assert_title_contains, 'Authenticate to')
    click_button(get_element(name='yes'))

