from itertools import count

from identityprovider.models import (
    Account,
    APIUser,
    EmailAddress,
    LPOpenIdIdentifier,
    Person,
    TeamParticipation,
    )
from identityprovider.const import (
    PERSON_VISIBILITY_PUBLIC,
    PERSON_VISIBILITY_PRIVATE_MEMBERSHIP,
)

from identityprovider.models.const import EmailStatus
from identityprovider.tests.utils import SSOBaseTestCase

__all__ = ['TestCaseWithFactory']


class SSOObjectFactory(object):
    """A factory for creating model objects for tests."""

    def __init__(self):
        self.counter = count()

    def get_unique_string(self, prefix=None):
        if prefix is None:
            prefix = "generic-string"
        return prefix + str(self.get_unique_integer())

    def get_unique_integer(self):
        """Return a unique integer for the test run."""
        return self.counter.next()

    def make_apiuser(self, username=None, password='test'):
        if username is None:
            username = self.get_unique_string(prefix='apiuser-')
        user, created = APIUser.objects.get_or_create(username=username)
        user.set_password(password)
        user.save()
        return user

    def make_account(self, displayname=None, username=None, email=None, password='test'):
        if displayname is None:
            displayname = self.get_unique_string(prefix='Account ')
        if username is None:
            username = self.get_unique_string(prefix='account-')
        account = Account.objects.create_account(displayname, username,
            password=password)
        self.make_email_for_account(account, email=email,
            status=EmailStatus.PREFERRED)
        return account

    def make_person(self, name=None, displayname=None, account=None):
        if name is None:
            name = self.get_unique_string(prefix='person-')
        if displayname is None:
            displayname = self.get_unique_string(prefix='Person ')
        lp_account = None
        if account is not None:
            lp_account = account.id

        person = Person.objects.create(name=name, displayname=displayname,
            lp_account=lp_account)
        if lp_account is not None:
            LPOpenIdIdentifier.objects.create(
                identifier=account.openid_identifier,
                lp_account=lp_account)
        return person

    def make_team(self, name=None, private=False, owner=None):
        if private:
            visibility = PERSON_VISIBILITY_PRIVATE_MEMBERSHIP
        else:
            visibility = PERSON_VISIBILITY_PUBLIC
        if owner is None:
            owner = self.make_person()
        if name is None:
            name = self.get_unique_string(prefix='team-')
        displayname = self.get_unique_string(prefix='Team ')
        team = self.make_person(name, displayname=displayname)
        team.visibility = visibility
        team.teamowner = owner
        team.save()
        return team

    def add_account_to_team(self, account, team):
        if account.person is None:
            self.make_person(account=account)
        elif account.person.in_team(team.name):
            return
        TeamParticipation.objects.create(person=account.person, team=team)

    def make_email_address(self):
        return self.get_unique_string(prefix='email-') + '@example.com'

    def make_email_for_account(self, account, email=None, status=None):
        if email is None:
            email = self.make_email_address()

        if status is None:
            status = EmailStatus.VALIDATED
        email, _ = EmailAddress.objects.get_or_create(email=email,
            defaults={'account': account, 'status': status})
        return email

class TestCaseWithFactory(SSOBaseTestCase):
    factory = SSOObjectFactory()
