import re

from BeautifulSoup import BeautifulSoup, SoupStrainer, PageElement
from django.conf import settings
from django.test import TransactionTestCase
from wsgi_intercept.zope_testbrowser import WSGI_Browser

from identityprovider.models.openidmodels import OpenIDRPConfig


class DuplicateIdError(Exception):
    """Raised by find_tag_by_id if more than one element has the given id."""


def find_tag_by_id(content, id):
    """Find and return the tag with the given ID"""
    if isinstance(content, PageElement):
        elements_with_id = content.findAll(True, {'id': id})
    else:
        elements_with_id = [tag for tag in BeautifulSoup(
                            content, parseOnlyThese=SoupStrainer(id=id))]
    if len(elements_with_id) == 0:
        return None
    elif len(elements_with_id) == 1:
        return elements_with_id[0]
    else:
        raise DuplicateIdError(
            'Found %d elements with id %r' % (len(elements_with_id), id))


class FunctionalTestCase(TransactionTestCase):
    functional = True
    fixtures = ['test']
    urls = 'identityprovider.tests.functional.urls'

    def new_browser(self):
        return WSGI_Browser()

    def setUp(self):
        self.anon_browser = self.new_browser()
        self.browser = self.new_browser()

        self._setup_django_settings()

    def _setup_django_settings(self):
        settings.DEBUG = True
        settings.SSO_ROOT_URL = 'http://openid.launchpad.dev/'
        settings.SSO_PROVIDER_URL = 'http://openid.launchpad.dev/+openid'
        settings.OPENID_PREAUTHORIZATION_ACL = (
            ('http://launchpad.dev/', 'http://launchpad.dev/'),
            ('http://openid.launchpad.dev/', 'http://openid.launchpad.dev/'),
        )
        settings.SSO_RESTRICT_RP = False

    def reset_browser(self):
        self.browser = self.new_browser()

    def login(self, email, password, browser=None):
        if browser is None:
            browser = self.browser

        browser.open('http://launchpad.dev/+login')

        browser.getControl(name='email').value = email
        browser.getControl(name='password').value = password
        browser.getControl(name='continue').click()

    def logout(self, browser=None):
        if browser is None:
            browser = self.browser

        if settings.BRAND == 'Ubuntu':
            logout_control = browser.getLink(id='logout-link')
        elif settings.BRAND == 'Launchpad':
            logout_control = browser.getControl(name='logout')
        else:
            self.fail('Invalid brand setting: %s' % settings.BRAND)
        logout_control.click()

    def admin_login(self, username, password, browser=None):
        if browser is None:
            browser = self.browser

        browser.open('http://launchpad.dev/admin/')

        browser.getControl(name='username').value = username
        browser.getControl(name='password').value = password
        browser.getControl(label='Log in').click()

    def admin_logout(self, browser=None):
        if browser is None:
            browser = self.browser

        browser.getLink('Log out').click()

    def assertUrl(self, url, browser=None):
        if not browser:
            browser = self.browser
        self.assertTrue(re.match(url, browser.url),
                          "Expected url %s doesn't match current one: %s" %
                          (url, browser.url))

    def assertContentType(self, expected_content_type, browser=None):
        if not browser:
            browser = self.browser
        content_type = browser.headers['content-type']
        if ';' in content_type:
            content_type = content_type.split(';', 1)[0]

        self.assertEquals(content_type, expected_content_type,
                          "Content type %s doesn't match expected %s" %
                          (content_type, expected_content_type))

    def assertRegexpMatches(self, string, regexp):
        if not re.search(regexp, string):
            msg = "Regular expression '%s' does not match '%s'"
            self.fail(msg % (regexp, string))

    def assertContains(self, content, browser=None, id=None):
        if not browser:
            browser = self.browser
        if id is None:
            if browser.contents.find(content) < 0:
                print browser.contents
                msg = "Retrieved page doesn't contain '%s'"
                self.fail(msg % content)
        else:
            soup = find_tag_by_id(browser.contents, id)
            if len(soup.findAll(text=content)) == 0:
                msg = "Retrieved page doesn't contain '%s' in tag with id=%s"
                self.fail(msg % (content, id))

    def assertNotContains(self, content, browser=None):
        if not browser:
            browser = self.browser
        if browser.contents.find(content) > 0:
            print browser.contents
            msg = "Retrieved page doesn't contain '%s'"
            self.fail(msg % content)

    def create_openid_rp_config(self, **extra):
        kwargs = {
            'trust_root': 'http://launchpad.dev/+openid-consumer',
            'displayname': 'test',
            'auto_authorize': False,
        }
        kwargs.update(extra)
        OpenIDRPConfig.objects.filter(trust_root=kwargs['trust_root']).delete()
        return OpenIDRPConfig.objects.create(**kwargs)
