from openid.consumer.consumer import Consumer
from openid.store.memstore import MemoryStore
from openid.consumer.discover import OPENID_2_0_TYPE as PROTOCOL_URI

from identityprovider.tests.functional.helpers import FunctionalTestCase
from identityprovider.tests.functional.openidhelpers import (
    complete_from_browser, make_endpoint)

class LogoutDuringLoginTestCase(FunctionalTestCase):

    def test(self):
        # = Handling Logouts During the Login Process =

        # If the user is logged in when the OpenID authentication request
        # begins, they are presented with a form that asks them if they want to
        # agree to log in to the remote site, without entering their password a
        # second time.

        # Now if the user manages to log out before posting this form, we can
        # not authenticate them to the remote site.  If this occurs, the user
        # should instead be prompted to log in.

        # There are a number of ways that this situation could occur:

        #  1. The user logs out in another browser window or tab.
        #  2. The user clicks the "I'm someone else" button, uses the browser's
        #     back button, then resubmits the form.

        # To test this, we will first authenticate to the Launchpad OpenID
        # provider:
        openid_store = MemoryStore()
        consumer = Consumer(session={}, store=openid_store)

        request = consumer.beginWithoutDiscovery(make_endpoint(
            PROTOCOL_URI, 'http://openid.launchpad.dev/+id/mark_oid'))
        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))

        self.browser.getControl(name='email').value = 'mark@example.com'
        self.browser.getControl(name='password').value = 'test'
        self.browser.getControl(name='continue').click()

        self.browser.getControl(name='yes').click()

        self.assertUrl('http://launchpad.dev/\+openid-consumer\?.*?')

        info = complete_from_browser(consumer, self.browser)

        self.assertEquals(info.status, 'success')
        self.assertEquals(info.endpoint.claimed_id,
                          'http://openid.launchpad.dev/+id/mark_oid')

        # Now lets begin another OpenID authentication request, which will
        # present a different form asking if we want to authenticate:

        request = consumer.beginWithoutDiscovery(make_endpoint(
            PROTOCOL_URI, 'http://openid.launchpad.dev/+id/mark_oid'))
        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))

        self.assertEquals(self.browser.title.strip(),
                          'Authenticate to http://launchpad.dev/')

        # We will now log out using the "Log Out" button, which will
        # present us with the login form.  We will then go back to the first
        # page:
        self.browser.getLink(id='logout-link').click()

        self.assertEquals(self.browser.title.strip(),
                          "You have been logged out")

        self.browser.getLink(id='login-link').click()

        self.assertEquals(self.browser.title.strip(), 'Log in')

        # From this point, we can log in as usual:

        self.browser.getControl(name='email').value = 'mark@example.com'
        self.browser.getControl(name='password').value = 'test'
        self.browser.getControl(name='continue').click()

        self.browser.getControl(name='yes').click()

        self.assertUrl('http://launchpad.dev/\+openid-consumer\?.*?')

        info = complete_from_browser(consumer, self.browser)

        self.assertEquals(info.status, 'success')
        self.assertEquals(info.endpoint.claimed_id,
                          'http://openid.launchpad.dev/+id/mark_oid')
