from openid.consumer.consumer import Consumer
from openid.fetchers import setDefaultFetcher
from openid.store.memstore import MemoryStore
from openid.consumer.discover import OPENID_2_0_TYPE as PROTOCOL_URI

from identityprovider.models.openidmodels import OpenIDRPConfig
from identityprovider.models.const import AccountCreationRationale
from identityprovider.const import LAUNCHPAD_TEAMS_NS

from identityprovider.tests.functional.helpers import FunctionalTestCase
from identityprovider.tests.functional.openidhelpers import complete_from_browser, make_endpoint


class OpenIDTeamsPrivateMembershipTestCase(FunctionalTestCase):

    def test(self):
        # = Launchpad OpenID Teams Extension Restrictions on Private Teams =

        # The Launchpad OpenID Teams Extension provides a way for relying
        # parties to check a user's team membership.  Launchpad also supports
        # the concept of private membership teams.  We do not want Launchpad's
        # OpenID code to disclose details of private membership teams to
        # unauthorized RPs.

        # The sample data contains a private team called "myteam", which has a
        # member called "member".

        # Let's set up an OpenID authentication request for this user,
        # checking to see whether they are a member of "myteam".  First we'll
        # set up the consumer:
        OpenIDRPConfig.objects.all().delete()

        openid_store = MemoryStore()
        consumer = Consumer(session={}, store=openid_store)

        # Now perform the authentication request, using the teams extension to
        # query membership of "myteam":

        endpoint = make_endpoint(
            PROTOCOL_URI, 'http://openid.launchpad.dev/+id/cCGE3LA')
        request = consumer.beginWithoutDiscovery(endpoint)

        request.message.namespaces.addAlias(LAUNCHPAD_TEAMS_NS, 'lp')

        request.addExtensionArg(
            LAUNCHPAD_TEAMS_NS, 'query_membership', 'myteam')

        self.browser.addHeader('Authorization', 'Basic member@canonical.com:test')
        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))
        self.browser.getControl(name='email').value = 'member@canonical.com'
        self.browser.getControl(name='password').value = 'test'
        self.browser.getControl(name='continue').click()
        # make sure the user doesn't have the option to authorize the
        # team, as it's private
        self.assertRaises(LookupError, self.browser.getControl, name='myteam')
        self.browser.getControl(name='yes').click()

        self.assertUrl('http://launchpad.dev/\+openid-consumer\?.*?')

        # The authentication request is successful:

        info = complete_from_browser(consumer, self.browser)

        self.assertEquals(info.status, 'success')

        # The authentication response includes a list of the teams the user is a
        # member of.  As "myteam" is private, Launchpad has not disclosed the
        # user's membership to the RP:

        self.assertEquals(info.getSigned(LAUNCHPAD_TEAMS_NS, 'is_member'), '')


        # == Trusted RPs ==

        # If the RP is known to Launchpad by having an OpenIDRPConfig and flagged
        # as being able to query private team membership (the
        # IOpenIDRPConfig.can_query_any_team flag), then it is allowed to do so.
        # To test this, lets create an RP config:

        self.create_openid_rp_config(
            trust_root='http://launchpad.dev/',
            displayname='Test RP', description="A test RP",
            creation_rationale=AccountCreationRationale.USER_CREATED)

        # Now perform a second authentication request:

        endpoint = make_endpoint(
            PROTOCOL_URI, 'http://openid.launchpad.dev/+id/cCGE3LA')
        request = consumer.beginWithoutDiscovery(endpoint)
        request.message.namespaces.addAlias(LAUNCHPAD_TEAMS_NS, 'lp')

        request.addExtensionArg(
            LAUNCHPAD_TEAMS_NS, 'query_membership', 'myteam')

        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))
        # make sure the user doesn't have the option to authorize the
        # team, as it's private
        self.assertRaises(LookupError, self.browser.getControl, name='myteam')
        self.browser.getControl(name='yes', index=0).click()

        self.assertUrl('http://launchpad.dev/\+openid-consumer\?.*?')

        # The authentication request is successful:
        info = complete_from_browser(consumer, self.browser)

        self.assertEquals(info.status, 'success')

        # However, even though the RP is known, it hasn't been given permission
        # to query private teams:

        self.assertEquals(info.getSigned(LAUNCHPAD_TEAMS_NS, 'is_member'), '')

        # So lets set the can_query_any_team flag on our RP config:

        rpconfig = OpenIDRPConfig.objects.get(trust_root='http://launchpad.dev/')
        rpconfig.can_query_any_team = True
        rpconfig.save()

        # Now if we authenticate again, Launchpad will informed us that
        # "member" is a member of "myteam":

        endpoint = make_endpoint(
            PROTOCOL_URI, 'http://openid.launchpad.dev/+id/cCGE3LA')
        request = consumer.beginWithoutDiscovery(endpoint)
        request.message.namespaces.addAlias(LAUNCHPAD_TEAMS_NS, 'lp')

        request.addExtensionArg(
            LAUNCHPAD_TEAMS_NS, 'query_membership', 'myteam')

        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))
        # authorize sending membership info
        self.browser.getControl(name='myteam').value = True
        self.browser.getControl(name='yes', index=0).click()
        info = complete_from_browser(consumer, self.browser)

        self.assertEquals(info.status, 'success')
        self.assertEquals(info.getSigned(LAUNCHPAD_TEAMS_NS, 'is_member'),
                          'myteam')

        rpconfig.delete()
