from openid.consumer.consumer import Consumer
from openid.store.memstore import MemoryStore
from openid.consumer.discover import OPENID_2_0_TYPE as PROTOCOL_URI

from identityprovider.tests.functional.helpers import FunctionalTestCase
from identityprovider.tests.functional.openidhelpers import make_identifier_select_endpoint


class RefererCookieTestCase(FunctionalTestCase):

    def get_cookie(self, cookie_name):
        # That whole code is to accommodate zope.testbrowser==3.5.1 which
        # doesn't provide browser.cookies attribute
        handler = [h for h in self.browser.mech_browser.handlers
                   if hasattr(h, 'cookiejar')][0]
        cookies = handler.cookiejar._cookies
        cookie = cookies[cookies.keys()[0]]
        cookie = cookie[cookie.keys()[0]]
        return cookie[cookie_name].value

    def test(self):
        # = Recording the Referer in a Cookie =

        # In order to maintain reasonable quality of service for important RPs,
        # we need some way that the load balancer can categorise connections by
        # the RP they originate from.

        # The first request in the OpenID authentication process can easily be
        # categorised by looking at the "Referer" request header, but that can't
        # be done for subsequent requests (whose "Referer" header will point at
        # the OpenID provider).

        # To solve this problem, we set a cookie at the beginning of the
        # authentication request containing the referer value.  The load
        # balancer can then use this to classify the subsequent requests.

        # First we'll set up our OpenID consumer:
        openid_store = MemoryStore()
        consumer = Consumer(session={}, store=openid_store)

        # Now when we start the OpenID authentication request, the "Referer"
        # header gets saved into a cookie.  The cookie has no expiry date set,
        # so will not last beyond the current web browser session.

        request = consumer.beginWithoutDiscovery(
            make_identifier_select_endpoint(PROTOCOL_URI))
        self.browser.addHeader('Referer', 'http://example.com/referer')
        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))

        self.assertEquals(self.get_cookie('openid_referer'),
                          '"http://example.com/referer"')

