from openid.extensions.sreg import SRegRequest, SRegResponse
from openid.consumer.consumer import Consumer
from openid.store.memstore import MemoryStore
from openid.consumer.discover import OPENID_2_0_TYPE as PROTOCOL_URI

from identityprovider.models.openidmodels import (
    OpenIDAuthorization,
    OpenIDRPConfig,
    OpenIDRPSummary,
    )

from identityprovider.tests.functional.openidhelpers import complete_from_browser, make_endpoint
from identityprovider.tests.functional.helpers import FunctionalTestCase


class RestrictedSregTestCase(FunctionalTestCase):

    def test(self):
        # = Restricted OpenID Simple Registration Extension support =

        # The Launchpad OpenID server has restricted support for the OpenID
        # Simple Registration Extension.  It will only provide a full set of
        # registration details to certain known trust roots.  The user's
        # launchpad username is share with all roots.

        # This is done in order to share the user details among the various
        # Canonical/Ubuntu sites participating in single sign-on.  The user's
        # nickname is published to every site, which is useful things like
        # weblog comments.


        # == Behaviour for unknown trust roots ==

        # If a relying party attempts to request user details via the
        # openid.sreg extension and Launchpad does not have a particular policy
        # configured, then only the user's approved fields are returned in the
        # response.

        # We will perform an OpenID authentication request asking for a few
        # user details:
        openid_store = MemoryStore()
        consumer = Consumer(session={}, store=openid_store)

        endpoint = make_endpoint(
            PROTOCOL_URI, 'http://openid.launchpad.dev/+id/mark_oid')

        request = consumer.beginWithoutDiscovery(endpoint)
        request.addExtension(SRegRequest(
            required=['email', 'country'],
            optional=['fullname', 'nickname']))

        redirect_url = request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer')

        self.assertRegexpMatches(redirect_url,
                                 'http://openid.launchpad.dev/\+openid\?.*?')

        self.browser.open(redirect_url)
        # log in
        self.browser.getControl(name='email').value = 'mark@example.com'
        self.browser.getControl(name='password').value = 'test'
        self.browser.getControl(name='continue').click()

        # authorize data
        # required fields are checked by default. don't authorize
        # anything
        email = self.browser.getControl(name='email')
        self.assertEqual(email.disabled, False)
        self.assertTrue(email.mech_control.items[0].selected)
        email.value = False
        self.browser.getControl(name='yes', index=0).click()

        self.assertUrl('http://launchpad.dev/\+openid-consumer\?.*?')

        # We have authenticated successfully:

        info = complete_from_browser(consumer, self.browser)

        self.assertEquals(info.status, 'success')
        self.assertEquals(info.endpoint.claimed_id,
                          'http://openid.launchpad.dev/+id/mark_oid')

        # But no fields are returned:

        sreg_response = SRegResponse.fromSuccessResponse(info)

        self.assertEquals(sreg_response, None)

        # If we attempt to authenticate again, we will be prompted to
        # confirm which fields we want to provide to the RP again,
        # but the defaults will be what we provided last time:

        request = consumer.beginWithoutDiscovery(endpoint)
        request.addExtension(SRegRequest(
            required=['email', 'country'],
            optional=['fullname', 'nickname']))
        redirect_url = request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer')
        self.assertRegexpMatches(redirect_url,
                                 'http://openid.launchpad.dev/\+openid\?.*?')
        self.browser.open(redirect_url)
        # No log in needed this time, we're directed straight to the confirm
        # screen.  Check that email is *not* selected by default this time:
        email = self.browser.getControl(name='email')
        self.assertFalse(email.mech_control.items[0].selected)

        # == Behaviour for known trust roots ==

        # If we create a Relying Party configuration for the trust root, things
        # play out a bit differently:

        OpenIDRPSummary.objects.all().delete()
        OpenIDAuthorization.objects.all().delete()

        self.create_openid_rp_config(
            trust_root='http://launchpad.dev/',
            displayname='Test RP', description='A test RP',
            allowed_sreg=','.join(['fullname', 'nickname', 'email', 'timezone']))

        # Now begin another identical OpenID authentication request:

        request = consumer.beginWithoutDiscovery(endpoint)
        request.addExtension(SRegRequest(
            required=['email', 'country'],
            optional=['fullname', 'nickname']))

        redirect_url = request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer')

        self.assertRegexpMatches(redirect_url,
                                 'http://openid.launchpad.dev/\+openid\?.*?')

        self.browser.open(redirect_url)
        # authorize data
        # required fields cannot be unchecked.
        self.assertEqual(self.browser.getControl(name='email').disabled, True)
        # authorize nickname
        self.browser.getControl(name='nickname').value = True
        # unauthorize fullname (checked by default)
        raw_control = self.browser.getControl(name='fullname').mech_control
        self.assertEqual(raw_control.get().attrs['checked'], 'checked')
        self.browser.getControl(name='fullname').value = False
        self.browser.getControl(name='yes', index=0).click()

        self.assertUrl('http://launchpad.dev/\+openid-consumer\?.*?')

        # Again, the authentication request is successful:

        info = complete_from_browser(consumer, self.browser)

        self.assertEquals(info.status, 'success')
        self.assertEquals(info.endpoint.claimed_id,
                          'http://openid.launchpad.dev/+id/mark_oid')

        # But now we have some user details.

        sreg_response = SRegResponse.fromSuccessResponse(info)

        self.assertEquals(list(sorted(sreg_response.items())),
                          [('email', 'mark@example.com'),
                           ('nickname', 'mark')])

        # == Behaviour for known trust roots with auto_authorize ==

        # enable auto-authorize for the rpconfig
        rpconfig = OpenIDRPConfig.objects.get(
            trust_root='http://launchpad.dev/')
        rpconfig.auto_authorize = True
        rpconfig.allowed_sreg=','.join(['fullname', 'email', 'timezone'])
        rpconfig.save()

        # Now begin another identical OpenID authentication request:

        request = consumer.beginWithoutDiscovery(endpoint)
        request.addExtension(SRegRequest(
            required=['email', 'country'],
            optional=['fullname', 'nickname']))

        redirect_url = request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer')

        self.assertRegexpMatches(redirect_url,
                                 'http://openid.launchpad.dev/\+openid\?.*?')

        self.browser.open(redirect_url)

        self.assertUrl('http://launchpad.dev/\+openid-consumer\?.*?')

        # Again, the authentication request is successful:

        info = complete_from_browser(consumer, self.browser)

        self.assertEquals(info.status, 'success')
        self.assertEquals(info.endpoint.claimed_id,
                          'http://openid.launchpad.dev/+id/mark_oid')

        # Again, we have some user details, but this time, the optional
        # sreg fields are also included automatically.

        sreg_response = SRegResponse.fromSuccessResponse(info)

        self.assertEquals(list(sorted(sreg_response.items())),
                          [('email', 'mark@example.com'),
                           ('fullname', 'Mark Shuttleworth')])
