from openid.consumer.consumer import Consumer
from openid.store.memstore import MemoryStore
from openid.consumer.discover import OPENID_2_0_TYPE as PROTOCOL_URI

from identityprovider.models import OpenIDRPSummary

from identityprovider.tests.functional.openidhelpers import (
    complete_from_browser, make_identifier_select_endpoint)
from identityprovider.tests.functional.helpers import FunctionalTestCase


class SSOWorkflowAuthorizeTestCase(FunctionalTestCase):

    def test(self):
        # = Launchpad Single-Signon Workflow: Authorize =

        # If a user has logged into Launchpad and then wants to log in to a
        # Launchpad-SSO web site, they are not prompted for a password.

        # First we will set up the helper view that lets us test the final
        # portion of the authentication process:

        # The authentication process is started by the relying party issuing a
        # checkid_setup request, sending the user to Launchpad.  We will
        # authenticate before starting the login procedure:
        OpenIDRPSummary.objects.filter(
           openid_identifier='http://openid.launchpad.dev/+id/mark_oid').delete()


        self.browser.open('http://openid.launchpad.dev/+login')
        self.browser.getControl(name='email').value = 'mark@example.com'
        self.browser.getControl(name='password').value = 'test'
        self.browser.getControl(name='continue').click()

        openid_store = MemoryStore()
        consumer = Consumer(session={}, store=openid_store)
        request = consumer.beginWithoutDiscovery(
            make_identifier_select_endpoint(PROTOCOL_URI))

        # Use localhost instead of launchpad.dev because Launchpad is not
        # recorded as an OpenIDRPSummary.
        self.browser.open(request.redirectURL(
            'http://localhost/', 'http://localhost/+openid-consumer'))

        # Each OpenID authentication is recorded as an OpenIDRPSummary. This user
        # has not ever used Launchpad to authenticate with a relying party.

        summaries = OpenIDRPSummary.objects.filter(
             openid_identifier='http://openid.launchpad.dev/+id/mark_oid')

        self.assertEquals(summaries.count(), 0)

        # At this point, the user is presented with a form asking if they want
        # to authenticate to the relying party

        self.assertEquals(self.browser.title.strip(),
                          'Authenticate to http://localhost/')

        # By clicking the "Sign In" button, the user is returned to the relying
        # party with their identity URL:

        self.browser.getControl(name='yes', index=0).click()

        self.assertUrl('http://localhost/\+openid-consumer\?.*?')

        info = complete_from_browser(
             consumer, self.browser, 'http://openid.launchpad.dev/+id/mark_oid')

        self.assertEquals(info.status, 'success')
        self.assertEquals(info.endpoint.claimed_id,
                          'http://openid.launchpad.dev/+id/mark_oid')

        # The authentication was recorded. There is a summary showing that the
        # user has used the replying part once.

        summaries = OpenIDRPSummary.objects.filter(
            openid_identifier='http://openid.launchpad.dev/+id/mark_oid')

        self.assertEquals(summaries.count(), 1)
        summary = summaries[0]

        self.assertEquals(summary.openid_identifier,
                          u'http://openid.launchpad.dev/+id/mark_oid')
        self.assertEquals(summary.trust_root, u'http://localhost/')
        self.assertEquals(summary.total_logins, 1)

        # == Declining to Authenticate ==

        # Alternatively, the user could have declined to authenticate.  This
        # time, we will cancel the authentication request:

        request = consumer.beginWithoutDiscovery(
            make_identifier_select_endpoint(PROTOCOL_URI))
        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))


        self.assertEquals(self.browser.title.strip(),
                          "Authenticate to http://launchpad.dev/")
        self.browser.getLink(url="+cancel").click()

        self.assertUrl('http://launchpad.dev/\+openid-consumer\?.*?')

        info = complete_from_browser(
             consumer, self.browser, 'http://openid.launchpad.dev/+id/mark_oid')

        self.assertEquals(info.status, 'cancel')

        OpenIDRPSummary.objects.filter(
           openid_identifier='http://openid.launchpad.dev/+id/mark_oid').delete()
