from BeautifulSoup import BeautifulSoup

from openid.consumer.consumer import Consumer
from openid.store.memstore import MemoryStore
from openid.consumer.discover import OPENID_2_0_TYPE as PROTOCOL_URI

from identityprovider.models.const import LoginTokenType
from identityprovider.models.authtoken import AuthToken
from identityprovider.models.account import Account

from identityprovider.tests.functional.openidhelpers import complete_from_browser, make_identifier_select_endpoint
from identityprovider.tests.functional.helpers import FunctionalTestCase


class SSOWorkflowCompleteTestCase(FunctionalTestCase):
    def test(self):
        account = Account.objects.get_by_email('new-user@example.com')
        if account:
            account.delete()

        # = Launchpad Single-Signon Workflow: The whole process =

        # If a user wants to use a Launchpad-SSO web site, but does not have a
        # Launchpad account, they can register directly from the login page.

        # By doing so they will get an SSO account, but that it will have no
        # associated person as they're not actually using Launchpad (yet).

        # First we will set up the helper view that lets us test the final
        # portion of the authentication process:

        # The authentication process is started by the relying party issuing a
        # checkid_setup request, sending the user to Launchpad:
        openid_store = MemoryStore()
        consumer = Consumer(session={}, store=openid_store)

        request = consumer.beginWithoutDiscovery(
            make_identifier_select_endpoint(PROTOCOL_URI))
        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))

        # At this point, we are at the login page.  Let's create a new account
        # for an email address that has not been registered, also entering the
        # name and password:
        self.browser.getLink(url='+new_account').click()
        self.browser.getControl(name='displayname').value = 'New User'
        self.browser.getControl(name='email').value = 'new-user@example.com'
        self.browser.getControl(name='password').value = 'testP4ss'
        self.browser.getControl(name='passwordconfirm').value = 'testP4ss'
        self.browser.getControl(name='continue').click()

        soup = BeautifulSoup(self.browser.contents)

        self.assertEquals(
            soup.find('h1', {'class': 'main'}).renderContents(),
            "Registration mail sent")

        # Following the link sent by email will take the user to the page where
        # they confirm the registration.
        token = AuthToken.objects.filter(
            email='new-user@example.com',
            token_type=LoginTokenType.NEWPERSONLESSACCOUNT,
            date_consumed=None).order_by('-date_created')[0].token
        token_url = 'http://openid.launchpad.dev/confirm-account/%s/new-user@example.com' % token
        self.browser.open(token_url)

        # It doesn't redirect:

        self.assertUrl('http://openid.launchpad.dev/confirm-account/.*?/'
                       'new-user@example.com')

        self.browser.getControl(name='continue').click()

        # Now the user is logged in with their new account, and has been
        # directed back to the original site:

        self.assertEquals(
            self.browser.getLink(url='http://launchpad.dev/').text,
            "http://launchpad.dev/")

        self.browser.getControl(name='yes').click()

        self.assertUrl('http://launchpad.dev/\+openid-consumer\?.*?')

        # And the response matches the new OpenID:
        account = Account.objects.get_by_email('new-user@example.com')
        expected_claimed_id = (
            'http://openid.launchpad.dev/+id/' + account.openid_identifier)

        info = complete_from_browser(consumer, self.browser, expected_claimed_id)

        self.assertEquals(info.status, 'success')
        self.assertEquals(info.endpoint.claimed_id, expected_claimed_id)

        # Since this account was created using OpenID, we will not create an entry in
        # the Person table for it.
        self.assertTrue(account.person is None)

        # That account can obviously be used in further OpenID interactions.
        self.reset_browser()

        request = consumer.beginWithoutDiscovery(
            make_identifier_select_endpoint(PROTOCOL_URI))
        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))

        self.browser.getControl(name='email').value = 'new-user@example.com'
        self.browser.getControl(name='password').value = 'testP4ss'
        self.browser.getControl(name='continue').click()

        # Here we see the OpenID interaction completed successfully.
        self.assertEquals(
            self.browser.getLink(url='http://launchpad.dev/').text,
            "http://launchpad.dev/")

        self.browser.getControl(name='yes').click()
        info = complete_from_browser(consumer, self.browser, expected_claimed_id)

        self.assertEquals(info.status, 'success')

        # If the user forgets their password, it's possible to reset it.

        self.reset_browser()
        request = consumer.beginWithoutDiscovery(
            make_identifier_select_endpoint(PROTOCOL_URI))
        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))

        self.browser.getLink(url='+forgot_password').click()
        self.browser.getControl(name='email').value = 'new-user@example.com'
        self.browser.getControl(name='continue').click()

        soup = BeautifulSoup(self.browser.contents)

        self.assertEquals(
            soup.find('h1', {'class': 'main'}).renderContents(),
            "Forgotten your password?")

        token = AuthToken.objects.filter(
            email='new-user@example.com',
            token_type=LoginTokenType.PASSWORDRECOVERY,
            date_consumed=None).order_by('-date_created')[0].token
        token_url = 'http://openid.launchpad.dev/token/%s' % token
        self.browser.open(token_url)

        self.assertUrl('http://.*?/\+resetpassword/new-user@example.com')

        self.browser.getControl(name='password').value = 'test2...'
        self.browser.getControl(name='passwordconfirm').value = 'test2...'
        self.browser.getControl(name='continue').click()

        # Here we see the OpenID interaction completed successfully.
        self.assertEquals(
            self.browser.getLink(url='http://launchpad.dev/').text,
            "http://launchpad.dev/")
        self.browser.getControl(name='yes').click()
        info = complete_from_browser(consumer, self.browser, expected_claimed_id)

        self.assertEquals(info.status, 'success')

        # And then log back in, using the new password.

        self.reset_browser()
        request = consumer.beginWithoutDiscovery(
            make_identifier_select_endpoint(PROTOCOL_URI))
        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))

        self.browser.getControl(name='email').value = 'new-user@example.com'
        self.browser.getControl(name='password').value = 'test2...'
        self.browser.getControl(name='continue').click()

        # Here we see the OpenID interaction completed successfully.
        self.assertEquals(
            self.browser.getLink(url='http://launchpad.dev/').text,
            "http://launchpad.dev/")

        self.browser.getControl(name='yes').click()
        info = complete_from_browser(consumer, self.browser, expected_claimed_id)

        self.assertEquals(info.status, 'success')

        # == Cleanup ==
        AuthToken.objects.filter(email='new-user@example.com').delete()
        account = Account.objects.get_by_email('new-user@example.com')
        account.emailaddress_set.all()[0].delete()
        account.delete()
