from BeautifulSoup import BeautifulSoup

from openid.consumer.consumer import Consumer
from openid.store.memstore import MemoryStore
from openid.consumer.discover import OPENID_2_0_TYPE as PROTOCOL_URI

from identityprovider.tests.functional.openidhelpers import complete_from_browser, make_identifier_select_endpoint
from identityprovider.tests.functional.helpers import FunctionalTestCase


class SSOWorkflowLoginTestCase(FunctionalTestCase):

    def test(self):
        # = Launchpad Single-Signon Workflow: Login =

        # A user with an existing account may log into a Launchpad-SSO web site
        # simply by entering their password on the login site.

        # First we will set up the helper view that lets us test the final
        # portion of the authentication process:

        # The authentication process is started by the relying party issuing a
        # checkid_setup request, sending the user to Launchpad:
        openid_store = MemoryStore()
        consumer = Consumer(session={}, store=openid_store)

        request = consumer.beginWithoutDiscovery(
            make_identifier_select_endpoint(PROTOCOL_URI))
        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))

        # As the user already has an account, they can enter their email address
        # and password.  The user must enter a valid email address, or they will
        # receive an error:

        self.browser.getControl(name='email').value = 'not an email address'
        self.browser.getControl(name='password').value = 'test'
        self.browser.getControl(name='continue').click()

        soup = BeautifulSoup(self.browser.contents)

        self.assertEquals(soup.find('span', 'error').renderContents(),
                          "Invalid email.")

        # Leaving the email address field blank results in the same error:

        self.browser.getControl(name='email').value = ''
        self.browser.getControl(name='password').value = 'test'
        self.browser.getControl(name='continue').click()

        soup = BeautifulSoup(self.browser.contents)

        self.assertEquals(soup.find('span', 'error').renderContents(),
                          "Required field.")

        # If the user provides a non-ASCII password, they receive an incorrect
        # password error.

        self.browser.getControl(name='email').value = 'mark@example.com'
        self.browser.getControl(name='password').value = '\xc2\xa0blah'
        self.browser.getControl(name='continue').click()

        soup = BeautifulSoup(self.browser.contents)

        self.assertEquals(soup.find('span', 'error').renderContents(),
                          "Password didn't match.")

        # If the password does not match the given email address, an error is
        # shown:

        self.browser.getControl(name='email').value = 'mark@example.com'
        self.browser.getControl(name='password').value = 'not the password'
        self.browser.getControl(name='continue').click()

        soup = BeautifulSoup(self.browser.contents)

        self.assertEquals(soup.find('span', 'error').renderContents(),
                          "Password didn't match.")

        # Finally, if the email address and password match, the user is logged
        # in and returned to the relying party, with the user's identity URL:

        self.browser.getControl(name='password').value = 'test'
        self.browser.getControl(name='continue').click()

        self.assertEquals(
            self.browser.getLink(url='http://launchpad.dev/').text,
            "http://launchpad.dev/")

        self.browser.getControl(name='yes').click()

        self.assertUrl('http://launchpad.dev/\+openid-consumer\?.*?')

        info = complete_from_browser(
            consumer, self.browser, 'http://openid.launchpad.dev/+id/mark_oid')

        self.assertEquals(info.status, 'success')
        self.assertEquals(info.endpoint.claimed_id,
                          'http://openid.launchpad.dev/+id/mark_oid')


