from BeautifulSoup import BeautifulSoup

from openid.consumer.consumer import Consumer
from openid.store.memstore import MemoryStore
from openid.consumer.discover import OPENID_2_0_TYPE as PROTOCOL_URI

from identityprovider.models.const import (AccountCreationRationale,
                                           LoginTokenType)
from identityprovider.models.authtoken import AuthToken
from identityprovider.models.account import Account

from identityprovider.tests.functional.openidhelpers import (
    complete_from_browser, make_identifier_select_endpoint)
from identityprovider.tests.functional.helpers import FunctionalTestCase


class SSOWorkflowRegisterTestCase(FunctionalTestCase):

    def test(self):
        account = Account.objects.get_by_email('new-user@example.com')
        if account:
            account.delete()
        # = Launchpad Single-Signon Workflow: Registration =

        # If a user wants to use a Launchpad-SSO web site, but does not have a
        # Launchpad account, they can register directly from the login page.

        # First we will set up the helper view that lets us test the final
        # portion of the authentication process:

        # The authentication process is started by the relying party issuing a
        # checkid_setup request, sending the user to Launchpad:
        openid_store = MemoryStore()
        consumer = Consumer(session={}, store=openid_store)

        request = consumer.beginWithoutDiscovery(
            make_identifier_select_endpoint(PROTOCOL_URI))

        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))

        # When a new account is created we'll use the creation rationale
        # specified for the trust_root given by the relying party.  We will set
        # up an RP configuration that uses the UBUNTU_SHOP creation rationale:

        rationale = AccountCreationRationale.OWNER_CREATED_UBUNTU_SHOP
        self.create_openid_rp_config(
            trust_root='http://launchpad.dev/',
            displayname='The Ubuntu Store from Canonical',
            description="For the Ubuntu Store, you need a Launchpad account "
                "so we can remember your order details and keep in "
                "touch with you about your orders.",
            creation_rationale=rationale)

        # At this point, we are at the login page.  Lets try to create a new
        # account for an email address that has already been registered. This
        # shouldn't result in an error, because we don't want to reveal
        # (non-)existing email addresses:
        self.browser.getLink('New account').click()
        self.browser.getControl(name='displayname').value = 'Tester'
        self.browser.getControl(name='email').value = 'test@canonical.com'
        self.browser.getControl(name='password').value = 'Testing123'
        self.browser.getControl(name='passwordconfirm').value = 'Testing123'
        self.browser.getControl(name='continue').click()

        soup = BeautifulSoup(self.browser.contents)
        self.assertEquals(
            soup.find('h1', {'class': 'main'}).renderContents(),
            "Registration mail sent")

        # If we instead pick a new email address, we can register an account:

        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))
        self.browser.getLink('New account').click()
        self.browser.getControl(name='displayname').value = 'New User'
        self.browser.getControl(name='email').value = 'new-user@example.com'
        self.browser.getControl(name='password').value = 'testP4ss'
        self.browser.getControl(name='passwordconfirm').value = 'testP4ss'
        self.browser.getControl(name='continue').click()

        soup = BeautifulSoup(self.browser.contents)
        self.assertEquals(
            soup.find('h1', {'class': 'main'}).renderContents(),
            "Registration mail sent")

        # The user would then check their email, and find a message.

        # Let's extract the URL from the email and follow the link:

        token = AuthToken.objects.filter(
            email='new-user@example.com',
            token_type=LoginTokenType.NEWPERSONLESSACCOUNT,
            date_consumed=None).order_by('-date_created')[0].token
        link = 'http://openid.launchpad.dev/confirm-account/' + \
            ('%s/new-user@example.com' % token)
        self.browser.open(link)

        self.assertUrl('http://openid.launchpad.dev/confirm-account/.*?/'
                       'new-user@example.com')

        # The user has already entered their full name and password, and only
        # needs to confirm.

        self.browser.getControl(name='continue').click()

        # Now the user is logged in with their new account, and has been
        # directed back to the original site:

        soup = BeautifulSoup(self.browser.contents)
        self.assertEquals(soup.find('h2').renderContents(),
                          "Sign in")

        self.browser.getControl(name='yes').click()

        self.assertUrl('http://launchpad.dev/\+openid-consumer\?.*?')

        # The creation rationale has been set correctly:
        account = Account.objects.get_by_email('new-user@example.com')
        expected_claimed_id = (
            'http://openid.launchpad.dev/+id/' + account.openid_identifier)

        self.assertEquals(account.creation_rationale,
                          AccountCreationRationale.OWNER_CREATED_UBUNTU_SHOP)

        # And the response matches the new OpenID:

        info = complete_from_browser(consumer, self.browser, expected_claimed_id)

        self.assertEquals(info.status, 'success')
        self.assertEquals(info.endpoint.claimed_id, expected_claimed_id)

        # Since this account was created using OpenID, we will not create an entry in
        # the Person table for it -- it will only be created when the user logs into
        # Launchpad.

        self.assertTrue(account.person is None)

        # == Cleanup ==
        AuthToken.objects.filter(email='new-user@example.com').delete()
        account = Account.objects.get_by_email('new-user@example.com')
        account.emailaddress_set.all()[0].delete()
        account.delete()

