from BeautifulSoup import BeautifulSoup

from openid.consumer.consumer import Consumer
from openid.store.memstore import MemoryStore
from openid.consumer.discover import OPENID_2_0_TYPE as PROTOCOL_URI

from identityprovider.models.const import LoginTokenType
from identityprovider.models.authtoken import AuthToken

from identityprovider.tests.functional.openidhelpers import complete_from_browser, make_identifier_select_endpoint
from identityprovider.tests.functional.helpers import FunctionalTestCase


class SSOWorkflowResetPasswordTestCase(FunctionalTestCase):

    def test(self):
        # = Launchpad Single-Signon Workflow: Password Reset =

        # If a user wants to use a Launchpad-SSO web site, but has forgotten
        # their password, they can ask for it to be emailed to them from the
        # login page.

        # First we will set up the helper view that lets us test the final
        # portion of the authentication process:

        # The authentication process is started by the relying party issuing a
        # checkid_setup request, sending the user to Launchpad:
        openid_store = MemoryStore()
        consumer = Consumer(session={}, store=openid_store)

        request = consumer.beginWithoutDiscovery(
            make_identifier_select_endpoint(PROTOCOL_URI))
        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))

        # At this point, we are at the login page.  Lets try to recover a
        # password for an unregistered email.  This shouldn't result in an error,
        # because we don't want to reveal (non-)existing email addresses:

        self.browser.getLink(url='+forgot_password').click()
        self.browser.getControl(name='email').value = 'no-account@example.com'
        self.browser.getControl(name='continue').click()

        soup = BeautifulSoup(self.browser.contents)
        self.assertEquals(
            soup.find('h1', {'class': 'main'}).renderContents(),
            "Forgotten your password?")

        # If the user tries to recover a password registered to a team, that
        # wouldn't fail either:

        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))
        self.browser.getLink(url='+forgot_password').click()
        self.browser.getControl(name='email').value = 'support@ubuntu.com'
        self.browser.getControl(name='continue').click()

        soup = BeautifulSoup(self.browser.contents)
        self.assertEquals(
            soup.find('h1', {'class': 'main'}).renderContents(),
            "Forgotten your password?")

        # Finally, lets try and recover the password for a test@canonical.com:

        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))
        self.browser.getLink(url='+forgot_password').click()
        self.browser.getControl(name='email').value = 'test@canonical.com'
        self.browser.getControl(name='continue').click()


        soup = BeautifulSoup(self.browser.contents)
        self.assertEquals(
            soup.find('h1', {'class': 'main'}).renderContents(),
            "Forgotten your password?")

        # The user would then check their email, and find a message:

        # Let's extract the URL from the email and follow the link:

        token = AuthToken.objects.filter(
            email='test@canonical.com',
            token_type=LoginTokenType.PASSWORDRECOVERY,
            date_consumed=None).order_by('-date_created')[0].token
        link = 'http://openid.launchpad.dev/token/%s' % token
        self.browser.open(link)

        self.assertUrl('http://openid.launchpad.dev/token/.*?/'
                       '\+resetpassword/test@canonical.com')

        # The user can now enter a new password:

        self.browser.getControl(name='password').value = 'new Password'
        self.browser.getControl(name='passwordconfirm').value = 'new Password'
        self.browser.getControl(name='continue').click()

        # Now the user is logged in, and has been directed back to the original
        # site:
        self.assertEquals(
            self.browser.getLink(url='http://launchpad.dev/').text,
            "http://launchpad.dev/")

        self.browser.getControl(name='yes').click()

        self.assertUrl('http://launchpad.dev/\+openid-consumer\?.*?')

        info = complete_from_browser(
            consumer, self.browser, 'http://openid.launchpad.dev/+id/name12_oid')

        self.assertEquals(info.status, 'success')
        self.assertEquals(info.endpoint.claimed_id,
                          "http://openid.launchpad.dev/+id/name12_oid")

        # == Cleanup ==
        AuthToken.objects.filter(email='test@canonical.com').delete()
