import re
from urllib import urlencode

from identityprovider.models.openidmodels import OpenIDRPConfig
from identityprovider.tests.functional.helpers import FunctionalTestCase


class BasicsTestCase(FunctionalTestCase):

    def test(self):
        # = OpenID =
        #
        # == Introduction ==
        #
        # Ubuntu SSO is an OpenID provider. If the URL is accessed by a web
        # browser, an informative message is displayed as per the OpenID spec.
        self.anon_browser.open('http://openid.launchpad.dev')

        self.assertContentType('text/html', self.anon_browser)

        # We are going to fake a consumer for these examples. In order to
        # ensure that the consumer is being fed the correct replies, we use a
        # view that renders the parameters in the response in an easily
        # testable format.
        self.anon_browser.open('http://launchpad.dev/+openid-consumer?foo=%2Bbar')

        self.assertEquals(self.anon_browser.contents,
                          "Consumer received GET\nfoo:+bar")

        # == associate Mode ==

        # Establish a shared secret between Consumer and Identity Provider.

        # After determining the URL of the OpenID server, the next thing a
        # consumer needs to do is associate with the server and get a shared
        # secret via a POST request.
        self.anon_browser.open(
             'http://openid.launchpad.dev/+openid',
             data=urlencode({
                 'openid.mode': 'associate',
                 'openid.assoc_type': 'HMAC-SHA1'
        }))

        self.assertContentType('text/plain', self.anon_browser)
        self.assertRegexpMatches(self.anon_browser.contents,
                                 'assoc_handle:\{HMAC-SHA1\}\{.*?\}\{.*?\}')
        self.assertRegexpMatches(self.anon_browser.contents,
                                 'assoc_type:HMAC-SHA1')
        self.assertRegexpMatches(self.anon_browser.contents,
                                 'expires_in:1209.*?')
        self.assertRegexpMatches(self.anon_browser.contents, 'mac_key:.*?')


        # Get the association handle, which we will need for later tests.
        [assoc_handle] = re.findall('assoc_handle:(.*)',
                                    self.anon_browser.contents)

        # == checkid_immediate Mode ==

        # Ask an Identity Provider if a End User owns the Claimed Identifier,
        # getting back an immediate "yes" or "can't say" answer.

        # Once the shared secret is negotiated, the consumer can send
        # checkid_immediate and checkid_setup GET requests.checkid_immediate
        # requests will currently return "can't say" as we are not yet logged
        # into Launchpad.

        args = urlencode({
            'openid.mode': 'checkid_immediate',
            'openid.identity':
            'http://openid.launchpad.dev/+id/mark_oid',
            'openid.assoc_handle': assoc_handle,
            'openid.return_to': 'http://launchpad.dev/+openid-consumer',
            })
        self.anon_browser.open('http://openid.launchpad.dev/+openid?%s' % args)

        self.assertContains("Consumer received GET", self.anon_browser)
        self.assertContains("openid.assoc_handle:", self.anon_browser)
        self.assertContains("openid.mode:id_res", self.anon_browser)
        self.assertContains(
            "openid.user_setup_url:http://openid.launchpad.dev/+openid?",
            self.anon_browser)

        # An error is returned to the consumer if an attempt to login as an
        # invalid user.
        args = urlencode({
            'openid.mode': 'checkid_immediate',
            'openid.identity': 'http://launchpad.dev/+id/limi_oid',
            'openid.assoc_handle': assoc_handle,
            'openid.return_to': 'http://launchpad.dev/+openid-consumer',
            })
        self.browser.open('http://openid.launchpad.dev/+openid?%s' % args)

        self.assertContains("openid.assoc_handle:")
        self.assertContains("openid.mode:id_res")
        self.assertContains(
            "openid.user_setup_url:http://openid.launchpad.dev/+openid?")

        # We test a success below in the 'Sticky Authorization' section.

        # == checkid_setup Mode ==

        # checkid_setup is interactive with the user. We can extract the URL
        # for the checkid_setup from the result of the previous test.

        [setup_url] = re.findall(
             '(?m)^openid.user_setup_url:(.*)$', self.anon_browser.contents)

        # Lets start a new browser session so we don't have any credentials.
        # When we go to the OpenID setup URL, we are presented with a login
        # form.  By entering an email address and password, we are directed back
        # to the consumer, completing the OpenID request:

        self.reset_browser()
        self.browser.open(setup_url)

        self.assertUrl('http://openid.launchpad.dev/.*?/\+decide')

        # Sign into SSO itself
        self.browser.getControl(name='email').value = 'mark@example.com'
        self.browser.getControl(name='password').value = 'test'
        self.browser.getControl(name='continue').click()
        self.browser.getControl(name='yes').click()

        self.assertUrl('http://launchpad.dev/\+openid-consumer\?.*?')
        self.assertContains("Consumer received GET")
        self.assertContains("openid.assoc_handle:")
        self.assertContains(
            "openid.identity:http://openid.launchpad.dev/+id/mark_oid")
        self.assertContains("openid.mode:id_res")
        self.assertContains(
            "openid.op_endpoint:http://openid.launchpad.dev/+openid")
        self.assertContains("openid.response_nonce:")
        self.assertContains(
            "openid.return_to:http://launchpad.dev/+openid-consumer",)
        self.assertContains("openid.sig:")
        self.assertContains("openid.signed:")

        # We will record the signature from this response to use in the next
        # test:
        [sig] = re.findall('sig:(.*)', self.browser.contents)

        # If we had been logged into Launchpad, we would instead have seen a
        # simple approve/deny form since Launchpad already knows who we are.
        # This can be seen using the existing browser session:

        self.browser.open(setup_url)
        self.browser.open(self.browser.url[:-6]+'cancel')

        self.assertUrl('http://launchpad.dev/\+openid-consumer\?.*?')
        self.assertContains("Consumer received GET")
        self.assertContains("openid.mode:cancel")

        # Also, checkid_immediate should now give a positive assertion for RPs
        # that are set to auto-authorize.
        self.create_openid_rp_config(auto_authorize=True)

        args = urlencode({
            'openid.mode': 'checkid_immediate',
            'openid.identity':
                'http://openid.launchpad.dev/+id/mark_oid',
            'openid.assoc_handle': assoc_handle,
            'openid.return_to': 'http://launchpad.dev/+openid-consumer',
            })

        self.browser.open('http://openid.launchpad.dev/+openid?%s' % args)

        self.assertContains("openid.assoc_handle:")
        self.assertContains("openid.mode:id_res")
        self.assertContains(
            "openid.return_to:http://launchpad.dev/+openid-consumer")

        OpenIDRPConfig.objects.filter(
            trust_root='http://launchpad.dev/+openid-consumer').delete()

        # == check_authentication Mode ==
        #
        # Ask an Identity Provider if a message is valid. For dumb, stateless
        # Consumers or when verifying an invalidate_handle response.
        #
        # If an association handle is stateful (genereted using the associate
        # Mode), check_authentication will fail.

        args = urlencode({
            'openid.mode': 'check_authentication',
            'openid.assoc_handle': assoc_handle,
            'openid.sig': sig,
            'openid.signed':  'return_to,mode,identity',
            'openid.identity':
                'http://openid.launchpad.dev/+id/mark_oid',
            'openid.return_to': 'http://launchpad.dev/+openid-consumer',
            })
        self.browser.open('http://openid.launchpad.dev/+openid?%s' % args)

        self.assertEquals(self.browser.contents, "is_valid:false\n")

        # If we are a dumb consumer though, we must invoke the
        # check_authentication mode, passing back the association handle,
        # signature and values of all fields that were signed.
        args = urlencode({
            'openid.mode': 'checkid_setup',
            'openid.identity':
                'http://openid.launchpad.dev/+id/mark_oid',
            'openid.return_to': 'http://launchpad.dev/+openid-consumer',
            })
        self.browser.open('http://openid.launchpad.dev/+openid?%s' % args)
        self.browser.getControl(name='yes', index=0).click()

        self.assertContains("Consumer received GET")
        self.assertContains("openid.assoc_handle:")
        self.assertContains(
            "openid.identity:http://openid.launchpad.dev/+id/mark_oid")
        self.assertContains("openid.mode:id_res")
        self.assertContains("openid.op_endpoint:http://openid.launchpad.dev/+openid")
        self.assertContains("openid.response_nonce:")
        self.assertContains("openid.return_to:http://launchpad.dev/+openid-consumer")
        self.assertContains("openid.sig:")
        self.assertContains("openid.signed:")

        fields = dict(line.split(':', 1)
                      for line in self.browser.contents.split('\n')[1:]
                      if line.startswith('openid.'))
        signed = ['openid.' + name
                  for name in fields['openid.signed'].split(',')]
        message = dict((key, value) for (key, value) in fields.items()
                       if key in signed)
        message.update({
            'openid.mode': 'check_authentication',
            'openid.assoc_handle': fields['openid.assoc_handle'],
            'openid.sig': fields['openid.sig'],
            'openid.signed': fields['openid.signed'],
            })

        args = urlencode(message)
        self.browser.open('http://openid.launchpad.dev/+openid', args)
        self.assertEquals(self.browser.contents, "is_valid:true\n")

        # == Identity Ownership ==
        #
        # You cannot log in as someone elses identity. If you try to, you will be
        # prompted with a login screen to connect as the correct user.
        #
        # Immediate mode:

        args = urlencode({
            'openid.mode': 'checkid_immediate',
            'openid.identity':
                'http://openid.launchpad.dev/+id/stub_oid',
            'openid.assoc_handle': assoc_handle,
            'openid.return_to': 'http://launchpad.dev/+openid-consumer',
            })
        self.browser.open('http://openid.launchpad.dev/+openid?%s' % args)

        self.assertContains("Consumer received GET")
        self.assertContains("openid.assoc_handle:")
        self.assertContains("openid.mode:id_res")
        self.assertContains("openid.sig:")
        self.assertContains("openid.signed:")
        self.assertContains(
            "openid.user_setup_url:http://openid.launchpad.dev/+openid?")


        # Interactive mode:

        [setup_url] = re.findall(
            '(?m)^openid.user_setup_url:(.*)$', self.browser.contents
            )

        self.browser.open(setup_url)

        self.assertUrl("http://launchpad.dev/\+openid-consumer\?.*?")
        self.assertEquals(self.browser.contents,
                          "Consumer received GET\nopenid.mode:cancel")

        # == Invalid identities ==
        #
        # If you attempt interactive authentication with an invalid OpenID
        # identifier, you get a nice error page.

        # XXX: disable this test until all functional test can be
        # rewritten correctly
        # This test fails because the csrftoken cookie is assigned to a
        # different domain (launchpad.dev instead of
        # openid.launchpad.dev); possibly due to how all the tests are
        # depending on each order (order-dependent).

        #args = urlencode({
        #    'openid.mode': 'checkid_immediate',
        #    'openid.identity':
        #        'http://some/other/site',
        #    'openid.assoc_handle': assoc_handle,
        #    'openid.return_to': 'http://launchpad.dev/+openid-consumer',
        #    })
        #self.browser.open('http://openid.launchpad.dev/+openid?%s' % args)
        #[setup_url] = re.findall(
        #    '(?m)^openid.user_setup_url:(.*)$', self.browser.contents
        #    )
        #self.browser.open(setup_url)
        #self.browser.getControl(name='continue').click()

        #self.assertUrl(
        #    'http://launchpad.dev/\+openid-consumer\?openid.mode=cancel')


        # == Broken Consumers ==

        # Really bad requests might trigger a protocol error.  These are such
        # edge cases that I can't even be bothered to figure out how to prevent
        # the u'' unicode prefix from showing up.  We might want to figure it
        # out one day if we feel inclined.

        args = urlencode({
            'openid.mode': 'whoops',
            })
        self.browser.handleErrors = True
        self.browser.open('http://openid.launchpad.dev/+openid?%s' % args)

        self.assertContains("error:")
        self.assertContains("mode u'whoops'")
        self.assertContains("mode:error")

        # If there is a valid return_to, then the consumer gets notified.
        args = urlencode({
            'openid.mode': 'whoops',
            'openid.return_to': 'http://launchpad.dev/+openid-consumer',
            })
        self.browser.open('http://openid.launchpad.dev/+openid?%s' % args)

        self.assertContains("openid.error:")

        self.assertContains("error:")
        self.assertContains("mode u'whoops'")
        self.assertContains("mode:error")
