from openid.consumer.consumer import Consumer
from openid.store.memstore import MemoryStore

from identityprovider.utils import encrypt_launchpad_password
from identityprovider.models.account import Account

from identityprovider.tests.functional.openidhelpers import complete_from_browser, get_endpoints
from identityprovider.tests.functional.helpers import FunctionalTestCase


class DelegatedIdentityTestCase(FunctionalTestCase):

    def assertSupports(self, data, expected_supports):
        supports = [e['supports'] for e in data['endpoints']]

        self.assertEqual(supports, expected_supports)

    def assertEndpointData(self, data, expected_local_id, expected_server_url):
        for endpoint in data['endpoints']:
            self.assertEqual(endpoint['local_id'], expected_local_id)
            self.assertEqual(endpoint['server_url'], expected_server_url)

    def test(self):
        # = Launchpad Delegated Identity =
        #
        # The Launchpad profile page acts as an OpenID identifier through delegation.
        #
        # The profile page delegates to the underlying Single Sign On identifier:

        data = get_endpoints('http://launchpad.dev/~name12')

        self.assertEquals(data['claimed_id'], "http://launchpad.dev/~name12")
        self.assertEndpointData(data,
                                "http://openid.launchpad.dev/+id/name12_oid",
                                "http://openid.launchpad.dev/+openid")
        self.assertSupports(data,
                         [["http://specs.openid.net/auth/2.0/signon"],
                          ["http://openid.net/signon/1.1"],
                          ["http://openid.net/signon/1.0"]])

        # In case the client doesn't support YADIS discovery, the profile page also
        # includes <link> discovery:

        data = get_endpoints('http://launchpad.dev/~name12', yadis=False)

        self.assertEquals(data['claimed_id'], "http://launchpad.dev/~name12")
        self.assertEndpointData(data,
                                "http://openid.launchpad.dev/+id/name12_oid",
                                "http://openid.launchpad.dev/+openid")
        self.assertSupports(data, [["http://specs.openid.net/auth/2.0/signon"],
                                   ["http://openid.net/signon/1.1"]])


        # == Logging in ==

        # Users can login through their profile page:
        account = Account.objects.get_by_email('test@canonical.com')
        account.accountpassword.password = encrypt_launchpad_password('test')
        account.accountpassword.save()

        openid_store = MemoryStore()
        consumer = Consumer(session={}, store=openid_store)

        request = consumer.begin('http://launchpad.dev/~name12')
        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))
        self.browser.getControl(name='email').value = 'test@canonical.com'
        self.browser.getControl(name='password').value = 'test'
        self.browser.getControl(name='continue').click()
        self.browser.getControl(name='yes').click()
        info = complete_from_browser(consumer, self.browser)

        self.assertEquals(info.status, 'success')
        self.assertEquals(info.endpoint.claimed_id,
                          "http://launchpad.dev/~name12")

