from openid.consumer.consumer import Consumer
from openid.store.memstore import MemoryStore
from openid.extensions import pape

from identityprovider.utils import encrypt_launchpad_password
from identityprovider.models.account import Account

from identityprovider.tests.functional.helpers import FunctionalTestCase
from identityprovider.tests.functional.openidhelpers import complete_from_browser


class MaxAuthAgeTestCase(FunctionalTestCase):

    def test(self):
        # = OpenID Security Extensions: Max Auth Age =

        # Launchpad supports the max_auth_age extension parameter defined in
        # Provider Authentication Policy Extension (
        # http://openid.net/specs/openid-provider-authentication-policy-extension-1_0-07.html).
        # It offers a way for a relaying party to ensure that the user enters
        # their password, even if they have already logged into Launchpad. The
        # purpose of this is for sites for which the long-lived session policy
        # of Launchpad is problematic.
        account = Account.objects.get_by_email('test@canonical.com')
        account.accountpassword.password = encrypt_launchpad_password('test')
        account.accountpassword.save()

        self.browser.open('http://openid.launchpad.dev')

        self.browser.getControl(name='email').value = 'test@canonical.com'
        self.browser.getControl(name='password').value = 'test'
        self.browser.getControl(name='continue').click()

        self.assertContains("Sample Person")

        # Normally, if a relaying party asks to authenticates this user, the user
        # will be able to click the 'Sign In' button to complete the
        # authentication process.
        openid_store = MemoryStore()
        consumer = Consumer(session={}, store=openid_store)
        request = consumer.begin('http://openid.launchpad.dev')

        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))
        self.browser.getControl(name='yes')

        # The max_auth_age PAPE parameter can be used to force the user to
        # enter their password, even if they are already authenticated with
        # Launchpad.

        consumer = Consumer(session={}, store=openid_store)
        request = consumer.begin('http://openid.launchpad.dev')

        # Using 0 as max_auth_age value will always force the user to enter
        # their password.
        pape_request = pape.Request(max_auth_age=0)
        request.addExtension(pape_request)
        self.browser.open(request.redirectURL(
            'http://launchpad.dev/', 'http://launchpad.dev/+openid-consumer'))

        self.assertUrl('http://openid.launchpad.dev/.*?/\+decide')

        self.assertNotContains("Sample Person")

        # Since we know who the user is, the email field is pre-filled for
        # their convenience:

        self.browser.getControl(name='email').value = 'test@canonical.com'
        self.browser.getControl(name='password').value = 'test'
        self.browser.getControl(name='continue').click()
        self.browser.getControl(name='yes').click()

        self.assertUrl('http://launchpad.dev/\+openid-consumer\?.*?')

        info = complete_from_browser(
            consumer, self.browser, 'http://openid.launchpad.dev/+id/name12_oid')

        self.assertEquals(info.status, 'success')
        self.assertEquals(info.endpoint.claimed_id,
                          'http://openid.launchpad.dev/+id/name12_oid')

        # The auth_age parameter contains the time of the last login.
        pape_response = pape.Response.fromSuccessResponse(info)

        self.assertRegexpMatches(pape_response.auth_time, '2.*')
