from urllib import urlencode
from urllib2 import HTTPError

from identityprovider.tests.functional.helpers import FunctionalTestCase
from identityprovider.tests.functional.openidhelpers import (
    make_identifier_select_endpoint, PublisherFetcher)

from openid.consumer.discover import OPENID_2_0_TYPE
from openid.consumer.consumer import Consumer
from openid.fetchers import setDefaultFetcher
from openid.store.memstore import MemoryStore

from identityprovider.models import OpenIDRPConfig, OpenIDAuthorization


class PreAuthorizationTestCase(FunctionalTestCase):

    def test(self):
        # = Pre-authorized consumers =

        # In some cases we want consumers to be pre-authorized so that the
        # authorization step of the protocol is skipped.  This is the case for
        # canonical's training site: we want the canonical shop to pre-authorize
        # the training site when a user buys some sort of training.

        # To do so the shop has to craft a link which takes the user to the
        # training site through Launchpad.  This is what the link looks like:

        # https://openid.launchpad.net/+pre-authorize-rp?trust_root=..&callback=...

        # Where the value of trust_root should be the root URL of the consumer to
        # be pre-authorized and callback is the URL to which the user will be
        # redirected after the consumer is pre-authorized.

        # As an example we'll pre-authorize our own OpenID consumer:

        # First, some setup to ensure old data doesn't mess with the results
        OpenIDRPConfig.objects.all().delete()
        OpenIDAuthorization.objects.all().delete()

        args = urlencode({
            'trust_root': 'http://openid.launchpad.dev/',
            'callback': 'http://openid.launchpad.dev/+edit'})

        self.browser.handleErrors = True
        self.browser.addHeader('REFERER', 'http://openid.launchpad.dev/')
        self.browser.open('http://openid.launchpad.dev/+pre-authorize-rp?%s' % args)

        # Since the user is not logged in, he'll have to login first.
        self.assertUrl('http://openid.launchpad.dev/\+login.*?')

        self.browser.getControl(name='email', index=0).value = 'mark@example.com'
        self.browser.getControl(name='password').value = 'test'
        self.browser.handleErrors = False
        self.browser.getControl(name='continue').click()

        # Now he's redirected to the callback page and the RP has been
        # pre-authorized.
        self.assertUrl('http://openid.launchpad.dev/\+edit')

        # First we will set up the helper view that lets us test the OpenID
        # protocol.
        setDefaultFetcher(PublisherFetcher())

        # The authentication process is started by the relying party issuing a
        # checkid_setup request, sending the user to Launchpad.
        openid_store = MemoryStore()
        consumer = Consumer(session={}, store=openid_store)
        request = consumer.beginWithoutDiscovery(
            make_identifier_select_endpoint(OPENID_2_0_TYPE))
        self.browser.open(request.redirectURL(
            'http://openid.launchpad.dev/',
            'http://openid.launchpad.dev/+openid-consumer'))

        # The user (mark) was already logged into Launchpad, so at this point,
        # instead of being presented with a form asking if they want to
        # authenticate to the relying party, they are sent directly to the RP.

        self.assertUrl('http://openid.launchpad.dev/\+openid-consumer\?.*?')

        # If the HTTP referrer and the trust_root are not in our config's
        # openid_preauthorization_acl, we will not pre-authorize.

        self.browser.open('http://openid.launchpad.dev/+logout')
        self.browser.open('http://openid.launchpad.dev/+login')
        self.browser.getControl(name='email', index=0).value = 'test@canonical.com'
        self.browser.getControl(name='password').value = 'test'
        self.browser.getControl(name='continue').click()

        args = urlencode({
            'trust_root': 'http://blueprints.launchpad.dev/',
            'callback': 'http://openid.launchpad.dev/+edit'})

        self.assertRaises(HTTPError,
                          self.browser.open,
                          'http://openid.launchpad.dev/+pre-authorize-rp?%s' % args)

        self.browser.open(request.redirectURL(
            'http://blueprints.launchpad.dev/',
            'http://blueprints.launchpad.dev/+openid-consumer'))

        self.assertEquals(self.browser.title.strip(),
                          'Authenticate to http://blueprints.launchpad.dev/')

        self.browser.open('http://openid.launchpad.dev/+logout')
        self.browser.open('http://openid.launchpad.dev/+login')
        self.browser.getControl(name='email', index=0).value = 'test@canonical.com'
        self.browser.getControl(name='password').value = 'test'
        self.browser.getControl(name='continue').click()

        args = urlencode({
            'trust_root': 'http://example.com/',
            'callback': 'http://launchpad.dev/people/+me'})
        self.assertRaises(HTTPError, self.browser.open,
                          'http://openid.launchpad.dev/+pre-authorize-rp?%s' % args)

        self.browser.open(request.redirectURL(
            'http://example.com/', 'http://example.com/+openid-consumer'))

        self.assertEquals(self.browser.title.strip(),
                          'Authenticate to http://example.com/')

        args = urlencode({
            'trust_root': 'http://example.com/'})

        self.assertRaises(HTTPError, self.browser.open,
                          'http://openid.launchpad.dev/+pre-authorize-rp?%s' % args)

        self.browser.open(request.redirectURL(
            'http://example.com/', 'http://example.com/+openid-consumer'))

        self.assertEquals(self.browser.title.strip(),
                          'Authenticate to http://example.com/')
