from identityprovider.tests.functional.helpers import FunctionalTestCase

from identityprovider.models.account import Account
from identityprovider.models.openidmodels import OpenIDRPSummary


class HomePageTestCase(FunctionalTestCase):

    def test(self):
        # = Openid.launchpad.dev Home Page =

        # When visiting the openid.launchpad.dev home page anonymously, the
        # user is presented with a login form:

        self.browser.open('http://openid.launchpad.dev')
        self.assertContains('Log in to Ubuntu Single Sign On', id='auth')

        # However, if the user is logged in, they are presented with some
        # information about their account.

        self.browser.getControl(name='email').value = 'test@canonical.com'
        self.browser.getControl(name='password').value = 'test'
        self.browser.getControl(name='continue').click()

        self.assertEquals(self.browser.title.strip(), "Sample Person's details")

        self.assertContains("Full name", id="content")
        self.assertContains("test@canonical.com", id="content")
        self.assertContains("testing@canonical.com", id="content")
        self.assertContains("Sites you last authenticated to", id="content")

        # == Previously Visited Sites ==

        # When the person is logged in and points his browser to
        # openid.launchpad.dev, he will also see a list containing up to 10
        # sites in which he authenticated recently, using the OpenID service.

        # Pretend the user authenticated to an OpenID relying party.
        OpenIDRPSummary.objects.record(
            Account.objects.get_by_email('test@canonical.com'),
            'http://example.com/')

        # Now reload the page
        self.browser.open('http://openid.launchpad.dev')

        self.assertContains("Sites you last authenticated to", id='visited-sites')
        self.assertContains("Site", id='visited-sites')
        self.assertContains("Last authenticated", id='visited-sites')
        self.assertContains("http://example.com/", id='visited-sites')
        self.assertContains("https://shop.canonical.com", id='visited-sites')
        self.assertContains("2008-02-04", id='visited-sites')

        # == Editing User Details ==

        # The user can edit their details from the main account page:
        self.browser.getControl(name="displayname").value = "New name"
        self.browser.getControl(name="update").click()

        self.assertEquals(self.browser.url, "http://openid.launchpad.dev/")
        self.assertEquals(self.browser.title.strip(), "New name's details")

        # Including their password:
        self.browser.getControl(name="password").value = "TestPass23"
        self.browser.getControl(name="passwordconfirm").value = "TestPass23"
        self.browser.getControl(name="update").click()

        self.assertEquals(self.browser.url, "http://openid.launchpad.dev/")

        # User can not use a blank display name:

        self.browser.getControl(name="displayname").value = " "
        self.browser.getControl(name="update").click()

        self.assertContains('Required field')

        self.browser.getControl(name="displayname").value = "Sample Person "
        self.browser.getControl(name="update").click()

        self.assertNotContains('Required field')

        # == Logging Out ==

        # From the Home page, the user can also log out:

        self.browser.open('http://openid.launchpad.dev')
        self.browser.getLink("Log out").click()

        self.assertEquals(self.browser.title.strip(),
                          "You have been logged out")

        # == Logging back in with new password ==

        # To log back in the user will have to use their new password:

        self.browser.open('http://openid.launchpad.dev')
        self.browser.getControl(name='email').value = 'test@canonical.com'
        self.browser.getControl(name='password').value = 'TestPass23'
        self.browser.getControl(name='continue').click()

        self.assertEquals(self.browser.title.strip(), "Sample Person's details")

        # Cleanup
        account = Account.objects.get_by_email('test@canonical.com')
        account.displayname = "Sample Person"
        account.save()

        from identityprovider.utils import encrypt_launchpad_password
        account = Account.objects.get_by_email('test@canonical.com')
        account.accountpassword.password = encrypt_launchpad_password('test')
        account.accountpassword.save()

