from identityprovider.tests.functional.helpers import FunctionalTestCase

from identityprovider.models.account import Account
from identityprovider.models.emailaddress import EmailAddress
from identityprovider.models.const import LoginTokenType
from identityprovider.models import AuthToken


class StandaloneLoginTestCase(FunctionalTestCase):
    def test(self):
        # = Logging into openid.launchpad.dev without an OpenID request =

        # We allow users accessing openid.launchpad.dev directly (instead of
        # being sent there from an OpenID RP) to login, so that they can
        # see/change their details.

        self.browser.open('http://openid.launchpad.dev')

        self.assertContains("Log in to Ubuntu Single Sign On", id='content')
        self.assertContains("New account", id='content')

        def login(browser, email=None, password=None):
            browser.open('http://openid.launchpad.dev')
            browser.getControl(name='email').value = email
            browser.getControl(name='password').value = password
            browser.getControl(name='continue').click()

        login(self.browser, email='mark@example.com', password='test')

        # Once successfully logged in, they will see the details of their account.
        self.assertEquals(self.browser.url, "http://openid.launchpad.dev/")
        self.assertContains("Full name", id='content')
        self.assertContains("mark@example.com", id='content')


        # == Creating a new account ==

        def register(browser):
            try:
                EmailAddress.objects.filter(
                    email='new-user@example.com').delete()
            except:
                pass
            try:
                Account.objects.get_by_email('new-user@example.com').delete()
            except:
                pass
            browser.open('http://openid.launchpad.dev')
            browser.getLink('New account').click()
            browser.getControl(name='displayname').value = 'New User'
            browser.getControl(name='email').value = 'new-user@example.com'
            browser.getControl(name='password').value = 'testP4ss'
            browser.getControl(name='passwordconfirm').value = 'testP4ss'
            browser.getControl(name='continue').click()

        self.reset_browser()
        register(self.browser)

        self.assertContains("Registration mail sent")

        def find_my_token(type, email='new-user@example.com'):
            return AuthToken.objects.filter(
                email=email,
                token_type=type,
                date_consumed=None).order_by('-date_created')[0].token

        token = find_my_token(LoginTokenType.NEWPERSONLESSACCOUNT)
        link = 'http://openid.launchpad.dev/confirm-account/' + \
               ('%s/new-user@example.com' % token)

        self.browser.open(link)

        self.assertRegexpMatches(
            self.browser.url,
            'http://openid.launchpad.dev/confirm-account/.*?'
            '/new-user@example.com')

        # Finish the registration process.

        self.browser.getControl(name='continue').click()

        self.assertEquals(self.browser.url, "http://openid.launchpad.dev/")

        # Registering using this login form should not create a person, but
        # currently it does.
        person = Account.objects.get_by_email(email='new-user@example.com').person
        self.assertTrue(person is None)


        # === Confirm out of session ===

        # Begin registration as before. 
        self.reset_browser()
        register(self.browser)

        self.assertContains("Registration mail sent")


        # But now, try to confirm the new account from out of the original
        # session.
        self.reset_browser()
        token = find_my_token(LoginTokenType.NEWPERSONLESSACCOUNT)
        link = 'http://openid.launchpad.dev/confirm-account/' + \
               ('%s/new-user@example.com' % token)
        self.browser.open(link)

        # Continue as before.

        self.browser.getControl(name='continue').click()

        self.assertEquals(self.browser.url, "http://openid.launchpad.dev/")


        # == Resetting the password ==

        def request_reset(browser):
            browser.open('http://openid.launchpad.dev')
            browser.getLink('Forgot your password?').click()
            browser.getControl(name='email').value = 'new-user@example.com'
            browser.getControl(name='continue').click()

        self.reset_browser()
        request_reset(self.browser)

        self.assertContains("Forgotten your password?")

        token = find_my_token(LoginTokenType.PASSWORDRECOVERY)
        link = 'http://openid.launchpad.dev/token/%s' % token
        self.browser.open(link)

        self.assertRegexpMatches(
            self.browser.url,
            'http://openid.launchpad.dev/token/.*?'
            '/\+resetpassword/new-user@example.com')

        def set_new_password(browser):
            browser.getControl(name='password').value = 'new Password'
            browser.getControl(name='passwordconfirm').value = 'new Password'
            browser.getControl(name='continue').click()

        set_new_password(self.browser)

        self.assertEqual(self.browser.url, 'http://openid.launchpad.dev/')

        # Log in with the new password.
        self.reset_browser()
        login(self.browser, email='new-user@example.com',
              password='new Password')
        self.assertContains("New User", id='subheader')


        # === Confirm out of session ===

        # Start as before.
        self.reset_browser()
        request_reset(self.browser)

        self.assertContains("Forgotten your password?")

        # But now, try to finish the reset from out of the original session.
        self.reset_browser()
        token = find_my_token(LoginTokenType.PASSWORDRECOVERY)
        link = 'http://openid.launchpad.dev/token/%s' % token
        self.browser.open(link)

        # It should prompt for the e-mail that the token was sent to.
        self.assertRegexpMatches(self.browser.url,
                                 'http://.*?/\+enter_token\?confirmation_code=.*?')

        # Fill in the e-mail address, then continue as before.
        self.browser.getControl(name='email').value = 'new-user@example.com'
        self.browser.getControl(name='continue').click()

        self.assertRegexpMatches(self.browser.url,
                                 'http://openid.launchpad.dev/token/.*?'
                                 '/\+resetpassword/new-user@example.com')

        set_new_password(self.browser)

        self.assertEquals(self.browser.url, 'http://openid.launchpad.dev/')

        # Log in with the new password.
        self.reset_browser()
        login(self.browser, email='new-user@example.com',
              password='new Password')

        self.assertContains("New User", id='subheader')


        # == Adding an e-mail address ==
        self.reset_browser()
        login(self.browser, email='new-user@example.com',
              password='new Password')

        def add_email(browser, email):
            browser.open('http://openid.launchpad.dev/+edit')
            browser.getLink('Manage email addresses').click()
            browser.getControl(name='newemail').value = email
            browser.getControl(name='continue').click()

        add_email(self.browser, 'alpha@example.org')

        token = find_my_token(LoginTokenType.VALIDATEEMAIL,
                              email='alpha@example.org')
        self.browser.open('http://openid.launchpad.dev/token/%s' % token)

        self.assertRegexpMatches(self.browser.url,
                                 'http://openid.launchpad.dev/token/.*?'
                                 '/\+newemail/alpha@example.org')

        self.browser.getControl(name='continue').click()

        self.reset_browser()
        login(self.browser, email='alpha@example.org', password='new Password')

        self.assertContains("Full name", id='content')
        self.assertContains("new-user@example.com", id='content')

        # cleanup
        EmailAddress.objects.filter(email='alpha@example.org').delete()
