# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from django.contrib import admin
from django.contrib.admin.sites import AlreadyRegistered, NotRegistered

from identityprovider.admin import (AccountPasswordInline,
    EmailAddressInline)
from identityprovider.models import (EmailAddress,
                                     OpenIDRPConfig, APIUser, Person)
from identityprovider.models.account import (Account, AccountPassword,
                                             LPOpenIdIdentifier)
from identityprovider.models.const import EmailStatus
from identityprovider.utils import validate_launchpad_password

from identityprovider.tests.utils import SSOBaseTestCase


class AdminTestCase(SSOBaseTestCase):

    fixtures = ['admin', 'test']

    def setUp(self):
        self.disable_csrf()
        self.client.login(username="admin", password="admin007")

    def tearDown(self):
        self.reset_csrf()

    def test_registered_models(self):
        for model in (Account, OpenIDRPConfig, APIUser, EmailAddress):
            self.assertRaises(AlreadyRegistered, admin.site.register, model)

        for model in (AccountPassword, Person, LPOpenIdIdentifier):
            self.assertRaises(NotRegistered, admin.site.unregister, model)

    def test_openidrpconfig_allowed_sreg_checkboxes_postable(self):
        data = {'trust_root': 'http://localhost/bla/',
                'displayname': 'My Test RP',
                'description': 'Bla',
                'allowed_sreg': ['fullname', 'email'],
                'creation_rationale': '13',
                }
        response = self.client.get(
            '/admin/identityprovider/openidrpconfig/add/')
        response = self.client.post(
            '/admin/identityprovider/openidrpconfig/add/',
            data)
        self.assertEquals(302, response.status_code)
        # We don't get the ID back, so continue on to the next test to
        # verify that the checkbox appears on the model's screen.

    def test_openidrpconfig_allowed_sreg_checkboxes_getable(self):
        data = {'trust_root': 'http://localhost/bla/',
                'displayname': 'My Test RP',
                'description': 'Bla',
                'allowed_sreg': 'fullname',
                'creation_rationale': '13',
                }
        rpconfig = OpenIDRPConfig(**data)
        rpconfig.save()
        response = self.client.get('/admin/identityprovider/openidrpconfig/%s/'
                                   % rpconfig.id)
        self.assertContains(response, 'checked')

    def test_inline_models(self):
        expected_inlines = [AccountPasswordInline, EmailAddressInline]
        registered_inlines = admin.site._registry[Account].inlines
        self.assertEqual(registered_inlines, expected_inlines)

    def test_account_overview(self):
        r = self.client.get('/admin/identityprovider/account/')
        self.assertContains(r, 'mark@example.com')

    def test_account_change(self):
        account_id = 1
        url = '/admin/identityprovider/account/%s/' % account_id
        account = Account.objects.get(pk=account_id)
        email = account.preferredemail
        new_email = 'mark2@example.com'
        new_password = 'blah'
        r = self.client.post(url, {
            'creation_rationale': str(account.creation_rationale),
            'status': str(account.status),
            'displayname': account.displayname,
            'openid_identifier': account.openid_identifier,
            'accountpassword-TOTAL_FORMS': '1',
            'accountpassword-INITIAL_FORMS': '1',
            'accountpassword-0-id': str(account_id),
            'accountpassword-0-account': str(account_id),
            'accountpassword-0-password': new_password,
            'emailaddress_set-TOTAL_FORMS': '1',
            'emailaddress_set-INITIAL_FORMS': '1',
            'emailaddress_set-0-id': str(email.id),
            'emailaddress_set-0-account': str(account.id),
            'emailaddress_set-0-email': new_email,
            'emailaddress_set-0-status': str(email.status)})
        # Any non-error status code would be fine here, but right now
        # it redirects with a 302.
        self.assertEqual(r.status_code, 302)
        account = Account.objects.get(pk=1)
        self.assertEqual(account.preferredemail.email, new_email)
        self.assertTrue(validate_launchpad_password(
            new_password, account.accountpassword.password))

    def test_multiple_preferred_emails(self):
        account_id = 1
        url = '/admin/identityprovider/account/%s/' % account_id
        account = Account.objects.get(pk=account_id)
        email = account.preferredemail
        r = self.client.post(url, {
            'creation_rationale': str(account.creation_rationale),
            'status': str(account.status),
            'displayname': account.displayname,
            'openid_identifier': account.openid_identifier,
            'accountpassword-TOTAL_FORMS': '1',
            'accountpassword-INITIAL_FORMS': '1',
            'accountpassword-0-id': str(account_id),
            'accountpassword-0-account': str(account_id),
            'emailaddress_set-TOTAL_FORMS': '2',
            'emailaddress_set-INITIAL_FORMS': '1',
            'emailaddress_set-0-id': str(email.id),
            'emailaddress_set-0-account': str(account.id),
            'emailaddress_set-0-email': email.email,
            'emailaddress_set-0-status': str(email.status),
            'emailaddress_set-1-account': str(account.id),
            'emailaddress_set-1-email': 'failure@example.com',
            'emailaddress_set-1-status': str(email.status)})
        account = Account.objects.get(pk=1)
        self.assertContains(r, 'Only one email address can be preferred.')
        self.assertEqual(len(account.emailaddress_set.filter(
            status=EmailStatus.PREFERRED)), 1)
        self.assertEqual(account.preferredemail, email)

    def test_inline_forms(self):
        r = self.client.get('/admin/identityprovider/account/1/')
        self.assertEqual(r.status_code, 200)
        self.assertContains(r, '<input type="password" '
                            'name="accountpassword-0-password"')

    def test_add_api_user(self):
        r = self.client.post('/admin/identityprovider/apiuser/add/', {
            'username': 'api_user',
            'password': 'backend'})
        self.assertEqual(r.status_code, 302)
