# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import tempfile
import os
import shutil

from django.db.backends.postgresql_psycopg2 import base as postgresql_base
from django.conf import settings
from django.db import connection

from identityprovider.branding import current_brand
from identityprovider.tests import mockdb
from identityprovider.readonly import ReadOnlyManager

from identityprovider.tests.utils import SSOBaseTestCase, skipOnSqlite


test_db_connections = [
    {
         'DATABASE_ID': 'testmaster',
         'DATABASE_NAME': settings.DATABASE_NAME,
         'DATABASE_USER': settings.DATABASE_USER,
         'DATABASE_PASSWORD': settings.DATABASE_PASSWORD,
         'DATABASE_HOST': settings.DATABASE_HOST,
         'DATABASE_PORT': settings.DATABASE_PORT,
    },
    {
         'DATABASE_ID': 'testslave',
         'DATABASE_NAME': settings.DATABASE_NAME,
         'DATABASE_USER': settings.DATABASE_USER,
         'DATABASE_PASSWORD': settings.DATABASE_PASSWORD,
         'DATABASE_HOST': settings.DATABASE_HOST,
         'DATABASE_PORT': settings.DATABASE_PORT,
    },
]


@skipOnSqlite
class FailoverTestCase(SSOBaseTestCase):

    def __init__(self, methodName='runTest'):
        super(FailoverTestCase, self).__init__(methodName)
        self.backups = {}

    def backup(self, setting, default=None):
        value = getattr(settings, setting, default)
        self.backups[setting] = value

    def restore(self, setting):
        setattr(settings, setting, self.backups.get(setting))

    def setUp(self):
        self.original_db = postgresql_base.Database
        postgresql_base.Database = mockdb
        try:
            dsn = settings.TEST_DSN
        except AttributeError:
            print """*** To run the tests you need to provide a TEST_DSN
            setting that will be used throughout the tests, instead
            of your configured DB_CONNECTIONS.
            """
            raise
        mockdb.fixate_connection(dsn)

        self.rm = ReadOnlyManager()
        connection.connection = None

    def tearDown(self):
        self.rm.clear_readonly()
        for db in self.rm.connections:
            self.rm.clear_failed(db['DATABASE_ID'])
        mockdb.fail_next_N_connections(0)
        for setting in self.backups:
            self.restore(setting)
        postgresql_base.Database = self.original_db
        mockdb.release_connection()

    def test_retries(self):
        """ Check that we attempt to connect N times """
        self.backup('DBFAILOVER_ATTEMPTS')
        settings.DBFAILOVER_ATTEMPTS = 3
        mockdb.fail_next_N_connections(2)
        self.client.get('/')

    def test_master_fail(self):
        """Check that we switch to readonly when the master fails."""
        self.backup('DBFAILOVER_ATTEMPTS')
        self.backup('DB_CONNECTIONS')
        settings.DBFAILOVER_ATTEMPTS = 3
        settings.DB_CONNECTIONS = test_db_connections
        mockdb.fail_next_N_connections(4)
        self.client.get('/')
        self.assert_(settings.READ_ONLY_MODE)
        self.assertEquals('testslave', settings.DATABASE_ID)
        self.rm.clear_readonly()
        self.rm.clear_failed('testmaster')
        self.assertFalse(settings.READ_ONLY_MODE)

    def test_complete_fail(self):
        """Check that a friendly oops screen is presented when all connections
           fail, instead of a plain backtrace."""
        self.backup('DBFAILOVER_ATTEMPTS')
        self.backup('DB_CONNECTIONS')
        settings.DBFAILOVER_ATTEMPTS = 3
        settings.DB_CONNECTIONS = test_db_connections
        mockdb.fail_next_N_connections(6)
        response = self.client.get('/')
        apology = (u"Sorry, something just went wrong in %s." %
                   current_brand.name)
        self.assertContains(response, apology, status_code=500)

    def test_files_are_in_the_right_place(self):
        """Check that flag files are created in DBFAILOVER_FLAG_DIR."""
        self.backup('DBFAILOVER_FLAG_DIR', '/tmp')
        self.backup('DB_CONNECTIONS')
        self.backup('DBFAILOVER_ATTEMPTS')
        flagdir = tempfile.mkdtemp(dir='/tmp')
        try:
            settings.DBFAILOVER_FLAG_DIR = flagdir
            settings.DB_CONNECTIONS = test_db_connections
            settings.DBFAILOVER_ATTEMPTS = 3
            mockdb.fail_next_N_connections(4)
            self.client.get('/')
            flags = os.listdir(flagdir)
            self.assert_(len(flags) == 2)
        finally:
            shutil.rmtree(flagdir, ignore_errors=True)
