# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import urllib2
from datetime import datetime

from django import forms
from django.test import TestCase
from django.conf import settings
from mock import (
    patch,
    Mock,
    )

from canonical.isd.tests.django import switch_settings
from canonical.isd.tests.wsgi_intercept import WSGIInterceptedTestCase

from identityprovider.fields import (
    CommaSeparatedField,
    OneTimePasswordField,
    )


class CommaSeparatedFieldTestCase(TestCase):

    def test_clean(self):
        choices = (('1', 'One'), ('2', 'Two'), ('3', 'Three'))
        field = CommaSeparatedField(choices=choices)
        value = field.clean(["1", "2", "3"])
        self.assertEquals(value, "1,2,3")


class OneTimePaswordFieldTestCase(TestCase):

    YUBIKEY_RESPONSE_REPLAY = (
            'h=LnOtAtozUA1ZjDKlHaVUIQIxkJM=\r\n'
            't=2011-05-10T11:57:11Z0777\r\n'
            'otp=cccccccitngfckhffcuudlkhjkkbtjklvubibhghvdel\r\n'
            'nonce=3df9ec99ef2796f6770a084c54883b\r\n'
            'status=REPLAYED_OTP\r\n\r\n')

    YUBIKEY_RESPONSE_OK = (
            'h=LnOtAtozUA1ZjDKlHaVUIQIxkJM=\r\n'
            't=2011-05-10T11:57:11Z0777\r\n'
            'otp=cccccccitngfckhffcuudlkhjkkbtjklvubibhghvdel\r\n'
            'nonce=3df9ec99ef2796f6770a084c54883b\r\n'
            'status=OK\r\n\r\n')

    def do_clean(self, value, mock=None):
        otp_field = OneTimePasswordField()
        mock = mock or Mock(return_value=dict(status='OK'))
        otp_field.request_otp_validation = mock

        return otp_field.clean(value)

    def test_valid_chars_accepted(self):
        good_otp = 'cccccccitngfckhffcuudlkhjkkbtjklvubibhghvdel'

        self.assertEqual(good_otp, self.do_clean(good_otp))

    def test_invalid_length_not_accepted(self):
        less_than_34 = 'cmcccccitngflibeilnkcdhcutrjn'
        more_than_48 = 'cccccccitngflibeilnkcdhcutrjnkbkbbejggfrvubcccccccc'

        self.assertRaises(forms.ValidationError, self.do_clean, less_than_34)
        self.assertRaises(forms.ValidationError, self.do_clean, more_than_48)

    def test_server_validation_fail_raises_error(self):
        otp_field = OneTimePasswordField()
        good_otp = 'cccccccitngfckhffcuudlkhjkkbtjklvubibhghvdel'
        # We want to check exactly what urlopen is called with.
        mock_urlopen = Mock()
        mock_urlopen.return_value.read.return_value = (
            self.YUBIKEY_RESPONSE_REPLAY)

        with patch('urllib2.urlopen', mock_urlopen):
            self.assertRaises(
                forms.ValidationError, otp_field.clean, good_otp)

    def test_convert_response_to_dict(self):
        otp_field = OneTimePasswordField()

        result = otp_field.convert_response_to_dict(
            self.YUBIKEY_RESPONSE_REPLAY)

        self.assertEqual(dict(
            h='LnOtAtozUA1ZjDKlHaVUIQIxkJM=', t='2011-05-10T11:57:11Z0777',
            otp='cccccccitngfckhffcuudlkhjkkbtjklvubibhghvdel',
            nonce='3df9ec99ef2796f6770a084c54883b',
            status='REPLAYED_OTP'), result)

    def test_urlopen_called_correctly(self):
        otp_field = OneTimePasswordField()
        otp = 'my_otp11111111111111111111111111111'
        # Ensure we control what the nonce will be.
        mock_sha1 = Mock()
        mock_sha1.return_value.hexdigest.return_value = 'mock_nonce'
        # We want to check exactly what urlopen is called with.
        mock_urlopen = Mock()
        mock_urlopen.return_value.read.return_value = (
            self.YUBIKEY_RESPONSE_OK)

        with patch('urllib2.urlopen', mock_urlopen):
            with patch('hashlib.sha1', mock_sha1):
                result = otp_field.clean(otp)

        mock_urlopen.assert_called_with(
            settings.YUBICO_API_URL.format(
                otp=otp, client_id=settings.YUBICO_CLIENT_ID,
                nonce='mock_nonce'))
        self.assertEqual(otp, result)

    def test_urlopen_exception(self):
        # clean raises an appropriate validation error if urlopen raises
        # a URLError.
        otp_field = OneTimePasswordField()
        otp = 'my_otp11111111111111111111111111111'
        mock_urlopen = Mock(side_effect=urllib2.URLError('Bang!'))

        with patch('urllib2.urlopen', mock_urlopen):
            self.assertRaises(
                forms.ValidationError, otp_field.clean, otp)


class OneTimePaswordFieldWithMockServerTestCase(TestCase, WSGIInterceptedTestCase):

    def setUp(self):

        # add yubi_test_server to the PYTHONPATH
        import identityprovider
        import sys
        sys.path.append(identityprovider.__path__)
        from yubi_test_server import make_app

        callbacks = {
            'default': make_app,
            ('localhost', 8080): make_app}
        self.setup_intercept(callbacks, intercept_api=True)

        self.old_settings = switch_settings(
            YUBICO_API_URL='http://localhost:8080/wsapi/2.0/verify?id={client_id}&otp={otp}&nonce={nonce}')

    def tearDown(self):
        self.teardown_intercept()
        switch_settings(**self.old_settings)

    def test_ok(self):
        otp_field = OneTimePasswordField()
        otp = '0000'
        resp_fields = otp_field.request_otp_validation(otp)
        status = resp_fields['status']
        self.assertEqual(status, 'OK')

    def test_bad_otp(self):
        otp_field = OneTimePasswordField()
        otp = '1111'
        resp_fields = otp_field.request_otp_validation(otp)
        status = resp_fields['status']
        self.assertEqual(status, 'BAD_OTP')

    def test_replayed_otp(self):
        otp_field = OneTimePasswordField()
        otp = '2222'
        resp_fields = otp_field.request_otp_validation(otp)
        status = resp_fields['status']
        self.assertEqual(status, 'REPLAYED_OTP')

    def test_bad_signature(self):
        otp_field = OneTimePasswordField()
        otp = '3333'
        resp_fields = otp_field.request_otp_validation(otp)
        status = resp_fields['status']
        self.assertEqual(status, 'BAD_SIGNATURE')

    def test_missing_parameter(self):
        otp_field = OneTimePasswordField()
        otp = '4444'
        resp_fields = otp_field.request_otp_validation(otp)
        status = resp_fields['status']
        self.assertEqual(status, 'MISSING_PARAMETER')

    def test_no_such_client(self):
        otp_field = OneTimePasswordField()
        otp = '5555'
        resp_fields = otp_field.request_otp_validation(otp)
        status = resp_fields['status']
        self.assertEqual(status, 'NO_SUCH_CLIENT')

    def test_operation_not_allowed(self):
        otp_field = OneTimePasswordField()
        otp = '6666'
        resp_fields = otp_field.request_otp_validation(otp)
        status = resp_fields['status']
        self.assertEqual(status, 'OPERATION_NOT_ALLOWED')

    def test_backend_error(self):
        otp_field = OneTimePasswordField()
        otp = '7777'
        resp_fields = otp_field.request_otp_validation(otp)
        status = resp_fields['status']
        self.assertEqual(status, 'BACKEND_ERROR')

    def test_not_enough_answers(self):
        otp_field = OneTimePasswordField()
        otp = '8888'
        resp_fields = otp_field.request_otp_validation(otp)
        status = resp_fields['status']
        self.assertEqual(status, 'NOT_ENOUGH_ANSWERS')

    def test_replayed_request(self):
        otp_field = OneTimePasswordField()
        otp = '9999'
        resp_fields = otp_field.request_otp_validation(otp)
        status = resp_fields['status']
        self.assertEqual(status, 'REPLAYED_REQUEST')

    @patch('identityprovider.fields.datetime')
    def test_request_otp_validation_use_utc(self, mock_datetime):
        mock_datetime.utcnow.return_value = datetime.utcnow()
        mock_datetime.side_effect = lambda *args, **kw: datetime(*args, **kw)

        otp_field = OneTimePasswordField()
        otp = '0000'
        resp_fields = otp_field.request_otp_validation(otp)
        status = resp_fields['status']

        self.assertEqual(status, 'OK')
        self.assertEqual(mock_datetime.utcnow.called, True)
