# -*- coding: utf-8 -*-

# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from django.conf import settings
from django.http import HttpRequest
from django_openid_auth.teams import TeamsRequest
from openid.extensions.sreg import SRegRequest

from identityprovider.models.account import Account
from identityprovider.forms import (
    EditAccountForm,
    LoginForm,
    ResetPasswordForm,
    SRegRequestForm,
    TeamsRequestForm,
    TokenForm,
)
from identityprovider.api10.forms import WebserviceCreateAccountForm
from identityprovider.models.openidmodels import OpenIDRPConfig

from identityprovider.tests.utils import SSOBaseTestCase, patch_settings
from identityprovider.tests.factory import TestCaseWithFactory


class EditAccountFormTest(SSOBaseTestCase):

    fixtures = ["test"]

    def test_initial_displayname(self):
        account = Account.objects.get(openid_identifier='name12_oid')
        form = EditAccountForm(account=account)
        self.assertEqual(form.initial.get('displayname', ''), "Sample Person")

    def test_nonascii_password(self):
        account = Account.objects.get(openid_identifier='name12_oid')
        data = {'password': 'changüí'}
        form = EditAccountForm(account=account, data=data)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['password'][0],
                         'Invalid characters in password')

    def test_no_preferredemail(self):
        # replace preferredemail property
        _preferredemail = Account.preferredemail
        Account.preferredemail = None

        account = Account.objects.get(openid_identifier='name12_oid')
        form = EditAccountForm(account=account)
        self.assertEqual(form.initial.get('displayname', ''), "Sample Person")

        Account.preferredemail = _preferredemail

    def test_displayname_validation(self):
        account = Account.objects.get(openid_identifier='name12_oid')
        form = EditAccountForm(account=account, data={'displayname': ' '})
        self.assertFalse(form.is_valid())
        self.assertTrue(form.errors['displayname'] == [u'Required field'])
        form = EditAccountForm(account=account,
                                data={'displayname': 'Sample Person'})
        self.assertFalse(form.is_valid())
        self.assertFalse('displayname' in form.errors)

    def test_newpassword_mismatch_validation(self):
        account = Account.objects.get(openid_identifier='name12_oid')
        form = EditAccountForm(account=account,
                               data={'password': 'tooweak',
                                     'passwordconfirm': 'other'})
        self.assertFalse(form.is_valid())
        self.assertEqual(1, len(form.errors['password']))
        self.assertFalse('passwordconfirm' in form.errors)
        self.assertEqual(0, len(form.non_field_errors()))

    def test_newpassword_validation(self):
        account = Account.objects.get(openid_identifier='name12_oid')
        form = EditAccountForm(account=account,
                               data={'password': 'tooweak',
                                     'passwordconfirm': 'tooweak'})
        self.assertFalse(form.is_valid())
        self.assertEqual(1, len(form.errors['password']))
        self.assertFalse('passwordconfirm' in form.errors)
        self.assertEqual(0, len(form.non_field_errors()))


class WebServiceCreateAccountFormTest(SSOBaseTestCase):

    def setUp(self):
        self.old_dcv = settings.DISABLE_CAPTCHA_VERIFICATION
        settings.DISABLE_CAPTCHA_VERIFICATION = True

    def tearDown(self):
        settings.DISABLE_CAPTCHA_VERIFICATION = self.old_dcv

    def test_nonascii_password(self):
        data = {'password': 'Curuzú Cuatiá',
                'remote_ip': '127.0.0.1'}
        form = WebserviceCreateAccountForm(data=data)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['password'][0],
                         'Invalid characters in password')

    def test_default_platform(self):
        form = WebserviceCreateAccountForm()
        self.assertEqual(form.fields['platform'].empty_value, 'desktop')

    def test_default_cleaned_platform(self):
        data = {
            'email': 'some@email.com',
            'password': 'password1A',
            'captcha_id': '1',
            'captcha_solution': '2',
            'remote_ip': '127.0.0.1',
        }
        form = WebserviceCreateAccountForm(data)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['platform'], 'desktop')

    def test_captcha_checked_for_whitelist(self):
        data = {
            'email': 'canonicaltest@gmail.com',
            'password': 'password1A',
            'captcha_id': '1',
            'captcha_solution': '2',
            'remote_ip': '127.0.0.1',
        }
        pattern = '^canonicaltest(?:\+.+)?@gmail\.com$'
        overrides = dict(
            DISABLE_CAPTCHA_VERIFICATION=False,
            EMAIL_WHITELIST_REGEXP_LIST=[pattern],
        )
        with patch_settings(**overrides):
            form = WebserviceCreateAccountForm(data)
            self.assertTrue(form.is_valid())

    def test_default_cleaned_validate_redirect_to(self):
        data = {
            'email': 'some@email.com',
            'password': 'password1A',
            'captcha_id': '1',
            'captcha_solution': '2',
            'remote_ip': '127.0.0.1',
        }
        form = WebserviceCreateAccountForm(data)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['validate_redirect_to'], None)


class ResetPasswordFormTest(SSOBaseTestCase):

    def test_nonascii_password(self):
        data = {'password': 'Ñandú'}
        form = ResetPasswordForm(data=data)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['password'][0],
                         'Invalid characters in password')


class SRegRequestFormTest(SSOBaseTestCase):

    def _get_request_with_post_args(self, args={}):
        request = HttpRequest()
        request.user = self.test_user
        request.POST = args
        request.META = {'REQUEST_METHOD': 'POST'}
        return request

    def setUp(self):
        self.test_user = Account.objects.create_account('My name',
            'me@test.com', 'password')
        self.rpconfig = OpenIDRPConfig.objects.create(
            trust_root='http://localhost/', description="Some description")

    def test_no_approved_fields_without_post_request(self):
        """The server should not generate a list of approved fields when the
        request is not a POST request.
        """
        request = self._get_request_with_post_args()
        request.META['REQUEST_METHOD'] = 'GET'
        form = SRegRequestForm(
            request,
            SRegRequest(required=['fullname', 'email']),
            self.rpconfig)
        self.assertEqual(len(form.data_approved_for_request), 0)

    def test_required_fields_for_trusted_site(self):
        """The server should always return values for required fields to trusted
        sites, regardless of the state of the checkbox in the UI.  Optional
        fields should not be returned if the user has unchecked them.
        """
        self.rpconfig.allowed_sreg = 'fullname,email'
        form = SRegRequestForm(
            self._get_request_with_post_args(),
            SRegRequest(required=['fullname'], optional=['email']),
            self.rpconfig)
        self.assertTrue('fullname' in form.data_approved_for_request)
        self.assertFalse('email' in form.data_approved_for_request)

    def test_optional_fields_for_trusted_site(self):
        """The server should return values for optional fields to trusted
        sites only when the user checks the checkbox in the UI.
        """
        self.rpconfig.allowed_sreg = 'fullname,email'
        post_args = {'email': 'email'}
        form = SRegRequestForm(
            self._get_request_with_post_args(post_args),
            SRegRequest(optional=['fullname', 'email']),
            self.rpconfig)
        self.assertFalse('fullname' in form.data_approved_for_request)
        self.assertTrue('email' in form.data_approved_for_request)

    def test_required_fields_for_untrusted_site(self):
        """The server should return values for required fields to untrusted
        sites only when the user checks the checkbox in the UI.
        """
        post_args = {'email': 'email'}
        form = SRegRequestForm(
            self._get_request_with_post_args(post_args),
            SRegRequest(required=['fullname', 'email']),
            None)
        self.assertFalse('fullname' in form.data_approved_for_request)
        self.assertTrue('email' in form.data_approved_for_request)

    def test_optional_fields_for_untrusted_site(self):
        """The server should return values for optional fields to untrusted
        sites only when the user checks the checkbox in the UI.
        """
        post_args = {'fullname': 'fullname'}
        form = SRegRequestForm(
            self._get_request_with_post_args(post_args),
            SRegRequest(optional=['fullname', 'email']),
            None)
        self.assertTrue('fullname' in form.data_approved_for_request)
        self.assertFalse('email' in form.data_approved_for_request)

    def test_checkbox_status_for_trusted_site(self):
        """Checkboxes are always checked if the site is trusted
        """
        form = SRegRequestForm(
            self._get_request_with_post_args(),
            SRegRequest(required=['fullname'], optional=['email']),
            self.rpconfig)
        self.assertTrue(form.check_test('fullname'))
        self.assertTrue(form.check_test('email'))

    def test_checkbox_status_for_trusted_site_with_approved_data(self):
        """If the user has previously approved sending data to a trusted site
        the same checkbox settings should be returned on the next request unless
        those conflict with the required fields.
        """
        approved_data = {
            'requested': ['fullname', 'email'],
            'approved': ['email']}
        form1 = SRegRequestForm(
            self._get_request_with_post_args(),
            SRegRequest(required=['fullname'], optional=['email']),
            self.rpconfig, approved_data=approved_data)
        self.assertTrue(form1.check_test('fullname'))
        self.assertTrue(form1.check_test('email'))

        approved_data['approved'] = []
        form2 = SRegRequestForm(
            self._get_request_with_post_args(),
            SRegRequest(required=['fullname'], optional=['email']),
            self.rpconfig, approved_data=approved_data)
        self.assertFalse(form2.check_test('email'))

    def test_checkbox_status_for_untrusted_site(self):
        """Checkboxes are only checked on untrusted site requests if the field
        is required
        """
        form = SRegRequestForm(
            self._get_request_with_post_args(),
            SRegRequest(required=['fullname'], optional=['email']),
            None)
        self.assertTrue(form.check_test('fullname'))
        self.assertFalse(form.check_test('email'))

    def test_checkbox_status_for_untrusted_site_with_approved_data(self):
        """If the user has previously approved sending data to an untrusted site
        the same checkbox settings should be returned on the next request.
        """
        approved_data = {
            'requested': ['fullname', 'email'],
            'approved': ['email']}
        form = SRegRequestForm(
            self._get_request_with_post_args(),
            SRegRequest(required=['fullname'], optional=['email']),
            None, approved_data=approved_data)
        self.assertFalse(form.check_test('fullname'))
        self.assertTrue(form.check_test('email'))


class TeamsRequestFormTest(SSOBaseTestCase):

    fixtures = ["test"]

    def _get_request_with_post_args(self, args={}):
        request = HttpRequest()
        request.user = self.account
        request.POST = args
        request.META = {'REQUEST_METHOD': 'POST'}
        return request

    def setUp(self):
        self.account = Account.objects.get(pk=1)
        self.rpconfig = OpenIDRPConfig.objects.create(
            trust_root='http://localhost/', description="Some description",
            can_query_any_team=True)

    def test_selected_teams_for_trusted_sites(self):
        """If a user checks a requested team in the form for a trusted consumer,
        it should be in the list of teams approved by the user.
        """
        post_args = {'ubuntu-team': 'ubuntu-team'}
        form = TeamsRequestForm(
            self._get_request_with_post_args(post_args),
            TeamsRequest(query_membership=['ubuntu-team']),
            self.rpconfig)
        self.assertTrue('ubuntu-team' in form.teams_approved_by_user)

    def test_unselected_teams_for_trusted_sites(self):
        """If a user unchecks a requested team in the form for a trusted
        consumer, it should not be in the list of teams approved by the user.
        """
        form = TeamsRequestForm(
            self._get_request_with_post_args(),
            TeamsRequest(query_membership=['ubuntu-team']),
            self.rpconfig)
        self.assertFalse('ubuntu-team' in form.teams_approved_by_user)

    def test_selected_teams_for_untrusted_sites(self):
        """If a user checks a requested team in the form for an untrusted
        consumer, it should be in the list of teams approved by the user.
        """
        post_args = {'ubuntu-team': 'ubuntu-team'}
        form = TeamsRequestForm(
            self._get_request_with_post_args(post_args),
            TeamsRequest(query_membership=['ubuntu-team']),
            None)
        self.assertTrue('ubuntu-team' in form.teams_approved_by_user)

    def test_unselected_teams_for_untrusted_sites(self):
        """If a user unchecks a requested team in the form for an untrusted
        consumer, it should not be in the list of teams approved by the user.
        """
        form = TeamsRequestForm(
            self._get_request_with_post_args(),
            TeamsRequest(query_membership=['ubuntu-team']),
            None)
        self.assertFalse('ubuntu-team' in form.teams_approved_by_user)

    def test_checkbox_status_for_trusted_site(self):
        """Checkboxes should always be checked by default for trusted sites.
        """
        form = TeamsRequestForm(
            self._get_request_with_post_args(),
            TeamsRequest(query_membership=['ubuntu-team']),
            self.rpconfig)
        self.assertTrue(form.check_test('ubuntu-team'))

    def test_checkbox_status_for_trusted_site_with_approved_data(self):
        """Checkboxes should respect user preferences on trusted sites where
        available.
        """
        approved_data = {
            'requested': ['ubuntu-team', 'myteam'],
            'approved': ['myteam']}
        form = TeamsRequestForm(
            self._get_request_with_post_args(),
            TeamsRequest(query_membership=['ubuntu-team', 'myteam']),
            self.rpconfig, approved_data=approved_data)
        self.assertFalse(form.check_test('ubuntu-team'))
        self.assertTrue(form.check_test('myteam'))

    def test_checkbox_status_for_untrusted_site(self):
        """Checkboxes should always be checked by default for trusted sites.
        """
        form = TeamsRequestForm(
            self._get_request_with_post_args(),
            TeamsRequest(query_membership=['ubuntu-team']),
            None)
        self.assertFalse(form.check_test('ubuntu-team'))

    def test_checkbox_status_for_untrusted_site_with_approved_data(self):
        """Checkboxes should respect user preferences on untrusted sited where
        available.
        """
        approved_data = {
            'requested': ['ubuntu-team'],
            'approved': ['ubuntu-team']}
        form1 = TeamsRequestForm(
            self._get_request_with_post_args(),
            TeamsRequest(query_membership=['ubuntu-team', 'myteam']),
            None, approved_data=approved_data)
        self.assertTrue(form1.check_test('ubuntu-team'))

        approved_data['approved'] = []
        form2 = TeamsRequestForm(
            self._get_request_with_post_args(),
            TeamsRequest(query_membership=['ubuntu-team', 'myteam']),
            None, approved_data=approved_data)
        self.assertFalse(form2.check_test('ubuntu-team'))


class TokenForrmTest(SSOBaseTestCase):

    def test_confirmation_code_error(self):
        data = {'confirmation_code': 'BOGUS', 'email': 'fake@example.com'}
        form = TokenForm(data=data)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.non_field_errors(), [])
        self.assertEqual(form.errors,
                         {'confirmation_code': [u'Unknown confirmation code.']})


class LoginFormTestCase(TestCaseWithFactory):
    def test_clean_email_mixed_case(self):
        self.factory.make_account(email='Test.User@test.com')
        data = {'email': 'Test.User@test.com', 'password': 'test'}
        form = LoginForm(data=data)
        self.assertTrue(form.is_valid())

        data = {'email': 'test.user@test.com', 'password': 'test'}
        form = LoginForm(data=data)
        self.assertTrue(form.is_valid())
