# encoding: utf-8
import base64
import json
from django.utils import simplejson
from mock import patch
from oauth.oauth import (
    OAuthRequest,
    OAuthConsumer,
    OAuthToken,
    OAuthSignatureMethod_PLAINTEXT,
    )
from lazr.restfulclient.errors import HTTPError

from identityprovider.api10.handlers import AuthenticationHandler
from identityprovider.tests.utils import AnonAPITestCase
from identityprovider.models import (
    Account,
    APIUser,
    AuthTokenFactory,
    EmailAddress,
)
from identityprovider.models.const import (
    AccountStatus,
    EmailStatus,
    LoginTokenType,
)
from identityprovider.tests.factory import TestCaseWithFactory


class MockRequest(object):
    def __init__(self, data=None, user=None):
        self.user = user
        if data is None:
            data = {}
        self.data = data
        self.environ = {'REMOTE_ADDR': '127.0.0.1'}


class RegistrationHandlerTestCase(AnonAPITestCase):

    def test_authtoken_without_requester_is_handled_properly(self):
        authtoken = AuthTokenFactory().new(None, None, "test@example.com",
                                           LoginTokenType.PASSWORDRECOVERY,
                                           None)

        response = self.api.registrations.set_new_password(
            email='test@example.com', token=authtoken.token, new_password='pass')

        self.assertEquals(response['status'], "error")

    def test_reset_token_when_email_is_invalid(self):
        try:
            self.api.registrations.request_password_reset_token(
                email='non-existing@example.com')
        except HTTPError, e:
            self.assertEqual(e.response.status, 403)
            self.assertTrue(e.content.startswith("CanNotResetPasswordError:"))

    def test_reset_token_when_account_is_disabled(self):
        email_address = self.factory.make_email_address()
        account = self.factory.make_account(email=email_address)
        account.status = AccountStatus.SUSPENDED
        account.save()

        try:
            self.api.registrations.request_password_reset_token(
                email=email_address)
        except HTTPError, e:
            self.assertEqual(e.response.status, 403)
            self.assertTrue(e.content.startswith("CanNotResetPasswordError:"))

    def test_reset_password_when_not_existing_token_is_passed(self):
        email_address = self.factory.make_email_address()
        try:
            self.api.registrations.set_new_password(
                email=email_address, token='not-existing-token',
                new_password='password')
        except HTTPError, e:
            self.assertEqual(e.response.status, 403)
            self.assertTrue(e.content.startswith(
                "CanNotSetPasswordError: Invalid token."))

    def test_register_email_error_is_in_list(self):
        EmailAddress.objects.create(email="register@example.com",
            status=EmailStatus.NEW)

        response = self.api.registrations.register(email='register@example.com',
            password='blogdf3Daa', captcha_solution='solution', captcha_id='id')

        self.assertEquals(response['errors'],
            {'email': ["Email already registered"]})

    def test_register_without_displayname(self):
        email = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id')
        self.assertEqual(response['status'], 'ok')

        account = Account.objects.get_by_email(email)
        self.assertEqual(account.displayname, '')

    def test_register_with_displayname(self):
        email = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id',
            displayname='Test User')
        self.assertEqual(response['status'], 'ok')

        account = Account.objects.get_by_email(email)
        self.assertEqual(account.displayname, 'Test User')

    @patch('identityprovider.models.authtoken.AuthToken.sendNewUserEmail')
    def test_register_without_platform(self, mock_sendNewUserEmail):
        email = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id')
        self.assertEqual(response['status'], 'ok')

        mock_sendNewUserEmail.assert_called_with(
            template='desktop-newuser.txt')

    @patch('identityprovider.models.authtoken.AuthToken.sendNewUserEmail')
    def test_register_with_platform(self, mock_sendNewUserEmail):
        email = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id',
            platform='mobile')
        self.assertEqual(response['status'], 'ok')

        mock_sendNewUserEmail.assert_called_with(
            template='mobile-newuser.txt')

    @patch('identityprovider.models.authtoken.AuthToken.sendNewUserEmail')
    def test_register_with_invalid_platform(self, mock_sendNewUserEmail):
        email = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id',
            platform='foo')

        self.assertEqual(response['status'], 'error')
        self.assertTrue('platform' in response['errors'])

    @patch('identityprovider.models.authtoken.AuthTokenFactory.'
        'new_api_email_validation_token')
    def test_register_without_redirection_url(self,
            mock_new_api_email_validation_token):
        email_address = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email_address, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id',
            platform='desktop')

        self.assertEqual(response['status'], 'ok')
        email = EmailAddress.objects.get(email=email_address)
        mock_new_api_email_validation_token.assert_called_with(
            email.account, email_address, None)

    @patch('identityprovider.models.authtoken.AuthTokenFactory.'
        'new_api_email_validation_token')
    def test_register_from_desktop_with_redirection_url(self,
            mock_new_api_email_validation_token):
        email_address = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email_address, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id',
            platform='desktop', validate_redirect_to='http://foo')

        self.assertEqual(response['status'], 'ok')
        email = EmailAddress.objects.get(email=email_address)
        mock_new_api_email_validation_token.assert_called_with(
            email.account, email_address, 'http://foo')

    @patch('identityprovider.models.authtoken.AuthTokenFactory.'
        'new_api_email_validation_token')
    def test_register_from_mobile_with_redirection_url(self,
        mock_new_api_email_validation_token):
        email_address = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email_address, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id',
            platform='mobile', validate_redirect_to='http://foo')

        self.assertEqual(response['status'], 'ok')
        account = Account.objects.get_by_email(email_address)
        self.assertEqual(None, account.preferredemail)
        unverified = account.unverified_emails()
        self.assertEqual([email_address],
            [emailaddress.email for emailaddress in unverified])

    @patch('identityprovider.models.authtoken.AuthTokenFactory.'
        'new_api_email_validation_token')
    def test_register_from_mobile_without_redirection_url(self,
        mock_new_api_email_validation_token):
        email_address = self.factory.make_email_address()

        response = self.api.registrations.register(
            email=email_address, password='MySecretPassword1',
            captcha_solution='foobar', captcha_id='id',
            platform='mobile')

        self.assertEqual(response['status'], 'ok')
        account = Account.objects.get_by_email(email_address)
        self.assertEqual(None, account.preferredemail)
        unverified = account.unverified_emails()
        self.assertEqual([email_address],
            [emailaddress.email for emailaddress in unverified])


class AuthenticationTestCase(TestCaseWithFactory):

    def setUp(self):
        self.authentication = AuthenticationHandler()

    def test_account_by_openid_with_valid_openid(self):
        account = self.factory.make_account()
        openid = account.openid_identifier

        request = MockRequest(user=APIUser(), data={'openid': openid})

        account = self.authentication.account_by_openid(request)

        self.assertEquals(account['openid_identifier'], openid)

    def test_account_by_openid_with_invalid_openid(self):
        request = MockRequest(user=APIUser(), data={'openid': "bad-openid"})

        account = self.authentication.account_by_openid(request)

        self.assertTrue(account is None)

    def test_authenticate_unquoted_token(self):
        email_address = self.factory.make_email_address()
        self.factory.make_account(email=email_address)
        expected_keys = ['consumer_key', 'consumer_secret', 'name', 'token',
            'token_secret']

        extra = {'HTTP_AUTHORIZATION': 'basic ' +
            base64.b64encode(email_address + ':test')}
        response = self.client.get('/api/1.0/authentications', data={
            'ws.op': 'authenticate', 'token_name': 'some-token'}, **extra)

        content = simplejson.loads(response.content)
        self.assertEqual(set(content.keys()), set(expected_keys))
        self.assertEqual(content['name'], 'some-token')

    def test_authenticate_quoted_token(self):
        email_address = self.factory.make_email_address()
        self.factory.make_account(email=email_address)
        expected_keys = ['consumer_key', 'consumer_secret', 'name', 'token',
            'token_secret']

        extra = {'HTTP_AUTHORIZATION': 'basic ' +
            base64.b64encode(email_address + ':test')}
        response = self.client.get('/api/1.0/authentications', data={
            'ws.op': 'authenticate', 'token_name': '"some-token"'}, **extra)


        content = simplejson.loads(response.content)
        self.assertEqual(set(content.keys()), set(expected_keys))
        self.assertEqual(content['name'], 'some-token')

    def test_team_memberships_public_teams(self):
        team = self.factory.make_team()
        otherteam = self.factory.make_team()
        self.factory.make_apiuser(username='foobar')
        account = self.factory.make_account()
        self.factory.add_account_to_team(account, team)

        # Request membership via API
        extra = {'HTTP_AUTHORIZATION': 'basic ' +
            base64.b64encode('foobar:test')}
        response = self.client.get('/api/1.0/authentications', data={
            'ws.op': 'team_memberships',
            'team_names': '["%s", "%s"]' % (team.name, otherteam.name),
            'openid_identifier': account.openid_identifier,
            }, **extra)
        self.assertContains(response, team.name)
        self.assertNotContains(response, otherteam.name)

    def test_team_membership_private_teams(self):
        team = self.factory.make_team(private=True)
        self.factory.make_apiuser(username='foobar')
        account = self.factory.make_account()
        self.factory.add_account_to_team(account, team)

        # Request membership via API
        extra = {'HTTP_AUTHORIZATION': 'basic ' +
            base64.b64encode('foobar:test')}
        response = self.client.get('/api/1.0/authentications', data={
            'ws.op': 'team_memberships',
            'team_names': '["%s"]' % team.name,
            'openid_identifier': account.openid_identifier,
            }, **extra)
        self.assertContains(response, team.name)

    def test_team_membership_basic_auth_protected(self):
        response = self.client.get('/api/1.0/authentications', data={
            'ws.op': 'team_memberships',
            'team_names': '["something"]',
            'openid_identifier': 'test',
            })
        self.assertEqual(401, response.status_code)

    def test_team_membership_fails_for_regular_user(self):
        email_address = self.factory.make_email_address()
        self.factory.make_account(email=email_address)
        extra = {'HTTP_AUTHORIZATION': 'basic ' +
            base64.b64encode(email_address + ':test')}
        response = self.client.get('/api/1.0/authentications', data={
            'ws.op': 'team_memberships',
            'team_names': '["something"]',
            'openid_identifier': 'test',
            }, **extra)
        self.assertEqual(403, response.status_code)


class AccountsTestCase(TestCaseWithFactory):

    def auth_header(self, token, url, realm='OAuth'):
        oaconsumer = OAuthConsumer(token.consumer.key, token.consumer.secret)
        oatoken = OAuthToken(token.token, token.token_secret)
        oarequest = OAuthRequest.from_consumer_and_token(
            oaconsumer, oatoken, http_url=url)
        oarequest.sign_request(OAuthSignatureMethod_PLAINTEXT(),
            oaconsumer, oatoken)
        header = oarequest.to_header(realm)
        return {'HTTP_AUTHORIZATION': header['Authorization']}


    def test_me_is_oauth_protected(self):
        email_address = self.factory.make_email_address()
        extra = {'HTTP_AUTHORIZATION': 'basic ' +
            base64.b64encode(email_address + ':test')}
        response = self.client.get('/api/1.0/accounts',
            data={'ws.op': 'me'}, **extra)
        self.assertEqual(401, response.status_code)

    def test_me_success(self):
        name = u'Ñânḑù'
        account = self.factory.make_account(displayname=name)
        token = account.create_oauth_token('token-name')
        url = '/api/1.0/accounts'

        response = self.client.get(url,
            data={'ws.op': 'me'}, **self.auth_header(token, url))

        data = json.loads(unicode(response.content))
        expected = set(['username', 'preferred_email', 'displayname',
            'unverified_emails', 'verified_emails', 'openid_identifier'])
        self.assertEqual(expected, set(data.keys()))
        self.assertEqual(data['displayname'], name)

    def test_team_membership_public_teams(self):
        account = self.factory.make_account()
        team = self.factory.make_team()
        otherteam = self.factory.make_team()
        self.factory.add_account_to_team(account, team)
        token = account.create_oauth_token('token-name')
        url = '/api/1.0/accounts'

        response = self.client.get(url, data={'ws.op': 'team_memberships',
            'team_names': '["%s", "%s"]' % (team.name, otherteam.name)},
            **self.auth_header(token, url))

        self.assertContains(response, team.name)
        self.assertNotContains(response, otherteam.name)

    def test_team_membership_private_teams(self):
        account = self.factory.make_account()
        team = self.factory.make_team(private=True)
        self.factory.add_account_to_team(account, team)
        token = account.create_oauth_token('token-name')
        url = '/api/1.0/accounts'

        response = self.client.get(url, data={'ws.op': 'team_memberships',
            'team_names': '["%s"]' % team.name},
            **self.auth_header(token, url))

        self.assertContains(response, team.name)

    def test_validate_email_bad_token(self):
        account = self.factory.make_account()
        token = account.create_oauth_token('token-name')
        url = '/api/1.0/accounts'

        response = self.client.get(url, data={'ws.op': 'validate_email',
            'email_token': 'ABCD'}, **self.auth_header(token, url))

        self.assertContains(response, 'Bad email token!')

    def test_validate_email_success(self):
        account = self.factory.make_account()
        email = self.factory.make_email_for_account(account,
            status=EmailStatus.NEW)
        authtoken = AuthTokenFactory().new_api_email_validation_token(
                account, email.email)

        token = account.create_oauth_token('token-name')
        url = '/api/1.0/accounts'

        response = self.client.get(url, data={'ws.op': 'validate_email',
            'email_token': authtoken.token}, **self.auth_header(token, url))

        self.assertContains(response, email.email)
        updatedemail = EmailAddress.objects.get(email=email.email)
        self.assertEqual(EmailStatus.VALIDATED, updatedemail.status)
