# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from datetime import datetime, timedelta

from django.conf import settings
from django.contrib.auth.models import User
from django.test import TestCase

from mock import patch
from oauth_backend.models import Consumer

from identityprovider.models.account import (Account, AccountPassword,
    AccountStatus, LPOpenIdIdentifier)
from identityprovider.models import account as a
from identityprovider.models.emailaddress import EmailAddress
from identityprovider.models.person import Person
from identityprovider.models.const import AccountCreationRationale, EmailStatus
from identityprovider.readonly import ReadOnlyManager
from identityprovider.utils import encrypt_launchpad_password, generate_salt
from identityprovider.tests.utils import SSOBaseTestCase


class MockBackend(object):

    def __init__(self, value):
        self.value = value

    def has_module_perms(self, account, app_label):
        return self.value

    def has_perm(self, account, perm):
        return self.value


class AccountTestCase(SSOBaseTestCase):

    fixtures = ["test"]

    def test_objects_get_by_email_fails(self):
        account = Account.objects.get_by_email('somethin@which.not.exists.com')

        self.assertTrue(account is None)

    def test_last_login_when_user_doesnt_exists(self):
        account = self.get_new_account()
        self.assertTrue(account.last_login is not None)

    def test_last_login_when_user_exists(self):
        account = self.get_existing_account()
        User.objects.get_or_create(username=account.openid_identifier)

        self.assertTrue(account.last_login is not None)

    def test_set_last_login_when_user_does_not_exists(self):
        account = self.get_new_account()

        # we use now() here because last_login maps to django's auth model
        # which uses localtime.
        account.last_login = datetime.now()

        self.assertTrue(account.last_login is not None)

    def test_set_last_login_when_readonly(self):
        readonly_manager = ReadOnlyManager()
        account = self.get_existing_account()
        account.last_login = last_login = datetime(2010, 01, 01)
        self.assertEqual(account.last_login, last_login)

        self.assertFalse(settings.READ_ONLY_MODE)
        readonly_manager.set_readonly()

        try:
            # we use now() here because last_login maps to django's auth model
            # which uses localtime.
            account.last_login = now = datetime.now()
            self.assertEqual(account.last_login, last_login)
            self.assertNotEqual(account.last_login, now)
        finally:
            readonly_manager.clear_readonly()
            self.assertFalse(settings.READ_ONLY_MODE)

    def test_set_last_login_when_no_preferredemail(self):
        account = self.get_new_account()
        account.emailaddress_set.all().update(status=EmailStatus.NEW)
        # we use now() here because last_login maps to django's auth model
        # which uses localtime.
        account.last_login = datetime.now()
        # Changed from assertRaises, as using it to test for database
        # exceptions causes transaction to be aborted and all tests following
        # it will fail. (It's only the case if the test case inherits
        # django.test.TestCase)
        self.assertEqual(
            0, User.objects.filter(username=account.openid_identifier).count())

    def test_is_active(self):
        account = self.get_new_account()
        account.status = AccountStatus.ACTIVE

        self.assertTrue(account.is_active)

    def test_is_staff(self):
        account = self.get_new_account()

        self.assertFalse(account.is_staff)

    def test_is_authenticated(self):
        account = self.get_new_account()

        self.assertTrue(account.is_authenticated)

    def test_is_superuser(self):
        account = self.get_new_account()

        self.assertFalse(account.is_superuser)

    def test_first_name(self):
        account = self.get_existing_account()
        self.assertEqual(account.first_name, "Sample")

    def test_full_name(self):
        account = self.get_existing_account()
        self.assertEqual(account.get_full_name(), "Sample Person")

    def test_has_module_perms_when_not_active(self):
        account = self.get_new_account()
        account.status = AccountStatus.NOACCOUNT

        self.assertFalse(account.has_module_perms('app'))

    def test_has_module_perms_returning_true(self):
        account, get_backends = self.prepare_perms(True)
        try:
            self.assertTrue(account.has_module_perms('app'))
        finally:
            a.get_backends = get_backends

    def test_has_module_perms_returning_false(self):
        account, get_backends = self.prepare_perms(False)
        try:
            self.assertFalse(account.has_module_perms('app'))
        finally:
            a.get_backends = get_backends

    def test_has_perm_returning_true(self):
        account, get_backends = self.prepare_perms(True)
        try:
            self.assertTrue(account.has_perm('can_change_account'))
        finally:
            a.get_backends = get_backends

    def test_has_perm_returning_false(self):
        account, get_backends = self.prepare_perms(False)
        try:
            self.assertFalse(account.has_perm('can_change_account'))
        finally:
            a.get_backends = get_backends

    def test_has_perm_when_account_is_not_active(self):
        account = self.get_new_account()
        account.status = AccountStatus.NOACCOUNT

        self.assertFalse(account.has_perm('can_change_account'))

    def prepare_perms(self, has_perm):
        account = self.get_new_account()
        account.status = AccountStatus.ACTIVE

        def mock_get_backends():
            return [MockBackend(has_perm)]

        get_backends = a.get_backends
        a.get_backends = mock_get_backends
        return account, get_backends

    def get_new_account(self):
        return Account.objects.create_account(
            'test account', 'test-email@example.com', 'password')

    def get_existing_account(self):
        return Account.objects.get_by_email('test@canonical.com')

    def test_new_account_does_not_have_preferred_email(self):
        account = self.create_account()
        self.assertTrue(account.preferredemail is None)

    def create_account(self):
        Account.objects.filter(displayname='Foo Bar').delete()
        account = Account.objects.create(
            creation_rationale=AccountCreationRationale.USER_CREATED,
            status=AccountStatus.ACTIVE,
            displayname='Foo Bar')
        return account

    def test_validated_email_address_is_preferred_email_by_default(self):
        account = self.create_account()
        email_address = account.emailaddress_set.create(
            email='foobar@canonical.com',
            status=EmailStatus.VALIDATED)
        self.assertEqual(account.preferredemail.email, 'foobar@canonical.com')

        email_address = EmailAddress.objects.get(pk=email_address.id)
        self.assertEqual(email_address.status, EmailStatus.PREFERRED)


    def test_person_openid_email_address_is_preferred_email(self):
        account = self.create_account()
        account.emailaddress_set.create(
            email='foobar@canonical.com',
            status=EmailStatus.VALIDATED)
        # account is set up with validated email

        # now set up person with preferred email, but no account
        person = Person.objects.create(
            name="Sample", 
            lp_account=100)

        # link person to account via open id
        LPOpenIdIdentifier.objects.create(
            identifier=account.openid_identifier,
            lp_account=person.lp_account)

        # add an address to the person, but NOT the account
        EmailAddress.objects.create(
            email='pref@bar.com',
            account=None,
            lp_person=person,
            status=EmailStatus.PREFERRED
        )

        self.assertEqual(account.preferredemail.email, 'pref@bar.com')

    
    def test_multi_persons_via_email_address_correct_preferred_email(self):
        """Test for advanced use case for
        https://bugs.launchpad.net/canonical-identity-provider/+bug/862944"""
        account = self.create_account()
        account.emailaddress_set.create(
            email='foobar@canonical.com',
            status=EmailStatus.VALIDATED)

        # now set up 2 persons
        person1 = Person.objects.create(
            name="Sample1",
            lp_account=100)
        person2 = Person.objects.create(
            name="Sample2",
            lp_account=101)

        # link person1 to account via open id
        LPOpenIdIdentifier.objects.create(
            identifier=account.openid_identifier,
            lp_account=person1.lp_account)

        # add an preferred address to person2, but NOT person1
        EmailAddress.objects.create(
            email='pref@bar.com',
            account=account,
            lp_person=person2,
            status=EmailStatus.PREFERRED
        )

        self.assertEqual(account.preferredemail.email, 'pref@bar.com')


    def test_verified_emails_has_preferred_first(self):
        account = Account.objects.get_by_email('test@canonical.com')
        emails = account.verified_emails()
        self.assertTrue(len(emails) > 1)
        self.assertEqual(emails[0].status, EmailStatus.PREFERRED)
        self.assertEqual(emails[1].status, EmailStatus.VALIDATED)

    def test_account_suspend_reset_password(self):
        account = self.create_account()
        AccountPassword.objects.create(account=account, password='valid')
        account.status = AccountStatus.SUSPENDED
        account.save()

        account_password = AccountPassword.objects.get(account=account)
        self.assertEqual(account_password.password, 'invalid')

    def test_save_when_readonly(self):
        readonly_manager = ReadOnlyManager()
        account = self.get_existing_account()
        status = account.status
        self.assertNotEqual(status, AccountStatus.SUSPENDED)

        self.assertFalse(settings.READ_ONLY_MODE)
        readonly_manager.set_readonly()

        try:
            account.status = AccountStatus.SUSPENDED
            account.save()

            # refresh account from db
            account = self.get_existing_account()
            self.assertEqual(account.status, status)
        finally:
            readonly_manager.clear_readonly()
            self.assertFalse(settings.READ_ONLY_MODE)

    def test_save_when_no_password(self):
        account = self.get_existing_account()
        # remove all passwords
        AccountPassword.objects.filter(account=account).delete()

        # trigger test
        account.status = AccountStatus.SUSPENDED
        account.save()

        password_count = AccountPassword.objects.filter(account=account).count()
        self.assertEqual(password_count, 0)

    def test_create_account_with_rationale(self):
        salt = generate_salt()
        account = Account.objects.create_account('displayname', 'username',
            'password', AccountCreationRationale.USER_CREATED, salt=salt)

        self.assertEqual(account.displayname, 'displayname')
        self.assertEqual(account.creation_rationale,
            AccountCreationRationale.USER_CREATED)
        self.assertEqual(account.status, AccountStatus.ACTIVE)

        emails = EmailAddress.objects.filter(account=account)
        self.assertEqual(emails.count(), 1)
        email = emails[0]
        self.assertEqual(email.email, 'username')
        self.assertEqual(email.status, EmailStatus.PREFERRED)

        passwords = AccountPassword.objects.filter(account=account)
        self.assertEqual(passwords.count(), 1)
        password = passwords[0]
        self.assertEqual(password.password,
            encrypt_launchpad_password('password', salt=salt))

    def test_create_account_with_no_rationale(self):
        salt = generate_salt()
        account = Account.objects.create_account('displayname', 'username',
            'password', salt=salt)

        self.assertEqual(account.displayname, 'displayname')
        self.assertEqual(account.creation_rationale,
            AccountCreationRationale.OWNER_CREATED_LAUNCHPAD)
        self.assertEqual(account.status, AccountStatus.ACTIVE)

        emails = EmailAddress.objects.filter(account=account)
        self.assertEqual(emails.count(), 1)
        email = emails[0]
        self.assertEqual(email.email, 'username')
        self.assertEqual(email.status, EmailStatus.PREFERRED)

        passwords = AccountPassword.objects.filter(account=account)
        self.assertEqual(passwords.count(), 1)
        password = passwords[0]
        self.assertEqual(password.password,
            encrypt_launchpad_password('password', salt=salt))

    def test_create_account_with_email_validated(self):
        salt = generate_salt()
        account = Account.objects.create_account('displayname', 'username',
            'password', salt=salt, email_validated=False)

        self.assertEqual(account.displayname, 'displayname')
        self.assertEqual(account.creation_rationale,
            AccountCreationRationale.OWNER_CREATED_LAUNCHPAD)
        self.assertEqual(account.status, AccountStatus.ACTIVE)

        emails = EmailAddress.objects.filter(account=account)
        self.assertEqual(emails.count(), 1)
        email = emails[0]
        self.assertEqual(email.email, 'username')
        self.assertEqual(email.status, EmailStatus.NEW)

        passwords = AccountPassword.objects.filter(account=account)
        self.assertEqual(passwords.count(), 1)
        password = passwords[0]
        self.assertEqual(password.password,
            encrypt_launchpad_password('password', salt=salt))

    def test_sites_with_active_sessions_when_returns_empty_result(self):
        account = Account.objects.create_account(
            'sites-account', 'sites@example.com', 'password')
        sites_count = account.sites_with_active_sessions().count()
        self.assertEqual(sites_count, 0)

    def test_sites_with_active_sessions_when_one_site_is_active(self):
        account = Account.objects.create_account(
            'sites-account', 'sites@example.com', 'password')

        account.openidrpsummary_set.create(
            date_last_used=datetime.utcnow(),
            trust_root='http://trust-root.example.com',
            openid_identifier='openid-identifier')

        account.openidrpsummary_set.create(
            date_last_used=datetime.utcnow() - timedelta(days=3),
            trust_root='http://trust-root-2.example.com')

        sites_count = account.sites_with_active_sessions().count()
        self.assertEqual(sites_count, 1)

    @patch('identityprovider.models.account.datetime')
    def test_sites_with_active_sessions_use_utc(self, mock_datetime):
        mock_datetime.utcnow.return_value = datetime.utcnow()
        mock_datetime.side_effect = lambda *args, **kw: datetime(*args, **kw)

        account = Account.objects.create_account(
            'sites-account', 'sites@example.com', 'password')
        sites_count = account.sites_with_active_sessions().count()

        self.assertEqual(sites_count, 0)
        self.assertEqual(mock_datetime.utcnow.called, True)


class AccountPasswordTestCase(TestCase):

    def test_unicode(self):
        account = Account(openid_identifier='oid', displayname='displayname')
        password = AccountPassword(account=account, password='password')
        self.assertEqual(unicode(password), u'Password for displayname')


class CreateOAuthTokenForAccountTestCase(SSOBaseTestCase):

    def setUp(self):
        self.account = Account.objects.create_account(
            'test', 'test-oauth@example.com', 'password')
        self.user = User.objects.create_user(
            self.account.openid_identifier,
            'test@example.com', 'password')

    def test_when_account_has_associated_consumer(self):
        consumer, _ = Consumer.objects.get_or_create(user=self.user)

        token = self.account.create_oauth_token('new-token')

        self.assertEquals(token.consumer.id, consumer.id)

    def test_when_account_has_no_associated_consumer(self):
        Consumer.objects.filter(user=self.user).delete()

        token = self.account.create_oauth_token('new-token')

        self.assertEquals(token.consumer.user.id, self.user.id)
