# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import random
from datetime import datetime
from django.conf import settings

from mock import (
    Mock,
    patch,
)

from identityprovider.branding import Launchpad
from identityprovider.models.account import Account
from identityprovider.models.authtoken import (
    AuthToken,
    AuthTokenFactory,
    create_unique_token_for_table,
    send_validation_email_request,
    valid_email,
)
from identityprovider.models.const import EmailStatus, LoginTokenType
from identityprovider.models import authtoken
from identityprovider.tests.utils import SSOBaseTestCase


class MockRenderToString(object):

    def __init__(self, return_value=None):
        self.template = None
        self.context = None
        self.return_value = None

    def __call__(self, template, context):
        self.template = template
        self.context = context
        return self.return_value


class AuthTokenTestCase(SSOBaseTestCase):

    fixtures = ["test"]

    def setUp(self):
        def mock_send_email(self, from_name, subject, message):
            self.message = message

        super(AuthTokenTestCase, self).setUp()

        # override method to capture message
        self.old_send_email = AuthToken._send_email
        AuthToken._send_email = mock_send_email

    def tearDown(self):
        AuthToken._send_email = self.old_send_email

        super(AuthTokenTestCase, self).tearDown()

    def test_sendEmailValidationRequest(self):
        token = self._get_new_email_validation_token()

        self.assertFalse(hasattr(token, 'message'))
        token.sendEmailValidationRequest()
        self.assertTrue(hasattr(token, 'message'))
        self.assertTrue(settings.SUPPORT_FORM_URL in token.message)

    def test_sendEmailValidationToken(self):
        token = self._get_new_email_validation_token()
        self.assertFalse(hasattr(token, 'message'))
        token.sendEmailValidationToken()
        self.assertTrue(hasattr(token, 'message'))
        self.assertTrue(token.token in token.message)

    def test_sendEmailValidationToken_uses_provided_template(self):
        token = self._get_new_email_validation_token()
        mock_rts, rts = self._setup_mock_rendert_to_string()

        token.sendEmailValidationToken('new-template.txt')

        self.assertTrue(mock_rts.template.endswith('new-template.txt'))

        self._teardown_mock_render_to_string(rts)

    def test_sendNewUserEmail_uses_provided_template(self):
        token = self._get_new_account_token()
        mock_rts, rts = self._setup_mock_rendert_to_string()

        token.sendNewUserEmail('new-template.txt')

        self.assertTrue(mock_rts.template.endswith('new-template.txt'))

        self._teardown_mock_render_to_string(rts)

    def test_sendNewUserEmail_default_template(self):
        token = self._get_new_account_token()
        mock_rts, rts = self._setup_mock_rendert_to_string()

        token.sendNewUserEmail()

        self.assertTrue(mock_rts.template.endswith('web-newuser.txt'))

        self._teardown_mock_render_to_string(rts)

    @patch('identityprovider.models.authtoken.current_brand', Launchpad())
    def test_sendNewUserEmail_default_template_launchpad_theme(self):
        token = self._get_new_account_token()
        # If at that point nothing blew up it means that the template is ok
        token.sendNewUserEmail()

    def _get_new_account_token(self):
        account = Account.objects.get_by_email('test@canonical.com')
        token = AuthTokenFactory().new(account, 'test@cannical.com',
                                       'test@canonical.com',
                                       LoginTokenType.NEWACCOUNT, None)
        return token

    def _teardown_mock_render_to_string(self, rts):
        authtoken.render_to_string = rts

    def _setup_mock_rendert_to_string(self):
        rts = authtoken.render_to_string
        mock_rts = MockRenderToString()
        authtoken.render_to_string = mock_rts
        return mock_rts, rts

    def _get_new_email_validation_token(self):
        account = Account.objects.get_by_email('test@canonical.com')
        token = AuthTokenFactory().new_api_email_validation_token(
            account, 'email@domain.com')
        return token

    @patch('identityprovider.models.authtoken.AuthTokenFactory.new')
    def test_new_api_email_validation_token_redirection_url(self, mock_new):
        requester = Mock()
        email = Mock()
        redirection_url = Mock()
        AuthTokenFactory().new_api_email_validation_token(requester, email,
            redirection_url=redirection_url)
        mock_new.assert_called_with(requester, None, email,
            LoginTokenType.VALIDATEEMAIL, redirection_url)

    @patch('identityprovider.models.authtoken.AuthTokenFactory.new')
    def test_new_api_email_validation_token_no_redirection_url(self,
            mock_new):
        requester = Mock()
        email = Mock()
        AuthTokenFactory().new_api_email_validation_token(requester, email)
        mock_new.assert_called_with(requester, None, email,
            LoginTokenType.VALIDATEEMAIL, None)

    def test_sendEmailValidationToken_default_template(self):
        token = self._get_new_email_validation_token()
        mock_rts, rts = self._setup_mock_rendert_to_string()

        token.sendEmailValidationToken()

        self.assertTrue(mock_rts.template.endswith(
            'email-validation-token.txt'))

        self._teardown_mock_render_to_string(rts)

    def test_sendNewUserEmail(self):
        token = self._get_new_email_validation_token()

        self.assertFalse(hasattr(token, 'message'))
        token.sendNewUserEmail()
        self.assertTrue(hasattr(token, 'message'))
        self.assertTrue(token.get_direct_absolute_url() in token.message)

    def test_send_validation_email_request_no_preferredemail(self):
        def mock_sendEmailValidationRequest(self):
            self.__class__.sendEmailValidationRequest_called = True
        _sendEmailValidationRequest = AuthToken.sendEmailValidationRequest
        AuthToken.sendEmailValidationRequest = mock_sendEmailValidationRequest
        AuthToken.sendEmailValidationRequest_called = False

        account = Account.objects.get_by_email('test@canonical.com')
        # remove preferredemail address
        for email in account.emailaddress_set.all():
            email.status = EmailStatus.NEW
            email.save()

        email = 'some@email.com'
        redirection_url = 'http://localhost/'
        send_validation_email_request(account, email, redirection_url)
        self.assertTrue(AuthToken.sendEmailValidationRequest_called)

        del AuthToken.sendEmailValidationRequest_called
        AuthToken.sendEmailValidationRequest = _sendEmailValidationRequest

    def test_consume(self):
        email = 'mark@example.com'
        token_type = LoginTokenType.NEWACCOUNT
        # need more than one token to fully test the method
        token = AuthToken(email=email, token_type=token_type, token='a')
        token.save()
        token2 = AuthToken(email=email, token_type=token_type, token='b')
        token2.save()
        self.assertEqual(token.date_consumed, None)
        self.assertEqual(token2.date_consumed, None)

        token.consume()

        tokens = AuthToken.objects.filter(email=token.email,
            token_type=token.token_type, requester=token.requester)
        self.assertEqual(tokens.count(), 2)
        self.assertNotEqual(tokens[0].date_consumed, None)
        self.assertNotEqual(tokens[1].date_consumed, None)

    @patch('identityprovider.models.authtoken.datetime')
    def test_consume_use_utc(self, mock_datetime):
        now = datetime.utcnow()
        mock_datetime.datetime.utcnow.return_value = now

        email = 'mark@example.com'
        token_type = LoginTokenType.NEWACCOUNT
        # need more than one token to fully test the method
        token = AuthToken(email=email, token_type=token_type, token='a')
        token.save()
        token2 = AuthToken(email=email, token_type=token_type, token='b')
        token2.save()
        self.assertEqual(token.date_consumed, None)
        self.assertEqual(token2.date_consumed, None)

        token.consume()

        tokens = AuthToken.objects.filter(email=token.email,
            token_type=token.token_type, requester=token.requester)

        self.assertEqual(tokens.count(), 2)
        self.assertEqual(tokens[0].date_consumed, now)
        self.assertEqual(tokens[1].date_consumed, now)
        self.assertEqual(mock_datetime.datetime.utcnow.called, True)

    def test_new_token_invalid_type(self):
        factory = AuthTokenFactory()
        good_types = [LoginTokenType.PASSWORDRECOVERY,
                      LoginTokenType.NEWACCOUNT,
                      LoginTokenType.VALIDATEEMAIL,
                      LoginTokenType.NEWPERSONLESSACCOUNT]
        bad_types = [t[0] for t in LoginTokenType._get_choices() \
                     if t[0] not in good_types]
        for token_type in bad_types:
            self.assertRaises(ValueError, factory.new, None, '',
                'mark@example.com', token_type, '')

    def test_create_unique_token_for_table(self):
        # initialize randomizer
        random.seed(42)

        # create a token
        token1 = create_unique_token_for_table(1, AuthToken, 'token')
        # use that token within an AuthToken
        AuthToken(email='mark@example.com',
            token_type=LoginTokenType.NEWACCOUNT, token=token1).save()

        # reinitialize randomizer to force same token
        random.seed(42)
        token2 = create_unique_token_for_table(1, AuthToken, 'token')
        self.assertNotEqual(token1, token2)

    def test_valid_email(self):
        emails = {'valid': ['kiko.async@hotmail.com', 'kiko+async@hotmail.com',
                            'kiko-async@hotmail.com', 'kiko_async@hotmail.com',
                            'kiko@async.com.br', 'kiko@canonical.com',
                            'kiko@UBUNTU.COM', 'i@tv', 'kiko@gnu.info',
                            'user@z.de', 'bob=dobbs@example.com',
                            'keith@risby-family.co.uk'],
                  'invalid': ['user@z..de', 'user@.z.de',
                              'keith@risby-family-.co.uk',
                              'keith@-risby-family.co.uk']}
        for email in emails['valid']:
            self.assertTrue(valid_email(email))
        for email in emails['invalid']:
            self.assertFalse(valid_email(email))
