# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import os
import stat
import shutil
import tempfile
import socket
import urllib2
from StringIO import StringIO

from django.conf import settings
from django.contrib.auth.models import User
from django.utils import simplejson

from identityprovider.readonly import ReadOnlyManager
from identityprovider import models
from identityprovider.backend.base import DatabaseError
from identityprovider.views.readonly import (_remote_req, get_server_atts,
    update_server)
from identityprovider.tests.utils import SSOBaseTestCase, skipOnSqlite


class InReadOnlyTestCase(SSOBaseTestCase):

    def setUp(self):
        self.rm = ReadOnlyManager()
        self.rm.set_readonly()

    def tearDown(self):
        self.rm.clear_readonly()


class ReadOnlyTest(InReadOnlyTestCase):
    fixtures = ["test"]

    @skipOnSqlite
    def test_invalid_insert(self):
        tests = [(models.OpenIDRPConfig, {'trust_root': 'foo',
                    'displayname': 'foo', 'description': 'foo'}),
                 (models.LibraryFileContent, {'filesize': 0,
                    'sha1': 'foo', 'md5': 'foo'}),
                 (models.Person, {'displayname': 'foo', 'name': 'foo'}),
                 (models.EmailAddress, {'email': 'foo', 'status': 0}),
                 ]
        for cls, args in tests:
            self.assertRaises(DatabaseError,
                cls.objects.create, **args)

    @skipOnSqlite
    def test_invalid_update(self):
        p = models.Person.objects.get(pk=1)
        p.displayname = 'Something Different'
        self.assertRaises(DatabaseError, p.save)

    def test_current_dbid_with_settings(self):
        old_DATABASE_ID = settings.DATABASE_ID
        settings.DATABASE_ID = 'mydb'

        self.assertEqual(self.rm.current_dbid(), 'mydb')

        settings.DATABASE_ID = old_DATABASE_ID

    def test_current_dbid_without_settings(self):
        old_DATABASE_ID = settings.DATABASE_ID
        del settings._wrapped.DATABASE_ID

        self.assertTrue(len(self.rm.connections) > 0)

        self.assertEqual(self.rm.current_dbid(),
                         self.rm.connections[0]['DATABASE_ID'])

        settings.DATABASE_ID = old_DATABASE_ID


class RemoteRequestTest(SSOBaseTestCase):
    msg = 'hello'
    host = 'myhost'
    scheme = 'https'
    vhost = 'http://foobar.baz'

    def setup_mock_urlopen(self, func):
        self.orig_urlopen = urllib2.urlopen
        urllib2.urlopen = func

    def restore_orig_urlopen(self):
        urllib2.urlopen = self.orig_urlopen

    def mock_urlopen(self, req, data=None):
        self.req = req
        self.assertEquals(5, socket.getdefaulttimeout())
        return StringIO(self.msg)

    def test_plain_remote_req(self):
        self.setup_mock_urlopen(self.mock_urlopen)
        server = {'host': self.host}
        self.assertEquals(self.msg, _remote_req(**server))
        self.assertEquals(self.host, self.req.get_host())
        self.restore_orig_urlopen()

    def test_https_remote_req(self):
        self.setup_mock_urlopen(self.mock_urlopen)
        server = {'host': self.host,
                  'scheme': self.scheme,
                 }
        self.assertEquals(self.msg, _remote_req(**server))
        self.assertEquals(self.scheme, self.req.get_type())
        self.assertEquals(self.host, self.req.get_host())
        self.restore_orig_urlopen()

    def test_vhost_remote_req(self):
        self.setup_mock_urlopen(self.mock_urlopen)
        server = {'host': self.host,
                  'scheme': self.scheme,
                  'virtual_host': self.vhost,
                 }
        self.assertEquals(self.msg, _remote_req(**server))
        headers = {'Host': self.vhost}
        self.assertEquals(headers, self.req.headers)
        self.assertEquals(self.scheme, self.req.get_type())
        self.assertEquals(self.host, self.req.get_host())
        self.restore_orig_urlopen()

    def test_remote_req_urllib2_error(self):
        def mock_urlopen(req, data=None):
            raise urllib2.URLError((-1, 'error'))

        self.setup_mock_urlopen(mock_urlopen)
        server = {'host': self.host}
        self.assertEquals(None, _remote_req(**server))
        self.restore_orig_urlopen()


class ReadOnlyBackUp(SSOBaseTestCase):
    """A base TestCase for backing up and restoring our readonly settings"""

    def setUp(self):
        self.rm = ReadOnlyManager()
        self.orig_readonly_secret = settings.READONLY_SECRET
        settings.READONLY_SECRET = 'testsecret'
        self.orig_db_connections = settings.DB_CONNECTIONS
        settings.DB_CONNECTIONS = [{'DATABASE_ID': 'master',
                                    'DATABASE_HOST': settings.DATABASE_HOST,
                                    'DATABASE_NAME': settings.DATABASE_NAME,
                                    'DATABASE_USER': settings.DATABASE_USER,
                                    'DATABASE_PORT': settings.DATABASE_PORT,
                                    'DATABASE_PASSWORD': settings.DATABASE_PASSWORD,
                                   },
                                   {'DATABASE_ID': 'slave',
                                    'DATABASE_HOST': settings.DATABASE_HOST,
                                    'DATABASE_NAME': settings.DATABASE_NAME,
                                    'DATABASE_USER': settings.DATABASE_USER,
                                    'DATABASE_PORT': settings.DATABASE_PORT,
                                    'DATABASE_PASSWORD': settings.DATABASE_PASSWORD,
                                   }]
        self._DBFAILOVER_FLAG_DIR = getattr(settings, 'DBFAILOVER_FLAG_DIR',
                                            None)
        settings.DBFAILOVER_FLAG_DIR = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(settings.DBFAILOVER_FLAG_DIR, True)
        settings.DBFAILOVER_FLAG_DIR = self._DBFAILOVER_FLAG_DIR
        settings.READONLY_SECRET = self.orig_readonly_secret
        settings.DB_CONNECTIONS = self.orig_db_connections
        self.rm.clear_readonly()
        self.rm.clear_failed('master')
        self.rm.set_db(settings.DB_CONNECTIONS[0])


class ReadOnlyManagerTestCase(ReadOnlyBackUp):
    def test_set_db(self):
        db = {
            'DATABASE_ID': 'foo',
            'DATABASE_HOST': 'foo_host',
            'DATABASE_PORT': 'foo_port',
            'DATABASE_NAME': 'foo_name',
            'DATABASE_USER': 'foo_user',
            'DATABASE_PASSWORD': 'foo_password',
        }
        self.rm.set_db(db)

        from django.db import connection
        for setting in db:
            self.assertEqual(getattr(settings, setting), db[setting])
            self.assertEqual(connection.settings_dict[setting], db[setting])


class ReadOnlyRecoveryTestCase(ReadOnlyBackUp):
    def setUp(self):
        super(ReadOnlyRecoveryTestCase, self).setUp()
        self.rm = ReadOnlyManager()
        self.rm.set_readonly(automatic=True)
        self.old_dbrecover_interval = settings.DBRECOVER_INTERVAL
        self.old_dbrecover_attempts = settings.DBRECOVER_ATTEMPTS
        self.old_dbrecover_multiplier = settings.DBRECOVER_MULTIPLIER
        settings.DBRECOVER_ATTEMPTS = 5
        settings.DBRECOVER_INTERVAL = 0 # Recover immediately
        settings.DBRECOVER_MULTIPLIER = 2

    def tearDown(self):
        if self.rm.is_failed(self.rm.current_dbid()):
            self.rm.clear_failed(self.rm.current_dbid())
        self.rm.clear_readonly()
        settings.DBRECOVER_INTERVAL = self.old_dbrecover_interval
        settings.DBRECOVER_ATTEMPTS = self.old_dbrecover_attempts
        settings.DBRECOVER_MULTIPLIER = self.old_dbrecover_multiplier
        super(ReadOnlyRecoveryTestCase, self).tearDown()

    def test_readonly_is_automatic(self):
        self.assertTrue(self.rm.current_readonly_is_automatic())

    def test_readonly_doesnt_override_auto(self):
        self.rm.mark_current_failed(automatic=True)
        self.assertTrue(self.rm.current_readonly_is_automatic())

    def test_recovers(self):
        self.rm.check_readonly()
        self.assertFalse(settings.READ_ONLY_MODE)

    def test_recovery_fail_increments_attempts(self):
        self.rm.mark_current_failed(automatic=True)
        self.assertEqual(2, self.rm.current_readonly_attempts())
        self.rm.ping_current_connection = self.mock_ping_current_connection
        self.rm.check_readonly()
        self.assertTrue(settings.READ_ONLY_MODE)
        self.assertEqual(3, self.rm.current_readonly_attempts())

    def test_recovery_stops_after_attempts(self):
        self.rm.ping_current_connection = self.mock_ping_current_connection
        for i in range(settings.DBRECOVER_ATTEMPTS):
            self.assertTrue(self.rm.next_recovery_due() is not None)
            self.rm.check_readonly()
        self.assertTrue(self.rm.next_recovery_due() is None)

    def test_leave_readonly_doesnt_enable_db_connection(self):
        self.rm.mark_current_failed(automatic=True)
        self.rm.clear_readonly()
        self.rm.check_readonly()
        self.assertTrue(settings.READ_ONLY_MODE)

    def test_readonly_doesnt_break_with_no_json(self):
        f = open(self.rm.readonly_marker_file, 'w')
        f.close() # Thus leaving the file empty
        rm = ReadOnlyManager()
        rm.check_readonly()
        self.assertFalse(settings.READ_ONLY_MODE)

    def test_readonly_doesnt_break_with_invalid_json(self):
        f = open(self.rm.readonly_marker_file, 'w')
        f.write('}so 0o; invalid"#')
        f.close()
        rm = ReadOnlyManager()
        rm.check_readonly()
        self.assertFalse(settings.READ_ONLY_MODE)

    def mock_ping_current_connection(self):
        return False


class ReadOnlyFlagFilesTest(ReadOnlyBackUp):
    def test_flag_files_in_right_directory(self):
        self.rm.set_readonly()
        flags = os.listdir(settings.DBFAILOVER_FLAG_DIR)
        self.assertTrue('db.readonly' in flags)

    def test_readonly_flag_files_have_right_mode(self):
        self.rm.set_readonly()
        mode = os.stat(self.rm.readonly_marker_file)[stat.ST_MODE]
        self.assertTrue(mode & stat.S_IRGRP)
        self.assertTrue(mode & stat.S_IWGRP)

    def test_dbfail_flag_files_have_right_mode(self):
        self.rm.mark_failed('foo')
        flagfile = self.rm.marker_file_pattern % 'foo'
        mode = os.stat(flagfile)[stat.ST_MODE]
        self.assertTrue(mode & stat.S_IRGRP)
        self.assertTrue(mode & stat.S_IWGRP)


class ReadOnlyDataTest(ReadOnlyBackUp):
    def test_readonly_returns_404_for_get(self):
        response = self.client.get('/readonlydata')
        self.assertEquals(404, response.status_code)

    def test_readonly_returns_404_without_readonly_secret(self):
        response = self.client.post('/readonlydata')
        self.assertEquals(404, response.status_code)

    def test_readonly_returns_404_with_bad_secret(self):
        data = {'secret': 'something different'}
        response = self.client.post('/readonlydata', data)
        self.assertEquals(404, response.status_code)

    def test_readonly_returns_200_with_right_secret(self):
        data = {'secret': settings.READONLY_SECRET}
        response = self.client.post('/readonlydata', data)
        self.assertEquals(200, response.status_code)

    def test_readonly_data(self):
        post = {'secret': settings.READONLY_SECRET}
        response = self.client.post('/readonlydata', post)
        data = simplejson.loads(response.content)
        self.assertTrue('readonly' in data)
        self.assertTrue('connections' in data)

    def test_readonly_set_readonly(self):
        post = {'secret': settings.READONLY_SECRET, 'action': 'set'}
        response = self.client.post('/readonlydata', post)
        data = simplejson.loads(response.content)
        self.assertTrue(data['readonly'])

    def test_readonly_clear_readonly(self):
        post = {'secret': settings.READONLY_SECRET, 'action': 'clear'}
        response = self.client.post('/readonlydata', post)
        data = simplejson.loads(response.content)
        self.assertFalse(data['readonly'])

    def test_readonly_disable_master(self):
        post = {'secret': settings.READONLY_SECRET,
                'action': 'disable',
                'conn': 'master'}
        response = self.client.post('/readonlydata', post)
        data = simplejson.loads(response.content)
        self.assertTrue(data['connections'][0]['failed'])
        self.assertTrue(data['readonly'])

    def test_readonly_reenable_master(self):
        post = {'secret': settings.READONLY_SECRET,
                'action': 'disable',
                'conn': 'master'}
        response = self.client.post('/readonlydata', post)
        post['action'] = 'enable'
        response = self.client.post('/readonlydata', post)
        data = simplejson.loads(response.content)
        self.assertFalse(data['connections'][0]['failed'])
        self.assertTrue(data['readonly'])


class ReadOnlyViewsTestCase(ReadOnlyBackUp):

    def login_with_staff(self):
        user = User.objects.create(username='admin')
        user.is_staff = True
        user.set_password('password')
        user.save()
        r = self.client.login(username='admin', password='password')
        self.assertTrue(r)

    def test_readonly_admin(self):
        self.login_with_staff()

        old_APP_SERVERS = settings.APP_SERVERS
        settings.APP_SERVERS = [{'SERVER_ID': 'localhost', 'SCHEME': 'http',
                                 'HOST': 'localhost', 'VIRTUAL_HOST': '',
                                 'PORT': '8000'}]
        expected = [{'appservers': [{'name': 'localhost', 'reachable': False}],
                     'admin_media_prefix': settings.ADMIN_MEDIA_PREFIX,
                     'clear_all_readonly': False,
                     'set_all_readonly': False}]
        r = self.client.get('/readonly')
        self.assertTemplateUsed(r, 'admin/readonly.html')
        for item in r.context:
            self.assertEqual(item.dicts, expected)

        settings.APP_SERVERS = old_APP_SERVERS

    def test_get_server_atts_server_unreachable(self):
        servers = [{'SERVER_ID': 'localhost', 'SCHEME': 'http',
                    'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'}]
        expected = {'appservers': [{'name': 'localhost', 'reachable': False}],
                    'admin_media_prefix': settings.ADMIN_MEDIA_PREFIX,
                    'clear_all_readonly': False,
                    'set_all_readonly': False}
        atts = get_server_atts(servers)
        self.assertEqual(atts, expected)

    def test_get_server_atts_data_error(self):
        def mock_loads(data):
            raise ValueError()
        old_loads = simplejson.loads
        simplejson.loads = mock_loads
        def mock_urlopen(req, data=None):
            data = {}
            return StringIO(simplejson.dumps(data))
        old_urlopen = urllib2.urlopen
        urllib2.urlopen = mock_urlopen

        servers = [{'SERVER_ID': 'localhost', 'SCHEME': 'http',
                    'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'}]
        expected = {'appservers': [{'name': 'localhost', 'reachable': False}],
                    'admin_media_prefix': settings.ADMIN_MEDIA_PREFIX,
                    'clear_all_readonly': False,
                    'set_all_readonly': True}
        atts = get_server_atts(servers)
        self.assertEqual(atts, expected)

        simplejson.loads = old_loads
        urllib2.urlopen = old_urlopen

    def test_get_server_atts_readonly(self):
        def mock_urlopen(req, data=None):
            data = {'readonly': True}
            return StringIO(simplejson.dumps(data))
        old_urlopen = urllib2.urlopen
        urllib2.urlopen = mock_urlopen

        servers = [{'SERVER_ID': 'localhost', 'SCHEME': 'http',
                    'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'}]
        expected = {'appservers': [{'name': 'localhost', 'reachable': True,
                                    'readonly': True}],
                    'admin_media_prefix': settings.ADMIN_MEDIA_PREFIX,
                    'clear_all_readonly': True,
                    'set_all_readonly': False}
        atts = get_server_atts(servers)
        self.assertEqual(atts, expected)

        urllib2.urlopen = old_urlopen

    def test_readonly_confirm_get(self):
        self.login_with_staff()

        r = self.client.get('/readonly/localhost/set')
        self.assertTemplateUsed(r, 'admin/readonly_confirm.html')
        self.assertEqual(r.context['appserver'], 'localhost')
        self.assertEqual(r.context['action'], 'set')
        self.assertEqual(r.context['conn'], None)

    def test_readonly_confirm_post(self):
        def mock_urlopen(req, data=None):
            self.reqs.append(req)
            return StringIO(simplejson.dumps({}))
        old_urlopen = urllib2.urlopen
        urllib2.urlopen = mock_urlopen

        self.disable_csrf()
        self.login_with_staff()

        self.reqs = []
        r = self.client.post('/readonly/localhost/set')
        data = map(lambda x: x.data, self.reqs)
        self.assertEqual(data, [
            "action=set&conn=None&secret=%s" % settings.READONLY_SECRET
        ])

        self.assertRedirects(r, '/readonly')

        self.reset_csrf()

        urllib2.urlopen = old_urlopen

    def test_update_server_all_appservers(self):
        def mock_urlopen(req, data=None):
            self.reqs.append(req)
            return StringIO(simplejson.dumps({}))
        old_urlopen = urllib2.urlopen
        urllib2.urlopen = mock_urlopen

        old_APP_SERVERS = settings.APP_SERVERS
        settings.APP_SERVERS = servers = [
            {'SERVER_ID': 'localhost', 'SCHEME': 'http',
             'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'},
            {'SERVER_ID': 'otherhost', 'SCHEME': 'http',
             'HOST': 'otherhost', 'VIRTUAL_HOST': '', 'PORT': '8000'},
        ]

        self.reqs = []
        update_server('set')
        data = map(lambda x: x.data, self.reqs)
        self.assertEqual(data, [
            "action=set&conn=None&secret=%s" % settings.READONLY_SECRET,
            "action=set&conn=None&secret=%s" % settings.READONLY_SECRET
        ])

        settings.APP_SERVERS = old_APP_SERVERS
        urllib2.urlopen = old_urlopen

    def test_update_server_one_appserver(self):
        def mock_urlopen(req, data=None):
            self.reqs.append(req)
            return StringIO(simplejson.dumps({}))
        old_urlopen = urllib2.urlopen
        urllib2.urlopen = mock_urlopen

        old_APP_SERVERS = settings.APP_SERVERS
        settings.APP_SERVERS = servers = [
            {'SERVER_ID': 'localhost', 'SCHEME': 'http',
             'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'},
            {'SERVER_ID': 'otherhost', 'SCHEME': 'http',
             'HOST': 'otherhost', 'VIRTUAL_HOST': '', 'PORT': '8000'},
        ]

        self.reqs = []
        update_server('set', 'localhost')
        data = map(lambda x: x.data, self.reqs)
        self.assertEqual(data, [
            "action=set&conn=None&secret=%s" % settings.READONLY_SECRET
        ])

        settings.APP_SERVERS = old_APP_SERVERS
        urllib2.urlopen = old_urlopen

    def test_update_server_one_connection(self):
        def mock_urlopen(req, data=None):
            self.reqs.append(req)
            return StringIO(simplejson.dumps({}))
        old_urlopen = urllib2.urlopen
        urllib2.urlopen = mock_urlopen

        old_APP_SERVERS = settings.APP_SERVERS
        settings.APP_SERVERS = servers = [
            {'SERVER_ID': 'localhost', 'SCHEME': 'http',
             'HOST': 'localhost', 'VIRTUAL_HOST': '', 'PORT': '8000'},
        ]

        self.reqs = []
        update_server('set', 'localhost', 'master')
        data = map(lambda x: x.data, self.reqs)
        self.assertEqual(data, [
            "action=set&conn=master&secret=%s" % settings.READONLY_SECRET,
        ])

        settings.APP_SERVERS = old_APP_SERVERS
        urllib2.urlopen = old_urlopen


class ReadOnlyViews(InReadOnlyTestCase):

    def test_new_account_is_rendered_using_read_only_template(self):
        settings.READ_ONLY_MODE = True
        r = self.client.get('/+new_account')
        self.assertTemplateUsed(r, 'readonly.html')
