# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import urllib2

from django.test import TestCase
from django.utils.translation import ugettext as _

from identityprovider.models.person import Person
from identityprovider.models.account import LPOpenIdIdentifier
from identityprovider.models.emailaddress import EmailAddress
from identityprovider.utils import (canonical_url,
    get_person_and_account_by_email, password_policy_compliant,
    polite_form_errors, validate_launchpad_password,
    PersonAndAccountNotFoundException, http_request_with_timeout,
    format_address)


from identityprovider.tests.utils import SSOBaseTestCase, LPAccountTestCase, skipOnSqlite


class CanonicalUrlTestCase(SSOBaseTestCase):

    def test_when_object_is_not_person(self):
        self.assertTrue(canonical_url(None) is None)

    def test_when_view_name_is_none(self):
        url = canonical_url(Person(name="test"), view_name=None)
        self.assertEquals(url, "https://launchpad.net/~test")

    def test_when_view_name_is_not_none(self):
        url = canonical_url(Person(name="test"), view_name="TEST")
        self.assertEquals(url, "https://launchpad.net/~test/TEST")


class GetPersonAndAccountByEmailTestCase(SSOBaseTestCase):

    fixtures = ["test"]

    def test_account_and_person_is_none(self):
        email = EmailAddress.objects.get(email__iexact="mark@example.com")
        email.account = None
        email.lp_person = None
        email.save()

        self.assertRaises(PersonAndAccountNotFoundException,
                          get_person_and_account_by_email,
                          "mark@example.com")


@skipOnSqlite
class LPGetPersonAndAccountByEmailTestCase(LPAccountTestCase):

    fixtures = ["test"]

    def test_person_does_not_exist(self):
        LPOpenIdIdentifier.objects.all().delete()

        # test method
        person, account = get_person_and_account_by_email('mark@example.com')
        self.assertEquals(None, person)


class PasswordPolicyCompliantTestCase(SSOBaseTestCase):

    def test_when_password_is_to_short(self):
        self.assertFalse(password_policy_compliant('a'))


class PoliteFormErrorsTestCase(SSOBaseTestCase):

    def test_polite_errors(self):
        text = _("Enter a valid e-mail address.")
        errors = {'email': [text]}

        polite_form_errors(errors)

        self.assertTrue(errors['email'][0] != text)


class ValidateLaunchpadPasswordTestCase(SSOBaseTestCase):

    def test_when_there_is_binascii_error_is_raised(self):
        self.assertFalse(validate_launchpad_password('a', 'b'))


class HttpRequestWithTimeoutTestCase(TestCase):

    def setUp(self):
        self.exception = None
        self.output = None
        self.headers = None

        class MockDatafile(object):

            def __init__(self, output, headers):
                self.output = output
                self.headers = headers

            def read(self):
                return self.output

            def info(self):
                return self.headers

        def mock_urlopen(request):
            self.request = request
            if self.exception:
                raise urllib2.URLError("arg")
            return MockDatafile(self.output, self.headers)

        self.urlopen = urllib2.urlopen
        urllib2.urlopen = mock_urlopen

    def tearDown(self):
        urllib2.urlopen = self.urlopen

    def test_without_data_and_successful_outcome(self):
        self.output = "output"
        data, headers = http_request_with_timeout("http://example.com")

        self.assertEquals(data, "output")

    def test_without_data_and_failing(self):
        self.exception = True
        data, headers = http_request_with_timeout("http://example.com")
        self.assertEquals(data, None)

    def test_with_data_and_successful_outcome(self):
        self.output = "output"
        data, headers = http_request_with_timeout("http://example.com",
                                                  {"i1": "v1", "i2": "v2"})
        self.assertTrue("&" in self.request.data)


class FormatAddressTestCase(TestCase):
    def test_address_and_name(self):
        result = format_address('foo@bar.com', 'Name')
        self.assertEqual(result, 'Name <foo@bar.com>')

    def test_address_empty_name(self):
        result = format_address('foo@bar.com', '')
        self.assertEqual(result, 'foo@bar.com')

    def test_address_no_name(self):
        result = format_address('foo@bar.com')
        self.assertEqual(result, 'foo@bar.com')

    def test_unicode(self):
        result = format_address('foo.bar@canonical.com', u'F\xf4\xf4 Bar')
        self.assertEqual(result, '=?utf-8?b?RsO0w7QgQmFy?= <foo.bar@canonical.com>')

    def test_quoting(self):
        result = format_address('foo.bar@canonical.com', 'Foo [Baz] Bar')
        self.assertEqual(result, '"Foo \\[Baz\\] Bar" <foo.bar@canonical.com>')

    def test_long_name(self):
        result = format_address('long.name@example.com', 'a '*100)
        self.assertFalse('\n' in result)
