# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import logging

from django.core import mail

from oauth_backend.models import Token

from identityprovider.models import Account, EmailAddress
from identityprovider.models.const import AccountStatus, EmailStatus
from identityprovider.utils import validate_launchpad_password

from identityprovider.tests.utils import SSOBaseTestCase, AuthenticatedTestCase


class AccountViewsUnauthenticatedTestCase(SSOBaseTestCase):

    def test_index_unauthenticated(self):
        r = self.client.get('/+edit')
        self.assertEqual(r.status_code, 200)

    def test_index_xrds(self):
        accept = "application/xrds+xml; q=1, text/html; q=0.9"
        response = self.client.get('/', HTTP_ACCEPT=accept)
        self.assertTrue('Accept' in response['Vary'].split(','))
        self.assertFalse('X-XRDS-Location' in response)
        self.assertTrue('application/xrds+xml' in response['Content-Type'])

    def test_index_html(self):
        response = self.client.get('/')
        self.assertTrue('Accept' in response['Vary'].split(','))
        self.assertTrue('X-XRDS-Location' in response)
        self.assertTrue('text/html' in response['Content-Type'])


class AccountViewsTest(AuthenticatedTestCase):

    def setUp(self):
        super(AccountViewsTest, self).setUp(disableCSRF=True)
        logging.disable(logging.WARNING)

    def test_index_authenticated(self):
        r = self.client.get('/+edit')
        self.assertTemplateUsed(r, 'account/edit.html')

    def test_index_edit_displayname(self):
        r = self.client.post('/+edit', {'displayname': "New Display Name",
                                        'preferred_email': '16'})
        self.assertEqual(r.status_code, 302)

    def test_index_edit_password(self):
        r = self.client.post('/+edit', {'displayname': "New Display Name",
                                        'preferred_email': '16',
                                        'password': 'new-Password',
                                        'passwordconfirm': 'new-Password'})
        account = Account.objects.get(displayname='New Display Name')
        self.assertEqual(r.status_code, 302)
        self.assert_(validate_launchpad_password(
            'new-Password', account.accountpassword.password))

    def test_index_clears_up_message_from_session(self):
        self.client.session['message'] = 'test message'
        self.client.get('/+edit')

        self.assert_('message' not in self.client.session)

    def test_verify_email_already_verified(self):
        r = self.client.get('/+verify-email', {'id': 1})
        self.assertEqual(r.status_code, 404)

    def test_verify_email_other_users_email(self):
        other_account = Account.objects.get(pk=1)
        other_email = EmailAddress.objects.create(
            email='foo@bar.com',
            account=other_account,
            status=EmailStatus.NEW)
        r = self.client.get('/+verify-email', {'id': other_email.id})
        self.assertEqual(r.status_code, 404)

    def test_verify_email_success(self):
        account = Account.objects.get_by_email('test@canonical.com')
        email = EmailAddress.objects.create(
            email='footest@bar.com',
            account=account,
            status=EmailStatus.NEW)
        r = self.client.get('/+verify-email', {'id': email.id})
        self.assertEqual(r.status_code, 200)
        self.assertEqual(len(mail.outbox), 1)
        # Erasing emails. Necessary to make other tests pass.
        mail.outbox = []

    def test_new_email_get(self):
        r = self.client.get('/+new-email')
        self.assertTemplateUsed(r, 'account/new_email.html')

    def test_new_email_post(self):
        r = self.client.post('/+new-email',
                             {'newemail': "very-new-email@example.com"})
        self.assertEqual(r.status_code, 200)
        mail.outbox = []

    def test_new_email_post_with_token(self):
        self.client.post('/thisissuperrando/+new-email',
                         {'newemail': "very-new-email@example.com"})
        self.assertEqual(len(mail.outbox), 1)
        mail.outbox = []


class AccountDeletionViewTestCase(AuthenticatedTestCase):

    def setUp(self):
        super(AccountDeletionViewTestCase, self).setUp(disableCSRF=True)

        email = EmailAddress.objects.get(email__iexact=self.login_email)
        self.account = email.account

    def test_delete_email_get_form(self):
        email = self.account.emailaddress_set.create(
            email='test@test.com', status=EmailStatus.NEW)
        r = self.client.get('/+remove-email', {'id': email.id})
        self.assertTemplateUsed(r, 'account/delete_email.html')

    def test_delete_email_when_submiting_form(self):
        email = self.account.emailaddress_set.create(
            email='test@test.com', status=EmailStatus.NEW)
        # Due to limitation in Django 1.0 you can't pass directly GET arguments
        # to post method, that's why you need to use QUERY_STRING, this is
        # unnecessary in Django 1.1
        r = self.client.post('/+remove-email', QUERY_STRING='id=%s' % email.id)
        self.assertRedirects(r, '/+emails')


class AccountDeactivationViewTestCase(AuthenticatedTestCase):
    def setUp(self):
        super(AccountDeactivationViewTestCase, self).setUp(disableCSRF=True)
        self._refresh_account()

    def _refresh_account(self):
        email = EmailAddress.objects.get(email__iexact=self.login_email)
        self.account = email.account

    def test_deactivate_account(self):
        # make sure we have a preferredemail
        email = self.account.emailaddress_set.all()[0]
        self.account.preferredemail = email

        # deactivate account
        r = self.client.post('/+deactivate')
        self.assertRedirects(r, '/+deactivated')

        # re-request account object from db
        self._refresh_account()

        # test account has been deactivated
        self.assertEqual(AccountStatus.DEACTIVATED, self.account.status)
        self.assertEqual(None, self.account.preferredemail)
        for email in self.account.emailaddress_set.all():
            self.assertEqual(EmailStatus.NEW, email.status)

    def test_deactivate_account_no_preferredemail(self):
        # unset preferredemail
        for email in self.account.emailaddress_set.all():
            email.status = EmailStatus.NEW
            email.save()
        self.assertEqual(self.account.preferredemail, None)

        # deactivate account
        r = self.client.post('/+deactivate')
        self.assertRedirects(r, '/+deactivated')

        # re-request account object from db
        self._refresh_account()

        # test preferredemail status
        self.assertEqual(self.account.preferredemail, None)

    def test_deactivate_with_token(self):
        token = 'a' * 16
        # This strange dance with session is necessary to overcome way in which
        # test.Client returns session (recreating it on every Client.session
        # access)
        session = self.client.session
        session[token] = 'raw_orequest content'
        session.save()

        # deactivate account
        r = self.client.post('/%s/+deactivate' % token)
        self.assertRedirects(r, '/+deactivated')

        # test token is preserved
        self.assertEqual(self.client.session.get(token),
                          'raw_orequest content')


class ApplicationsTestCase(SSOBaseTestCase):

    fixtures = ["test"]

    def setUp(self):
        self.client.login(username="mark@example.com", password="test")
        self.disable_csrf()

    def tearDown(self):
        self.reset_csrf()

    def test_applications_page_is_rendered_using_right_template(self):
        r = self.client.get('/+applications')
        self.assertTemplateUsed(r, 'account/applications.html')

    def test_account_without_applications_does_not_renders_token_table(self):
        r = self.client.get('/+applications')
        self.assertNotContains(r, '<table')

    def test_revoking_token_removes_it_from_being_displayed(self):
        account = Account.objects.get_by_email("mark@example.com")
        token_1 = account.create_oauth_token("Token-1")
        token_2 = account.create_oauth_token("Token-2")

        r = self.client.get('/+applications')

        self.assertContains(r, "Token-1")
        self.assertContains(r, "Token-2")

        self.client.post('/+applications', {'token_id': token_2.token})

        r = self.client.get('/+applications')

        self.assertNotContains(r, token_2.token)

        token_1.delete()

    def test_revoking_token_which_does_not_belogs_to_an_account(self):
        account = Account.objects.get_by_email("test@canonical.com")
        token = account.create_oauth_token("Token")

        r = self.client.post('/+applications', {'token_id': token.token})

        self.assertRedirects(r, '/+applications')
        self.assertEqual(Token.objects.filter(token=token.token).count(), 1)
