# encoding: utf-8
# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import unittest

from openid import fetchers
from openid.consumer import consumer
from openid.consumer.discover import OpenIDServiceEndpoint, OPENID_2_0_TYPE
from openid.message import (OPENID1_URL_LIMIT, IDENTIFIER_SELECT)
from openid.yadis.constants import YADIS_CONTENT_TYPE
from urllib import quote_plus


# need to override fetcher before consumer module gets imported
class MockFetcher(fetchers.Urllib2Fetcher):

    called = False

    def fetch(self, url, body=None, headers=None):
        MockFetcher.called = True
        return super(MockFetcher, self).fetch(url, body, headers)

fetchers.Urllib2Fetcher = MockFetcher
from identityprovider.tests.utils import SSOBaseTestCase
from identityprovider.views import server
from identityprovider.views.consumer import (SREG_OPTIONAL,
    SREG_REQUIRED, getBaseURL, normalDict, renderIndexPage, startOpenID)


class DummyRequest(object):
    def __init__(self, META):
        self.META = META


class DummyDjangoRequest(object):
    def __init__(self):
        self.META = {
            'HTTP_HOST': "localhost",
            'SCRIPT_NAME': "http://localhost",
            'SERVER_PROTOCOL': "http",
        }
        self.POST = {
            'openid_identifier': "http://localhost/+id/abcd123",
        }
        self.session = {}


class FetcherTestCase(unittest.TestCase):
    def test_check_custom_fetcher_called(self):
        startOpenID(DummyDjangoRequest())
        self.assertTrue(MockFetcher.called)


class GetBaseURLTestCase(unittest.TestCase):
    def setUp(self):
        self.http_consumer_url = "http://localhost/"
        self.https_consumer_url = "https://localhost/"
        self.http_proto_req = DummyRequest({'HTTP_HOST': 'localhost',
                                            'SERVER_PROTOCOL': 'HTTP/1.1'})
        self.https_proto_req = DummyRequest({'HTTP_HOST': 'localhost',
                                             'SERVER_PROTOCOL': 'HTTPS/1.1'})
        self.https_req = DummyRequest({'HTTP_HOST': 'localhost',
                                       'SERVER_PROTOCOL': 'HTTP/1.1',
                                       'HTTPS': 1})

    def test_http_consumer(self):
        url = getBaseURL(self.http_proto_req)
        self.assertEqual(url, self.http_consumer_url)

    def test_https_consumer(self):
        url = getBaseURL(self.https_proto_req)
        self.assertEqual(url, self.https_consumer_url)

        url = getBaseURL(self.https_req)
        self.assertEqual(url, self.https_consumer_url)

    def test_consumer_on_port(self):
        req = DummyRequest({'HTTP_HOST': 'localhost',
                            'SERVER_PORT': '81',
                            'SERVER_PROTOCOL': 'HTTP/1.1'})
        base_url = 'http://localhost:81/'
        url = getBaseURL(req)
        self.assertEqual(url, base_url)


class RenderIndexPageTestCase(unittest.TestCase):
    def test_renderIndexPage_utf8(self):
        fullname = u'Dūmmẏ Ũsèṙ'.encode('utf-8')
        request = DummyRequest({
            'HTTP_HOST': 'localhost',
            'SERVER_PROTOCOL': 'HTTP/1.1'})

        template_args = {'sreg': [('fullname', fullname)],
                         'url': 'http://localhost/'}
        response = renderIndexPage(request, **template_args)
        headers = dict(response.items())
        self.assertEqual(headers['Content-Type'], 'text/html; charset=utf-8')
        self.assertTrue(fullname in response.content)


class ConsumerTestCase(SSOBaseTestCase):

    def setUp(self):
        super(ConsumerTestCase, self).setUp()

        self.req = DummyDjangoRequest()
        self.endpoint = OpenIDServiceEndpoint()
        self.endpoint.claimed_id = 'oid'
        self.endpoint.server_url = 'http://localhost/'

        class MockConsumer(consumer.Consumer):
            def begin(this, url):
                auth_request = consumer.AuthRequest(self.endpoint, None)
                return auth_request
        self.old_consumer = consumer.Consumer
        consumer.Consumer = MockConsumer

    def tearDown(self):
        consumer.Consumer = self.old_consumer

        super(ConsumerTestCase, self).tearDown()

    def test_normalDict(self):
        req = DummyDjangoRequest()
        self.assertEqual(normalDict(req.POST),
            {'openid_identifier': "http://localhost/+id/abcd123"})

    def test_startOpenID_get(self):
        r = self.client.get('/consumer/', **self.req.META)
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'consumer/index.html')

    def test_startOpenID_discovery_error(self):
        # we don't want the mock consumer for this test
        consumer.Consumer = self.old_consumer

        r = self.client.post('/consumer/', self.req.POST, **self.req.META)
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'consumer/index.html')
        self.assertTrue(r.context['error'].startswith(
            'OpenID discovery error'))

    def test_startOpenID_immediate(self):
        self.req.POST.update({'mode': 'immediate'})
        r = self.client.post('/consumer/', self.req.POST, **self.req.META)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.mode'], 'checkid_immediate')
        self.assertEqual(query['openid.return_to'], quote_plus('http://localhost/consumer/finish/'))

    def test_startOpenID_setup(self):
        self.req.POST.update({'mode': 'setup'})
        r = self.client.post('/consumer/', self.req.POST, **self.req.META)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.mode'], 'checkid_setup')
        self.assertEqual(query['openid.return_to'], quote_plus('http://localhost/consumer/finish/'))

    def test_startOpenID_sreg(self):
        self.req.POST.update({'mode': 'setup',
                              'sreg': 'yes',
                              'sreg_fullname': SREG_REQUIRED,
                              'sreg_email': SREG_REQUIRED,
                              'sreg_nickname': SREG_OPTIONAL})
        r = self.client.post('/consumer/', self.req.POST, **self.req.META)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.sreg.required'],
                         quote_plus('fullname,email'))
        self.assertEqual(query['openid.sreg.optional'], 'nickname')

    def test_startOpenID_teams(self):
        request_teams = 'team1,team2,team3'
        self.req.POST.update({
            'mode': 'setup',
            'teams': 'yes',
            'request_teams': request_teams
        })
        r = self.client.post('/consumer/', self.req.POST, **self.req.META)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.lp.query_membership'],
                         quote_plus(request_teams))

    def test_startOpenID_should_redirect(self):
        self.req.POST['mode'] = 'setup'
        r = self.client.post('/consumer/', self.req.POST, **self.req.META)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.trust_root'],
                         quote_plus('http://localhost/consumer/'))
        self.assertEqual(query['openid.return_to'],
                         quote_plus('http://localhost/consumer/finish/'))

    def test_startOpenID_render_template(self):
        # replace consumer with OpenID 2.0 compliant one
        class MockConsumer(consumer.Consumer):
            def begin(self, url):
                endpoint = OpenIDServiceEndpoint()
                endpoint.claimed_id = 'oid'
                endpoint.server_url = 'http://localhost/'
                endpoint.type_uris = [OPENID_2_0_TYPE]
                auth_request = consumer.AuthRequest(endpoint, None)
                return auth_request
        consumer.Consumer = MockConsumer

        self.req.POST['mode'] = 'setup'
        r = self.client.post('/consumer/', self.req.POST, **self.req.META)
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'consumer/request_form.html')
        self.assertTrue('openid_message'in r.context['html'])

    def test_finishOpenID_get(self):
        r = self.client.get('/consumer/finish/', **self.req.META)
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'consumer/index.html')

    def test_finishOpenID_post(self):
        r = self.client.post('/consumer/finish/', **self.req.META)
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'consumer/index.html')

    def test_finishOpenID_consumer_success(self):
        def mock_complete(this, request_args, return_to):
            request = {'openid.mode': 'checkid_setup',
                       'openid.trust_root': 'http://localhost/',
                       'openid.return_to': 'http://localhost/',
                       'openid.identity': IDENTIFIER_SELECT,
                       'openid.sreg.nickname': 'mynickname',
                       'openid.lp.is_member': 'myteam'}
            openid_server = server._get_openid_server()
            orequest = openid_server.decodeRequest(request)
            response = consumer.SuccessResponse(
                self.endpoint, orequest.message,
                signed_fields=['openid.sreg.nickname',
                               'openid.lp.is_member'])
            return response
        old_complete = consumer.Consumer.complete
        consumer.Consumer.complete = mock_complete

        r = self.client.post('/consumer/finish/', self.req.POST,
                             **self.req.META)
        self.assertEqual(r.context['url'], 'oid')
        self.assertEqual(r.context['sreg'], [('nickname', 'mynickname')])
        self.assertEqual(r.context['teams'], ['myteam'])

        consumer.Consumer.complete = old_complete

    def test_finishOpenID_consumer_cancel(self):
        def mock_complete(this, request_args, return_to):
            response = consumer.CancelResponse(self.endpoint)
            return response
        old_complete = consumer.Consumer.complete
        consumer.Consumer.complete = mock_complete

        r = self.client.post('/consumer/finish/', self.req.POST,
                             **self.req.META)
        self.assertEqual(r.context['message'],
                         'OpenID authentication cancelled.')

        consumer.Consumer.complete = old_complete

    def test_finishOpenID_consumer_failure(self):
        def mock_complete(this, request_args, return_to):
            response = consumer.FailureResponse(self.endpoint,
                                                message='some error')
            return response
        old_complete = consumer.Consumer.complete
        consumer.Consumer.complete = mock_complete

        r = self.client.post('/consumer/finish/', self.req.POST,
                             **self.req.META)
        self.assertEqual(r.context['error'],
                         'OpenID authentication failed.')
        self.assertEqual(r.context['failure_reason'], 'some error')

        consumer.Consumer.complete = old_complete

    def test_rpXRDS(self):
        r = self.client.get('/consumer/xrds/', **self.req.META)
        self.assertTemplateUsed(r, 'openidapplication-xrds.xml')
        self.assertEqual(r['Content-Type'], YADIS_CONTENT_TYPE)

    def test_long_query_string(self):
        self.req.POST['mode'] = 'setup'
        self.req.POST['forcelongurl'] = '1'
        r = self.client.post('/consumer/', self.req.POST, **self.req.META)
        query = self.get_query(r)
        self.assertTrue(len(query['openid.return_to']) > OPENID1_URL_LIMIT)
