"""
Tests for SAML2 Processors.

A lot of this is borrowed directly from the saml2idp tests.
"""
import base64

from BeautifulSoup import BeautifulSoup
from django.conf import settings
from saml2idp import saml2idp_settings
from saml2idp.tests import salesforce, google_apps

from identityprovider.models import EmailAddress
from identityprovider.tests.utils import add_user_to_team, AuthenticatedTestCase


SAML_TEAM = 'saml2team' # Name of team enabled for SAML 2.0

def add_test_user_to_saml_team(address):
    """
    Adds test user with email address to the SAML_TEAM.
    """
    email = EmailAddress.objects.get(email=address)
    add_user_to_team(email.account, SAML_TEAM)


def find_cert_path():
    """
    Find the right path to certificates
    """
    from identityprovider import tests
    return tests.__path__[0] + '/cert/'

def change_preferred_email(login_email, alternate_email):
    """
    Finds the user identified by login_email and sets his preferredemail to
    alternate_email.
    """
    email = EmailAddress.objects.get(email=login_email)
    account = email.account
    target = EmailAddress.objects.get(email=alternate_email)
    account.preferredemail = target


class GoogleAppsAssertionTestCase(AuthenticatedTestCase):
    """
    Test SAML Assertions for a Google Apps Service Point.

    Sub-classes with identical behavior, but different ACS_URL or REQUEST_DATA
    simply specify their own attributes on the class.
    """
    ACS_URL = google_apps.GOOGLE_APPS_ACS
    REQUEST_DATA = google_apps.REQUEST_DATA

    def setUp(self, *args, **kwargs):
        super(GoogleAppsAssertionTestCase, self).setUp(disableCSRF=True, *args, **kwargs)
        # Save settings.
        self._old_cert_file = saml2idp_settings.SAML2IDP_CERTIFICATE_FILE
        self._old_priv_file = saml2idp_settings.SAML2IDP_PRIVATE_KEY_FILE
        self._old_signing = saml2idp_settings.SAML2IDP_SIGNING
        self._old_acs = saml2idp_settings.SAML2IDP_VALID_ACS
        self._old_teams = settings.SAML2_TEAMS
        path = find_cert_path()
        cert_file = path + 'certificate.pem'
        priv_file = path + 'private-key.pem'

        saml2idp_settings.SAML2IDP_CERTIFICATE_FILE = cert_file
        saml2idp_settings.SAML2IDP_PRIVATE_KEY_FILE = priv_file
        saml2idp_settings.SAML2IDP_SIGNING = True
        settings.SAML2_TEAMS = [ SAML_TEAM ]
        saml2idp_settings.SAML2IDP_VALID_ACS = [ self.ACS_URL ]

    def tearDown(self, *args, **kwargs):
        super(GoogleAppsAssertionTestCase, self).tearDown(*args, **kwargs)
        # Restore settings.
        saml2idp_settings.SAML2IDP_CERTIFICATE_FILE = self._old_cert_file
        saml2idp_settings.SAML2IDP_PRIVATE_KEY_FILE = self._old_priv_file
        saml2idp_settings.SAML2IDP_SIGNING = self._old_signing
        saml2idp_settings.SAML2IDP_VALID_ACS = self._old_acs
        settings.SAML2_TEAMS = self._old_teams

    def test_authnrequest_handled(self):
        """ AuthnRequest isn't handled by a Processor. """
        # Arrange/Act:
        response = self.client.get('/+saml', data=self.REQUEST_DATA, follow=False)

        # Assert:
        self.assertEqual(response.status_code, 302)

    def test_user_logged_in(self):
        """ Unable to produce a valid Assertion for SAML 2-enabled user. """
        # Act:
        add_test_user_to_saml_team(self.login_email)
        response = self.client.get('/+saml', data=self.REQUEST_DATA, follow=True)
        soup = BeautifulSoup(response.content)
        inputtag = soup.findAll('input', {'name':'SAMLResponse'})[0]
        encoded_response = inputtag['value']
        samlresponse = base64.b64decode(encoded_response)

        # Assert:
        self.assertTrue(self.login_email in samlresponse)

    def test_user_has_different_preferred_email(self):
        """ User's preferred email is not being passed as SAML identifier. """
        # Act:
        add_test_user_to_saml_team(self.login_email)
        alternate_email = 'test' + self.login_email
        change_preferred_email(self.login_email, alternate_email)
        response = self.client.get('/+saml', data=self.REQUEST_DATA, follow=True)
        soup = BeautifulSoup(response.content)
        inputtag = soup.findAll('input', {'name':'SAMLResponse'})[0]
        encoded_response = inputtag['value']
        samlresponse = base64.b64decode(encoded_response)

        # Assert:
        self.assertTrue(alternate_email in samlresponse)

    def test_user_not_enabled(self):
        """ Assertion produced for user who is not enabled for SAML 2.0. """
        # Arrange/Act:
        response = self.client.get('/+saml', data=self.REQUEST_DATA, follow=True)

        # Assert:
        self.assertTemplateUsed(response, 'saml2idp/invalid_user.html')


class SalesForceAssertionTestCase(GoogleAppsAssertionTestCase):
    """
    Test SAML Assertions for a SalesForce Service Point.
    """
    ACS_URL = salesforce.SALESFORCE_ACS
    REQUEST_DATA = salesforce.REQUEST_DATA
