# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import datetime
import urlparse
from random import randint

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.core.cache import cache
from django.http import HttpRequest
from django.test import TestCase
from mock import Mock
from openid.extensions import pape
from openid.extensions.sreg import SRegRequest

from openid.message import (Message, IDENTIFIER_SELECT, OPENID1_URL_LIMIT,
    OPENID2_NS)
from openid.yadis.constants import YADIS_HEADER_NAME
from urllib import quote, quote_plus

import identityprovider.signed as signed
from identityprovider.models import (
    Account,
    OpenIDAuthorization,
    OpenIDRPConfig,
    OpenIDRPSummary,
    Person,
)
from identityprovider.models.account import LPOpenIdIdentifier
from identityprovider.models.const import (
    AccountStatus,
    AccountCreationRationale,
)
from identityprovider.models.person import PersonLocation
from identityprovider.models.authtoken import create_token

from identityprovider.views import server

from identityprovider.tests.utils import SSOBaseTestCase, AuthenticatedTestCase


class DummyORequest(object):
    mode = 'checkid_setup'
    trust_root = 'http://localhost/'
    message = Message()
    identity = 'http://localhost/+id/mark_oid'

    def idSelect(self):
        return False


class DummySession(dict):

    @property
    def session_key(self):
        return 'abc'

    def flush(self):
        pass


class DummyRequest(object):

    def __init__(self):
        self.session = DummySession()
        self.COOKIES = {}
        self.META = {}


def delete_rpconfig_cache_entry(trust_root):
        cache.delete(OpenIDRPConfig.cache_key(trust_root))



class HandleOpenIDErrorTestCase(SSOBaseTestCase):

    # tests for the _handle_openid_error method

    def test_handle_openid_error_with_encode_url(self):
        params = {'openid.return_to': 'http://localhost/'}
        r = self.client.get('/+openid', params)
        query = self.get_query(r)
        error_msg = 'No+mode+value+in+message+'
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.mode'], 'error')
        self.assertTrue(query['openid.error'].startswith(error_msg))

    def test_handle_openid_error_other(self):
        params = {'openid.mode': 'checkid_setup'}
        r = self.client.get('/+openid', params)
        error_mode = "mode:error"
        error_msg = "error:Missing required field 'return_to'"
        self.assertEqual(r.status_code, 200)
        self.assertTrue(error_mode in r.content)
        self.assertTrue(error_msg in r.content)


class ProcessOpenIDRequestTestCase(SSOBaseTestCase):

    # tests for the _process_openid_request method

    def test_process_openid_request_no_orequest(self):
        r = self.client.get('/+openid')
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'server_info.html')


class HandleUserResponseTestCase(SSOBaseTestCase):
    fixtures = ['test']

    def setUp(self):
        super(HandleUserResponseTestCase, self).setUp()

        # create a trusted rpconfig
        self.rpconfig = OpenIDRPConfig(trust_root='http://localhost/')
        self.rpconfig.save()

        self.params = {'openid.trust_root': 'http://localhost/',
                       'openid.return_to': 'http://localhost/',
                       'openid.identity': IDENTIFIER_SELECT,
                       'openid.claimed_id': 'http://localhost/~userid',
                       'openid.ns': OPENID2_NS,
                       'openid.mode': 'checkid_setup'}

    # tests for the _handle_user_response method

    def test_handle_user_response_checkid_immediate(self):
        self.params['openid.mode'] = 'checkid_immediate'

        r = self.client.get('/+openid', self.params)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.mode'], 'setup_needed')

    def test_handle_user_response_no_valid_openid(self):
        self.params.update({'openid.identity': 'bogus',
                            'openid.claimed_id': 'bogus'})
        r = self.client.get('/+openid', self.params)
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'invalid_identifier.html')

    def test_handle_user_response_openid_is_authorized_idselect(self):
        # update rp to auto authorize
        self.rpconfig.auto_authorize = True
        self.rpconfig.save()

        account = Account.objects.get_by_email('mark@example.com')
        self.client.login(username='mark@example.com', password='test')
        r = self.client.get('/+openid', self.params)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.mode'], 'id_res')
        self.assertEqual(query['openid.identity'],
                         quote_plus(account.openid_identity_url))

    def test_handle_user_response_openid_is_authorized_other_id(self):
        self.rpconfig.auto_authorize = True
        self.rpconfig.save()

        self.params['openid.identity'] = \
            settings.SSO_ROOT_URL + '+id/mark_oid'
        self.client.login(username='mark@example.com', password='test')
        r = self.client.get('/+openid', self.params)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.mode'], 'id_res')
        self.assertEqual(query['openid.identity'],
                         quote_plus(self.params['openid.identity']))

    def test_handle_user_response_user_is_authenticated(self):
        self.params['openid.identity'] = \
            settings.SSO_ROOT_URL + '+id/other_oid'
        self.client.login(username='mark@example.com', password='test')
        r = self.client.get('/+openid', self.params)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.mode'], 'cancel')

    def test_handle_user_response_decide(self):
        r = self.client.get('/+openid', self.params)
        self.assertEqual(r.status_code, 302)
        self.assertTrue(r['Location'].endswith('+decide'))

    def test_handle_user_response_with_referer(self):
        META = {'HTTP_REFERER': 'http://localhost/'}
        r = self.client.get('/+openid', self.params, **META)
        openid_referer = self.client.cookies.get('openid_referer')
        self.assertEqual(r.status_code, 302)
        self.assertTrue(r['Location'].endswith('+decide'))
        self.assertEqual(openid_referer.value, META['HTTP_REFERER'])

    def get_login_after_redirect_from_consumer(self):
        self.rpconfig.logo = 'http://someserver/logo.png'
        self.rpconfig.save()
        r = self.client.get('/+openid', self.params, follow=True)
        return r

    def get_new_account_after_redirect_to_login_from_consumer(self):
        r = self.get_login_after_redirect_from_consumer()
        path_parts = r.request['PATH_INFO'].split('/')
        token = path_parts[1]
        new_account_url = '/%s/+new_account' % token
        r = self.client.get(new_account_url)
        return r

    def test_logo_for_rpconfig_on_decide_page(self):
        r = self.get_login_after_redirect_from_consumer()
        self.assertEqual(r.status_code, 200)
        self.assertContains(r, "<img src='http://someserver/logo.png'/>")

    def test_logo_for_rpconfig_on_new_account_page(self):
        r = self.get_new_account_after_redirect_to_login_from_consumer()
        self.assertContains(r, "<img src='http://someserver/logo.png'/>")

    def test_rpconfig_fetch_on_new_account_page(self):
        r = self.get_new_account_after_redirect_to_login_from_consumer()
        # new account is called with a get with no rpconfig sent in
        # rpconfig should be refetched
        self.assertEqual(r.context['rpconfig'], self.rpconfig)

    def test_rpconfig_fetch_on_login_page(self):
        r = self.get_login_after_redirect_from_consumer()
        self.assertEqual(r.status_code, 200)
        path_parts = r.request['PATH_INFO'].split('/')
        token = path_parts[1]
        login_url = '/%s/+login' % token
        # leave username and password blank to cause validation error
        r = self.client.post(login_url)
        # rpconfig should be refetched
        self.assertEqual(r.context['rpconfig'], self.rpconfig)


class MultiLangOpenIDTestCase(TestCase):
    def setUp(self):
        self.orig_language_code = settings.LANGUAGE_CODE
        settings.LANGUAGE_CODE = 'en'
        self.orig_supported_languages = settings.SUPPORTED_LANGUAGES
        settings.SUPPORTED_LANGUAGES = ['en', 'de']

    def flag_icon(self, lang):
        tag = '<img src="/assets/identityprovider/flags/%s.png" alt="%s" />'
        return tag % (lang, lang)

    def tearDown(self):
        settings.LANGUAGE_CODE = self.orig_language_code
        settings.SUPPORTED_LANGUAGES = self.orig_supported_languages

    def test_no_lang_specified(self):
        response = self.client.get('/+openid')
        self.assertContains(response, 'This is the Ubuntu SSO service')

    def test_german(self):
        response = self.client.get('/de/+openid')
        self.assertContains(response, self.flag_icon('de'))
        self.assertEqual('de', self.client.session['django_language'])

    def test_german_with_country_code(self):
        # This should default back to 'de'.
        response = self.client.get('/de_CH/+openid')
        self.assertContains(response, self.flag_icon('de'))
        self.assertEqual('de', self.client.session['django_language'])

    def test_unsupported_language(self):
        response = self.client.get('/sw/+openid')
        self.assertContains(response, self.flag_icon('en'))
        self.assertEqual('en', self.client.session['django_language'])

    def test_language_persists_in_session(self):
        self.client.get('/de/+openid')
        self.assertEqual('de', self.client.session['django_language'])
        # Requesting a second time will still bring back German
        # if no other setting takes precedence (like the user's preference)
        self.client.get('/+openid')
        self.assertEqual('de', self.client.session['django_language'])


class ValidOpenIDTestCase(SSOBaseTestCase):

    def test_is_valid_openid_idselect(self):
        valid = server._is_valid_openid_for_this_site(IDENTIFIER_SELECT)
        self.assertTrue(valid)

    def test_is_valid_openid_different_scheme(self):
        srvparts = urlparse.urlparse(settings.SSO_ROOT_URL)
        if srvparts.scheme == 'http':
            scheme = 'https'
        else:
            scheme = 'http'
        identity = settings.SSO_ROOT_URL.replace(srvparts.scheme, scheme)
        valid = server._is_valid_openid_for_this_site(identity)
        self.assertFalse(valid)

    def test_is_valid_openid_different_port(self):
        srvparts = urlparse.urlparse(settings.SSO_ROOT_URL)
        if srvparts.port is not None:
            port = int(srvparts.port) + 1
        else:
            port = 81
        identity = "%s:%s//%s%s" % (srvparts.scheme, port,
            srvparts.hostname, srvparts.path)
        valid = server._is_valid_openid_for_this_site(identity)
        self.assertFalse(valid)

    def test_is_valid_openid_different_hostname(self):
        identity = 'http://testserver/'
        valid = server._is_valid_openid_for_this_site(identity)
        self.assertFalse(valid)

    def test_is_valid_openid_path_patterns(self):
        identity = settings.SSO_ROOT_URL
        valid = server._is_valid_openid_for_this_site(identity)
        self.assertTrue(valid)

        identity += '+id/foo'
        valid = server._is_valid_openid_for_this_site(identity)
        self.assertTrue(valid)

        identity = settings.SSO_ROOT_URL + '~foo'
        valid = server._is_valid_openid_for_this_site(identity)
        self.assertTrue(valid)

        # try a non-normalized uri
        identity = settings.SSO_ROOT_URL.strip('/')
        valid = server._is_valid_openid_for_this_site(identity)
        self.assertTrue(valid)

    def test_is_valid_openid_other(self):
        identity = settings.SSO_ROOT_URL + '+foo'
        valid = server._is_valid_openid_for_this_site(identity)
        self.assertFalse(valid)

    def test_is_valid_openid_error(self):
        valid = server._is_valid_openid_for_this_site(None)
        self.assertFalse(valid)


class DecideTestCase(AuthenticatedTestCase):

    def _prepare_openid_token(self, param_overrides=None):
        request = {'openid.mode': 'checkid_setup',
                   'openid.trust_root': 'http://localhost/',
                   'openid.return_to': 'http://localhost/',
                   'openid.identity': IDENTIFIER_SELECT}
        if param_overrides:
            request.update(param_overrides)
        openid_server = server._get_openid_server()
        self.orequest = openid_server.decodeRequest(request)
        self.token = create_token(16)
        session = self.client.session
        session[self.token] = signed.dumps(self.orequest, settings.SECRET_KEY)
        session.save()

    def setUp(self):
        super(DecideTestCase, self).setUp(disableCSRF=True)
        self._prepare_openid_token()

    def test_decide_invalid(self):
        token = 'a' * 16
        r = self.client.get("/%s/+decide" % token)
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.content, 'Invalid OpenID transaction')

    def test_decide_authenticated(self):
        r = self.client.post("/%s/+decide" % self.token, {'yes': 'yes'})
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.mode'], 'id_res')

    def test_decide_authenticated_with_post(self):
        # Using a return-to URL that is more than OPENID1_URL_LIMIT
        # characters long will force the assertion to be sent as a
        # POSTing form (instead of a 302 redirect).
        return_to = 'http://localhost/' + (OPENID1_URL_LIMIT * 'a')
        self._prepare_openid_token({
            'openid.return_to': return_to,
            'openid.ns': OPENID2_NS,
            'openid.claimed_id': 'http://localhost/~userid'})
        r = self.client.post("/%s/+decide" % self.token, {'yes': ''})
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r['Content-Type'], 'text/html')
        self.assertContains(r, '<form ')
        self.assertContains(r, return_to)

    def test_decide_auto_authorize(self):
        # make sure rpconfig is set to auto authorize
        rpconfig = OpenIDRPConfig(trust_root='http://localhost/',
                                  auto_authorize=True)
        rpconfig.save()

        r = self.client.post("/%s/+decide" % self.token)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.mode'], 'id_res')

    def test_decide_process(self):
        r = self.client.post("/%s/+decide" % self.token)
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'decide.html')

    def test_decide_not_authenticated(self):
        self.client.logout()

        # create a trusted rpconfig
        rpconfig = OpenIDRPConfig(trust_root='http://localhost/')
        rpconfig.save()

        # start openid request
        params = {'openid.trust_root': 'http://localhost/',
                  'openid.return_to': 'http://localhost/',
                  'openid.identity': IDENTIFIER_SELECT,
                  'openid.claimed_id': 'http://localhost/~userid',
                  'openid.ns': OPENID2_NS,
                  'openid.mode': 'checkid_setup'}
        r = self.client.get('/+openid', params)
        # follow redirect
        path = r['Location'].split('http://testserver')[1]
        self.assertTrue(path.endswith('+decide'))
        r = self.client.get(path)
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'registration/login.html')

    def test_decide_multiple_openidrpsummary(self):
        # create multiple matching OpenIDRPSummary objects
        account = Account.objects.get_by_email(self.login_email)
        trust_root = 'http://localhost/'
        delete_rpconfig_cache_entry(trust_root)

        OpenIDRPSummary.objects.create(account=account, trust_root=trust_root,
            openid_identifier='http://localhost/~openid1')
        OpenIDRPSummary.objects.create(account=account, trust_root=trust_root,
            openid_identifier='http://otherhost/~openid1')

        r = self.client.post("/%s/+decide" % self.token)
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'decide.html')


    def test_decide_team_membership_with_auto_authorize(self):
        # make sure rpconfig is set to auto authorize
        self.client.login(username='mark@example.com',
                          password='test')

        rpconfig = OpenIDRPConfig(trust_root='http://localhost/',
                                  auto_authorize=True)
        rpconfig.save()

        param_overrides = {
            'openid.lp.query_membership': 'ubuntu-team',
        }
        self._prepare_openid_token(param_overrides=param_overrides)
        r = self.client.post("/%s/+decide" % self.token)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.mode'], 'id_res')
        self.assertEqual(query['openid.lp.is_member'], 'ubuntu-team')

    def test_yui_js_url(self):
        r = self.client.get("/%s/+decide" % self.token)
        self.assertContains(r,
            '{0}lazr-js/yui/yui-min.js'.format(settings.STATIC_URL))

    def test_check_team_membership_multiple_openidrpsummary(self):
        # create multiple matching OpenIDRPSummary objects
        account = Account.objects.get_by_email(self.login_email)
        trust_root = 'http://localhost/'
        OpenIDRPSummary.objects.create(account=account, trust_root=trust_root,
            openid_identifier='http://localhost/~openid1')
        OpenIDRPSummary.objects.create(account=account, trust_root=trust_root,
            openid_identifier='http://otherhost/~openid1')

        request = Mock()
        request.user = account
        oresponse = Mock()
        # _check_team_membership should not raise an error
        # no extra check is needed
        server._check_team_membership(request, self.orequest, oresponse)

    def test_list_of_details_is_complete(self):
        self.client.login(username='mark@example.com',
                          password='test')

        # create a trusted rpconfig
        rpconfig = OpenIDRPConfig(trust_root='http://localhost/',
            allowed_sreg='fullname,email,language',
            can_query_any_team=True,
            description="Some description",
        )
        rpconfig.save()
        param_overrides = {
            'openid.sreg.required': 'nickname,email,fullname,language',
            'openid.lp.query_membership': 'ubuntu-team',
        }
        self._prepare_openid_token(param_overrides=param_overrides)
        response = self.client.get('/%s/+decide' % self.token)
        self.assertContains(response, "Team membership")
        self.assertContains(response, "Full name")
        self.assertContains(response, "Email address")
        self.assertContains(response, "Preferred language")

    def test_state_of_checkboxes_and_data_formats_trusted(self):
        self.client.login(username='mark@example.com',
                          password='test')

        # create a trusted rpconfig
        rpconfig = OpenIDRPConfig(trust_root='http://localhost/',
            allowed_sreg='nickname,email,language',
            can_query_any_team=True,
            description="Some description",
        )
        rpconfig.save()
        param_overrides = {
            'openid.sreg.required': 'nickname,email,fullname',
            'openid.sreg.optional': 'language',
            'openid.lp.query_membership': 'ubuntu-team,launchpad-team,isd-team',
        }
        self._prepare_openid_token(param_overrides=param_overrides)
        response = self.client.get('/%s/+decide' % self.token)
        # checkbox checked and disabled for required fields and label is bold
        username_html = ('<li><input checked="checked" name="nickname" '
                         'value="mark" class="required" disabled="disabled" '
                         'type="checkbox" id="id_nickname" /> <label '
                         'for="id_nickname">Username: mark</label></li>')
        email_html = ('<li><input checked="checked" name="email" '
                      'value="mark@example.com" class="required" '
                      'disabled="disabled" type="checkbox" id="id_email" /> '
                      '<label for="id_email">Email address: '
                      'mark@example.com</label></li>')
        # checkbox checked and enabled for optional fields and label is plain
        language_html = ('<li><input checked="checked" type="checkbox" '
                         'name="language" value="en" id="id_language" /> '
                         '<label for="id_language">Preferred language: en'
                         '</label></li>')
        # team data is enabled and checked, the label is plain
        team_html_1 = ('<li><input checked="checked" type="checkbox" '
                       'name="ubuntu-team" value="ubuntu-team" id="id_'
                       'ubuntu-team" /> <label for="id_ubuntu-team">'
                       'ubuntu-team</label></li>')
        team_html_2 = ('<li><input checked="checked" type="checkbox" '
                       'name="isd-team" value="isd-team" id="id_isd-team" /> '
                       '<label for="id_isd-team">isd-team</label></li>')

        self.assertContains(response, username_html)
        self.assertContains(response, email_html)
        self.assertContains(response, language_html)
        self.assertContains(response, team_html_1)
        self.assertContains(response, team_html_2)

    def test_state_of_checkboxes_and_data_formats_untrusted(self):
        self.client.login(username='mark@example.com',
                          password='test')

        # create a trusted rpconfig
        trust_root = 'http://untrusted/'
        rpconfig = OpenIDRPConfig(trust_root=trust_root,
            can_query_any_team=True,
            description="Some description",
        )

        delete_rpconfig_cache_entry("http://localhost/")

        rpconfig.save()
        param_overrides = {
            'openid.sreg.required': 'nickname,email,fullname',
            'openid.sreg.optional': 'language',
            'openid.lp.query_membership': 'ubuntu-team',
        }
        self._prepare_openid_token(param_overrides=param_overrides)
        response = self.client.get('/%s/+decide' % self.token)
        # checkbox checked and *enabled* for required fields and label is bold
        username_html = ('<li><input checked="checked" name="nickname" '
                         'value="mark" class="required" type="checkbox" '
                         'id="id_nickname" /> <label for="id_nickname">'
                         'Username: mark</label></li>')
        email_html = ('<li><input checked="checked" name="email" '
                      'value="mark@example.com" class="required" type='
                      '"checkbox" id="id_email" /> <label for="id_email">'
                      'Email address: mark@example.com</label></li>')
        # checkbox *not checked* and enabled for optional fields & label's plain
        language_html = ('<li><input type="checkbox" name="language" value="'
                         'en" id="id_language" /> <label for="id_language">'
                         'Preferred language: en</label></li>')
        # team data is enabled and *not checked*, the label is plain
        team_html = ('<li><input type="checkbox" name="ubuntu-team" '
                     'value="ubuntu-team" id="id_ubuntu-team" /> '
                     '<label for="id_ubuntu-team">Team membership:</label> '
                     '<label for="id_ubuntu-team">ubuntu-team</label></li>')
        self.assertContains(response, username_html)
        self.assertContains(response, email_html)
        self.assertContains(response, language_html)
        self.assertContains(response, team_html)


class PreAuthorizeTestCase(AuthenticatedTestCase):
    def setUp(self):
        super(PreAuthorizeTestCase, self).setUp(disableCSRF=True)

    def test_pre_authorize_get(self):
        r = self.client.get('/+pre-authorize-rp')
        self.assertEqual(r.status_code, 400)

    def test_pre_authorize_unauthorized(self):
        r = self.client.post('/+pre-authorize-rp',
            {'trust_root': 'http://localhost/',
             'callback': 'http://localhost/'})
        self.assertEqual(r.status_code, 400)

    def test_pre_authorize_no_pre_authorized(self):
        old_OPENID_PREAUTHORIZATION_ACL = settings.OPENID_PREAUTHORIZATION_ACL
        settings.OPENID_PREAUTHORIZATION_ACL = []

        rpconfig = OpenIDRPConfig(trust_root='http://localhost/')
        rpconfig.save()

        extra = {'HTTP_REFERER': 'http://localhost/'}
        r = self.client.post('/+pre-authorize-rp',
            {'trust_root': 'http://localhost/',
             'callback': 'http://localhost/'},
            **extra)
        self.assertEqual(r.status_code, 400)

        settings.OPENID_PREAUTHORIZATION_ACL = old_OPENID_PREAUTHORIZATION_ACL

    def test_pre_authorize_authenticated(self):
        old_OPENID_PREAUTHORIZATION_ACL = settings.OPENID_PREAUTHORIZATION_ACL
        settings.OPENID_PREAUTHORIZATION_ACL = [
            ('http://localhost/', 'http://localhost/')
        ]
        rpconfig = OpenIDRPConfig(trust_root='http://localhost/')
        rpconfig.save()

        extra = {'HTTP_REFERER': 'http://localhost/'}
        r = self.client.post('/+pre-authorize-rp',
            {'trust_root': 'http://localhost/',
             'callback': 'http://localhost/'},
            **extra)
        self.assertRedirects(r, 'http://localhost/')

        settings.OPENID_PREAUTHORIZATION_ACL = old_OPENID_PREAUTHORIZATION_ACL

    def test_pre_authorize_not_authenticated(self):
        old_OPENID_PREAUTHORIZATION_ACL = settings.OPENID_PREAUTHORIZATION_ACL
        settings.OPENID_PREAUTHORIZATION_ACL = [
            ('http://localhost/', 'http://localhost/')
        ]
        rpconfig = OpenIDRPConfig(trust_root='http://localhost/')
        rpconfig.save()

        self.client.logout()
        extra = {'HTTP_REFERER': 'http://localhost/'}
        r = self.client.post('/+pre-authorize-rp',
            {'trust_root': 'http://localhost/',
             'callback': 'http://localhost/'},
            **extra)
        next_url = '/+login?next=' + quote('/+pre-authorize-rp?')
        self.assertRedirects(r, next_url)

        settings.OPENID_PREAUTHORIZATION_ACL = old_OPENID_PREAUTHORIZATION_ACL

    def test_pre_authorize_after_login(self):
        old_OPENID_PREAUTHORIZATION_ACL = settings.OPENID_PREAUTHORIZATION_ACL
        settings.OPENID_PREAUTHORIZATION_ACL = [
            ('http://localhost/', 'http://localhost/')
        ]
        rpconfig = OpenIDRPConfig(trust_root='http://localhost/')
        rpconfig.save()

        # make sure we are logged out
        self.client.logout()

        # attempt to pre-authorize
        extra = {'HTTP_REFERER': 'http://localhost/'}
        data = {'trust_root': 'http://localhost/',
                'callback': 'http://localhost/'}
        r = self.client.post('/+pre-authorize-rp', data, **extra)
        # we get redirected to login
        next_url = '/+login?next=' + quote('/+pre-authorize-rp?')
        self.assertRedirects(r, next_url)
        # and the referer info is stored in the session
        self.assertEqual(self.client.session['pre_auth_referer'],
                         'http://localhost/')
        self.assertTrue(self.client.session['pre_auth_referer_for'],
                        'http://localhost/')

        # login and redirect to pre-authorize again
        data.update({'email': 'mark@example.com', 'password': 'test',
                     'next': '/+pre-authorize-rp'})
        r = self.client.post('/+login', data, **extra)
        r = self.client.post('/+pre-authorize-rp', data, **extra)
        # we get effectively pre-authorized
        self.assertRedirects(r, 'http://localhost/')
        # and the pre-auth session data gets removed
        self.assertFalse('pre_auth_referer' in self.client.session)
        self.assertFalse('pre_auth_referer_for' in self.client.session)

        settings.OPENID_PREAUTHORIZATION_ACL = old_OPENID_PREAUTHORIZATION_ACL

    def test_pre_authorize_sess_referer_not_trust_root(self):
        old_OPENID_PREAUTHORIZATION_ACL = settings.OPENID_PREAUTHORIZATION_ACL
        settings.OPENID_PREAUTHORIZATION_ACL = [
            ('http://otherhost/', 'http://localhost/')
        ]
        rpconfig = OpenIDRPConfig(trust_root='http://localhost/')
        rpconfig.save()

        # make sure we are logged out
        self.client.logout()

        # attempt to pre-authorize
        extra = {'HTTP_REFERER': 'http://otherhost/'}
        data = {'trust_root': 'http://localhost/',
                'callback': 'http://localhost/'}
        r = self.client.post('/+pre-authorize-rp', data, **extra)
        # we get redirected to login
        next_url = '/+login?next=' + quote('/+pre-authorize-rp?')
        self.assertRedirects(r, next_url)
        # and the referer info is stored in the session
        self.assertEqual(self.client.session['pre_auth_referer'],
                         'http://otherhost/')
        self.assertTrue(self.client.session['pre_auth_referer_for'],
                        'http://localhost/')

        # attempt to pre-authorize for a different trust_root
        data['trust_root'] = 'http://otherhost/'
        r = self.client.post('/+pre-authorize-rp', data, **extra)
        # pre-authorization is denied
        self.assertEqual(r.status_code, 400)

        settings.OPENID_PREAUTHORIZATION_ACL = old_OPENID_PREAUTHORIZATION_ACL


class CancelTestCase(AuthenticatedTestCase):
    def setUp(self):
        super(CancelTestCase, self).setUp(disableCSRF=True)

        self.params = request = {'openid.mode': 'checkid_setup',
                                 'openid.trust_root': 'http://localhost/',
                                 'openid.return_to': 'http://localhost/',
                                 'openid.identity': IDENTIFIER_SELECT}
        openid_server = server._get_openid_server()
        self.orequest = openid_server.decodeRequest(request)
        self.token = create_token(16)
        session = self.client.session
        session[self.token] = signed.dumps(self.orequest,
                                      settings.SECRET_KEY)
        session.save()

        rpconfig = OpenIDRPConfig(trust_root='http://localhost')
        rpconfig.save()

    def test_cancel_invalid_openid_transaction(self):
        token = 'a' * 16
        r = self.client.get("/%s/+cancel" % token)
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.content, 'Invalid OpenID transaction')

    def test_cancel_user_is_authenticated(self):
        # cancel request
        r = self.client.get("/%s/+cancel" % self.token)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.mode'], 'cancel')

    def test_cancel_user_is_not_authenticated(self):
        # save session data
        session_data = self.client.session[self.token]

        # logout
        self.client.logout()
        # request something so we get a real session in the client
        self.client.get('/+openid', self.params)

        # manipulate session
        session = self.client.session
        session[self.token] = session_data
        session.save()

        # cancel request
        r = self.client.get("/%s/+cancel" % self.token)
        query = self.get_query(r)
        self.assertEqual(r.status_code, 302)
        self.assertEqual(query['openid.mode'], 'cancel')


class XRDSTestCase(SSOBaseTestCase):
    fixtures = ['test']

    def assert_xrds_document(self, response, template=None):
        if template is None:
            template = 'openidapplication-xrds.xml'
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['Content-Type'], 'application/xrds+xml')
        self.assertTemplateUsed(response, template)

    def test_xrds(self):
        r = self.client.get('/+xrds')
        self.assert_xrds_document(r)

    def test_root_page_also_returns_xrds_document_when_right_header_is_used(self):
        r = self.client.get('/', HTTP_ACCEPT="application/xrds+xml")
        self.assert_xrds_document(r)

    def test_identity_page_nonexisting_account(self):
        r = self.client.get('/+id/aaaaaaaaaaaaaaaa')
        self.assertEqual(r.status_code, 404)

    def test_identity_page_inactive_account(self):
        account = Account.objects.get_by_email('mark@example.com')
        account.status = AccountStatus.DEACTIVATED
        account.save()
        r = self.client.get("/+id/%s" % account.openid_identifier)
        self.assertEqual(r.status_code, 404)

    def test_identity_page_active_account(self):
        account = Account.objects.get_by_email('mark@example.com')
        r = self.client.get("/+id/%s" % account.openid_identifier)
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'person.html')
        self.assertEqual(r[YADIS_HEADER_NAME],
                         "%s/+xrds" % account.openid_identity_url)

    def test_xrds_identity_page_nonexisting_account(self):
        r = self.client.get('/+id/aaaaaaaaaaaaaaaa')
        self.assertEqual(r.status_code, 404)

    def test_xrds_identity_page_inactive_account(self):
        account = Account.objects.get_by_email('mark@example.com')
        account.status = AccountStatus.DEACTIVATED
        account.save()
        r = self.client.get("/+id/%s/+xrds" % account.openid_identifier)
        self.assertEqual(r.status_code, 404)

    def test_xrds_identity_page_active_account(self):
        account = Account.objects.get_by_email('mark@example.com')
        r = self.client.get("/+id/%s/+xrds" % account.openid_identifier)
        self.assert_xrds_document(r, template='person-xrds.xml')


class OpenIDAuthorizedTestCase(AuthenticatedTestCase):
    fixtures = ['test']

    def setUp(self):
        super(OpenIDAuthorizedTestCase, self).setUp()

        self.request = DummyRequest()
        self.request.user = Account.objects.get_by_email('mark@example.com')
        # we use now() here because last_login maps to django's auth model
        # which uses localtime.
        self.request.user.last_login = datetime.datetime.now()
        self.orequest = DummyORequest()
        self.orequest.identity = settings.SSO_ROOT_URL + '+id/mark_oid'

        @classmethod
        def mock_fromOpenIDRequest(cls, request):
            return None
        self.old_fromOpenIDRequest = pape.Request.fromOpenIDRequest
        pape.Request.fromOpenIDRequest = mock_fromOpenIDRequest

    def tearDown(self):
        pape.Request.fromOpenIDRequest = self.old_fromOpenIDRequest

        super(OpenIDAuthorizedTestCase, self).tearDown()

    def test_openid_is_authorized_not_authenticated(self):
        self.request.user = AnonymousUser()
        r = server._openid_is_authorized(self.request, self.orequest)
        self.assertFalse(r)

    def test_openid_is_authorized_id_owner(self):
        self.orequest.identity = 'http://localhost/+id/mark_oid'
        r = server._openid_is_authorized(self.request, self.orequest)
        self.assertFalse(r)

    def test_openid_is_authorized_should_reauthenticate(self):
        @classmethod
        def mock_fromOpenIDRequest(cls, request):
            class MockPapeRequest(object):
                max_auth_age = '0'
            return MockPapeRequest()
        old_fromOpenIDRequest = pape.Request.fromOpenIDRequest
        pape.Request.fromOpenIDRequest = mock_fromOpenIDRequest

        r = server._openid_is_authorized(self.request, self.orequest)
        self.assertFalse(r)

        pape.Request.fromOpenIDRequest = old_fromOpenIDRequest

    def test_openid_is_authorized_rpconfig(self):
        rpconfig = OpenIDRPConfig(trust_root=self.orequest.trust_root)
        rpconfig.auto_authorize = True
        rpconfig.save()

        r = server._openid_is_authorized(self.request, self.orequest)
        self.assertTrue(r)

    def test_openid_is_authorized_other(self):
        r = server._openid_is_authorized(self.request, self.orequest)
        self.assertFalse(r)

        expires = datetime.datetime.utcnow() + datetime.timedelta(1)
        OpenIDAuthorization.objects.authorize(self.request.user,
            self.orequest.trust_root, expires,
            self.request.session.session_key)

        r = server._openid_is_authorized(self.request, self.orequest)
        self.assertTrue(r)


class ShouldReauthenticateTestCase(SSOBaseTestCase):
    fixtures = ['test']

    def setUp(self):
        self.user = Account.objects.get_by_email('mark@example.com')
        self.orequest = DummyORequest()
        self.old_fromOpenIDRequest = pape.Request.fromOpenIDRequest

    def tearDown(self):
        pape.Request.fromOpenIDRequest = self.old_fromOpenIDRequest

    def test_should_reauthenticate_no_pape(self):
        # pape_request is None
        @classmethod
        def mock_fromOpenIDRequest(cls, orequest):
            return None
        pape.Request.fromOpenIDRequest = mock_fromOpenIDRequest

        r = server._should_reauthenticate(self.orequest, self.user)
        self.assertFalse(r)

    def test_should_reauthenticate_no_pape_max_auth_age(self):
        # pape_request.max_auth_age is None
        @classmethod
        def mock_fromOpenIDRequest(cls, request):
            class MockPapeRequest(object):
                max_auth_age = None
            return MockPapeRequest()
        pape.Request.fromOpenIDRequest = mock_fromOpenIDRequest

        r = server._should_reauthenticate(self.orequest, self.user)
        self.assertFalse(r)

    def test_should_reauthenticate_invalid_max_auth_age(self):
        # max_auth_age raise ValueError
        @classmethod
        def mock_fromOpenIDRequest(cls, request):
            class MockPapeRequest(object):
                max_auth_age = 'bad-value'
            return MockPapeRequest()
        pape.Request.fromOpenIDRequest = mock_fromOpenIDRequest

        r = server._should_reauthenticate(self.orequest, self.user)
        self.assertFalse(r)

    def test_should_reauthenticate_true(self):
        # last_login <= cutoff
        @classmethod
        def mock_fromOpenIDRequest(cls, request):
            class MockPapeRequest(object):
                max_auth_age = '0'
            return MockPapeRequest()
        pape.Request.fromOpenIDRequest = mock_fromOpenIDRequest
        # we use now() here because last_login maps to django's auth model
        # which uses localtime.
        self.user.last_login = datetime.datetime.now()

        r = server._should_reauthenticate(self.orequest, self.user)
        self.assertTrue(r)

    def test_should_reauthenticate_false(self):
        # last_login > cutoff
        @classmethod
        def mock_fromOpenIDRequest(cls, request):
            class MockPapeRequest(object):
                max_auth_age = '100'
            return MockPapeRequest()
        pape.Request.fromOpenIDRequest = mock_fromOpenIDRequest
        # we use now() here because last_login maps to django's auth model
        # which uses localtime.
        self.user.last_login = datetime.datetime.now()

        r = server._should_reauthenticate(self.orequest, self.user)
        self.assertFalse(r)


class UntrustedRPTest(SSOBaseTestCase):
    fixtures = ['test']

    def setUp(self):
        # Ensure that we're restricting RPs for these tests
        self.old_restrict = getattr(settings, 'SSO_RESTRICT_RP', True)
        settings.SSO_RESTRICT_RP = True

    def tearDown(self):
        settings.SSO_RESTRICT_RP = self.old_restrict

    def test_process_unknown_rp(self):
        orequest = DummyORequest()
        request = DummyRequest()
        response = server._process_openid_request(request, orequest, None)
        self.assertEquals(302, response.status_code)
        self.assert_(response['Location'].endswith('+untrusted'))

    def test_untrusted(self):
        request = {'openid.mode': 'checkid_setup',
                   'openid.trust_root': 'http://localhost/',
                   'openid.return_to': 'http://localhost/',
                   'openid.identity': IDENTIFIER_SELECT}
        openid_server = server._get_openid_server()
        orequest = openid_server.decodeRequest(request)
        token = create_token(16)

        # call up a session-modifying view to get a real session object
        r = self.client.login(username='mark@example.com', password='test')
        self.assertTrue(r)

        session = self.client.session
        session[token] = signed.dumps(orequest,
                                      settings.SECRET_KEY)
        session.save()

        r = self.client.get("/%s/+untrusted" % token)
        self.assertEqual(r.status_code, 200)
        self.assertTemplateUsed(r, 'untrusted.html')


class TestSregFields(SSOBaseTestCase):

    fixtures = ["test"]

    def setUp(self):
        super(TestSregFields, self).setUp()

        self.account = Account(
            creation_rationale=AccountCreationRationale.USER_CREATED,
            status=AccountStatus.ACTIVE,
            displayname='User')
        self.account.save()

        lp_account = randint(1, 9999)
        LPOpenIdIdentifier.objects.create(
            identifier=self.account.openid_identifier,
            lp_account=lp_account)
        person = Person(lp_account=lp_account)
        person.save()
        now = datetime.datetime.utcnow()
        personlocation = PersonLocation(date_created=now,
                                        person=person, time_zone='UTC',
                                        last_modified_by=person,
                                        date_last_modified=now)
        personlocation.save()
        self.sreg_request = SRegRequest()
        self.rpconfig = OpenIDRPConfig()
        self.request = DummyRequest()
        self.request.user = self.account


class MarkupTestCase(SSOBaseTestCase):

    fixtures = ["test"]

    def test_untrusted_rp_properly_shows_markup(self):
        self.client.login(username='mark@example.com', password='test')

        session = self.client.session
        token = 'a' * 16
        session[token] = signed.dumps(DummyORequest())
        session.save()

        delete_rpconfig_cache_entry("http://localhost/")

        r = self.client.get('/%s/+decide' % token)
        self.assertContains(r, '<em>not</em>')


class ApprovedDataTest(SSOBaseTestCase):
    fixtures = ['test']

    def _get_openid_request(self, with_sreg=True, with_teams=True):
        request = {
            'openid.mode': 'checkid_setup',
            'openid.trust_root': 'http://localhost/',
            'openid.return_to': 'http://localhost/',
            'openid.identity': IDENTIFIER_SELECT}
        if with_sreg:
            request['openid.sreg.required'] = 'email,fullname'
        if with_teams:
            request['openid.lp.query_membership'] = 'ubuntu-team'
        openid_server = server._get_openid_server()
        return openid_server.decodeRequest(request)

    def _get_request_with_post_args(self, args={}):
        request = HttpRequest()
        request.user = Account.objects.get(pk=1)
        request.POST = args
        request.META = {'REQUEST_METHOD': 'POST'}
        return request

    def setUp(self):
        # Ensure that we're restricting RPs for these tests
        self.old_restrict = getattr(settings, 'SSO_RESTRICT_RP', True)
        settings.SSO_RESTRICT_RP = False

    def tearDown(self):
        settings.SSO_RESTRICT_RP = self.old_restrict

    def test_approved_data_returns_none_for_no_request(self):
        result = server._get_approved_data(HttpRequest(), None)
        self.assertEqual(result, None)

    def test_approved_data_for_sreg_only(self):
        post_args = {'email': 'email', 'ubuntu-team': 'ubuntu-team'}
        result = server._get_approved_data(
            self._get_request_with_post_args(post_args),
            self._get_openid_request(True, False))
        self.assertEqual(sorted(result['sreg']['requested']),
            ['email', 'fullname'])
        self.assertEqual(result['sreg']['approved'], ['email'])
        self.assertFalse(result.has_key('teams'))

    def test_approved_data_for_teams_only(self):
        post_args = {'ubuntu-team': 'ubuntu-team'}
        result = server._get_approved_data(
            self._get_request_with_post_args(post_args),
            self._get_openid_request(False, True))
        self.assertEqual(result['teams']['requested'], ['ubuntu-team'])
        self.assertEqual(result['teams']['approved'], ['ubuntu-team'])
        self.assertFalse(result.has_key('sreg'))

    def test_approved_data_for_sreg_and_teams(self):
        post_args = {'email': 'email', 'ubuntu-team': 'ubuntu-team'}
        result = server._get_approved_data(
            self._get_request_with_post_args(post_args),
            self._get_openid_request())
        self.assertEqual(sorted(result['sreg']['requested']),
            ['email', 'fullname'])
        self.assertEqual(result['sreg']['approved'], ['email'])
        self.assertEqual(result['teams']['requested'], ['ubuntu-team'])
        self.assertEqual(result['teams']['approved'], ['ubuntu-team'])
