# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import urllib
import urllib2

from datetime import date
from cStringIO import StringIO

from canonical.isd.tests.wsgi_intercept import WSGIInterceptedTestCase
from django.conf import settings
from django.db import connection
from django.http import HttpRequest
from django.test import TestCase, TransactionTestCase
from django.utils.unittest import skipIf

from lazr.restfulclient.resource import ServiceRoot

from openid.consumer.discover import (
    OpenIDServiceEndpoint, OPENID_1_0_TYPE, OPENID_1_1_TYPE,
    OPENID_2_0_TYPE, OPENID_IDP_2_0_TYPE)
from openid.message import IDENTIFIER_SELECT

from identityprovider.wsgi import make_app


def generate_openid_identifier():
    from random import choice
    chars = '34678bcdefhkmnprstwxyzABCDEFGHJKLMNPQRTWXY'
    length = 7

    loop_count = 0
    while loop_count < 20000:
        oid = ''.join(choice(chars) for count in range(length))
        rv = connection.connection.execute(
            ("SELECT COUNT(*) AS num "
             "FROM Account WHERE openid_identifier = ?"), [oid])
        if rv.fetchone()[0] == 0:
            return oid
        loop_count += 1


def create_generate_openid_identifier_function():
    connection.connection.create_function(
        'generate_openid_identifier', 0, generate_openid_identifier)


class patch_settings(object):

    def __init__(self, **kwargs):
        self.marker = object()
        self.old_settings = {}
        self.kwargs = kwargs

    def __enter__(self):
        for setting, new_value in self.kwargs.items():
            old_value = getattr(settings, setting, self.marker)
            self.old_settings[setting] = old_value
            setattr(settings, setting, new_value)

    def __exit__(self, exc_type, exc_val, exc_tb):
        for setting, old_value in self.old_settings.items():
            if old_value is self.marker:
                delattr(settings, setting)
            else:
                setattr(settings, setting, old_value)


class SSOBaseTestCase(TestCase):

    def _pre_setup(self):
        super(SSOBaseTestCase, self)._pre_setup()
        if connection.vendor == 'sqlite':
            create_generate_openid_identifier_function()

    def get_query(self, response):
        query_start = response['Location'].find('?')
        query_str = response['Location'][query_start + 1:]
        query_items = map(tuple,
            [item.split('=') for item in query_str.split('&')])
        query = dict(query_items)
        return query

    def reset_csrf(self):
        settings.MIDDLEWARE_CLASSES = self._MIDDLEWARE_CLASSES

    def disable_csrf(self):
        self._MIDDLEWARE_CLASSES = settings.MIDDLEWARE_CLASSES
        settings.MIDDLEWARE_CLASSES = (x for x in self._MIDDLEWARE_CLASSES
                                         if not 'csrf' in x.lower())


class LPAccountTestCase(TransactionTestCase):

    def setUp(self):
        super(LPAccountTestCase, self).setUp()

        # disable person constraint
        cursor = connection.cursor()
        # Committing any transactions, as you can't alter table inside one
        cursor.execute("COMMIT")
        cursor.execute('ALTER TABLE emailaddress DROP CONSTRAINT '
                       'emailaddress_person_fkey')
        cursor.execute("BEGIN")

    def tearDown(self):
        # enable person constraint
        cursor = connection.cursor()
        cursor.execute("COMMIT")
        cursor.execute('ALTER TABLE emailaddress ADD CONSTRAINT '
                       'emailaddress_person_fkey FOREIGN KEY (person) '
                       'REFERENCES lp_person(id) DEFERRABLE INITIALLY '
                       'DEFERRED')
        super(LPAccountTestCase, self).tearDown()


class AuthenticatedTestCase(SSOBaseTestCase):

    fixtures = ["test"]

    def setUp(self, disableCSRF=False):
        # We allow classes that inherit this one to disable CSRF when needed,
        # as CSRF breaks Django's test client:
        # http://code.djangoproject.com/ticket/9172
        if disableCSRF:
            self.old_middlewares = settings.MIDDLEWARE_CLASSES
            settings.MIDDLEWARE_CLASSES = tuple(x for x in self.old_middlewares
                                                  if not 'csrf' in x.lower())
        else:
            self.old_middlewares = settings.MIDDLEWARE_CLASSES
        self.login_email = "test@canonical.com"
        self.login_password = "test"

        self.client.login(username=self.login_email,
                          password=self.login_password)

    def tearDown(self):
        settings.MIDDLEWARE_CLASSES = self.old_middlewares


def make_identifier_select_endpoint(protocol_uri):
    """Create an endpoint for use in OpenID identifier select mode.

    :arg protocol_uri: The URI for the OpenID protocol version.  This
        should be one of the OPENID_X_Y_TYPE constants.

    If the OpenID 1.x protocol is selected, the endpoint will be
    suitable for use with Launchpad's non-standard identifier select
    workflow.
    """
    assert protocol_uri in [
        OPENID_1_0_TYPE, OPENID_1_1_TYPE, OPENID_2_0_TYPE], (
        "Unexpected protocol URI: %s" % protocol_uri)

    endpoint = OpenIDServiceEndpoint()
    endpoint.server_url = settings.SSO_PROVIDER_URL
    if protocol_uri == OPENID_2_0_TYPE:
        endpoint.type_uris = [OPENID_IDP_2_0_TYPE]
    else:
        endpoint.type_uris = [protocol_uri]
        endpoint.claimed_id = IDENTIFIER_SELECT
        endpoint.local_id = IDENTIFIER_SELECT
    return endpoint


class MockRequest(HttpRequest):

    def __init__(self, path):
        super(HttpRequest, self).__init__()
        self.path = path

        class MockSession(dict):
            def flush(self):
                pass

        self.session = MockSession()

    def get_full_path(self):
        return self.path


class MockHandler(urllib2.HTTPHandler):

    def set_next_response(self, msg):
        self.next_response = msg

    def http_open(self, req):
        response = StringIO(self.next_response)
        resp = urllib.addinfourl(response, {}, req.get_full_url())
        resp.code = 200
        resp.msg = 'OK'
        return resp


def add_user_to_team(account, teamname, nickname=None):
    """ Note: will not work on the staging or production databases as those are
        replicated from LP and read-only.
    """

    from identityprovider.models.person import Person
    from identityprovider.models.account import LPOpenIdIdentifier
    from identityprovider.models.team import (
        TeamMembership, TeamParticipation)

    memberships = TeamMembership.get_team_memberships_for_user([teamname],
                                                               account, True)
    if len(memberships) > 0:
        return

    p = account.person
    if p is None:
        if nickname is None:
            nickname = get_unique_username()
        p = Person.objects.create(lp_account=account.id,
                                  displayname=account.displayname,
                                  name=nickname)

    try:
        LPOpenIdIdentifier.objects.get(lp_account=account.id)
    except LPOpenIdIdentifier.DoesNotExist:
        LPOpenIdIdentifier.objects.create(
            identifier=account.openid_identifier, lp_account=account.id)

    try:
        t = Person.objects.get(name=teamname)
    except Person.DoesNotExist:
        t = Person.objects.create(displayname="Team %s" % teamname,
                                  teamowner=p, name=teamname)

    now = date.today()
    exp = date(day=now.day, month=now.month, year=(now.year + 1))
    TeamMembership.objects.create(person=p, date_expires=exp, team=t, status=1)
    TeamParticipation.objects.create(team=t, person=p)


def get_unique_username():
    from identityprovider.models.person import Person
    i = 0
    while True:
        try:
            username = "user%d" % i
            i += 1
            Person.objects.get(name=username)
        except Person.DoesNotExist:
            break
    return username


skipOnSqlite = skipIf(connection.vendor == 'sqlite', "Skipping on SQLite")


class AnonAPITestCase(TransactionTestCase, WSGIInterceptedTestCase):
    # Circular import. utils.py is already imported into factory.
    from identityprovider.tests.factory import SSOObjectFactory
    factory = SSOObjectFactory()

    def setUp(self):
        super(AnonAPITestCase, self).setUp()
        callbacks = {
            'default': make_app,
            ('localhost', 8000): make_app,
            }
        self.setup_intercept(callbacks, intercept_api=True)
        self.api = ServiceRoot(None, 'http://localhost:8000/api/1.0')

    def tearDown(self):
        super(AnonAPITestCase, self).tearDown()
        self.teardown_intercept()
