# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import binascii
import hashlib
import random
import re
import urllib
import urllib2
import socket

from email.Header import Header
from email.Utils import formataddr

from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import ugettext as _

SALT_LENGTH = 20


class CannotResetPasswordException(Exception):
    pass


class PersonAndAccountNotFoundException(Exception):
    pass


def canonical_url(
    obj, request=None, rootsite=None, path_only_if_possible=False,
    view_name=None):
    """ Very dumb implementation of canonical_url() """
    from identityprovider.models import Person
    if isinstance(obj, Person):
        if view_name is None:
            return 'https://launchpad.net/~%s' % obj.name
        else:
            return 'https://launchpad.net/~%s/%s' % (obj.name, view_name)
    else:
        return None


def encrypt_launchpad_password(plaintext, salt=None):
    plaintext = str(plaintext)
    if salt is None:
        salt = generate_salt()
    v = binascii.b2a_base64(hashlib.sha1(plaintext + salt).digest() + salt)
    return v[:-1]


def format_address(address, name=None):
    r"""Formats a name and address to be used as an email header.

        >>> format_address('foo@bar.com', 'Name')
        'Name <foo@bar.com>'
        >>> format_address('foo@bar.com', '')
        'foo@bar.com'
        >>> format_address(u'foo@bar.com')
        'foo@bar.com'

    It handles unicode and characters that need quoting as well.

        >>> format_address('foo.bar@canonical.com', u'F\xf4\xf4 Bar')
        '=?utf-8?b?RsO0w7QgQmFy?= <foo.bar@canonical.com>'

        >>> format_address('foo.bar@canonical.com', 'Foo [Baz] Bar')
        '"Foo \\[Baz\\] Bar" <foo.bar@canonical.com>'

    Really long names doesn't get folded, since we're not constructing
    an e-mail header here.

        >>> formatted_address = format_address(
        ...     'long.name@example.com',
        ...     'a '*100)
        >>> '\n' in formatted_address
        False
    """
    if not name:
        return str(address)
    name = str(Header(unicode(name)))
    # Using Header to encode the name has the side-effect that long
    # names are folded, so let's unfold it again.
    name = ''.join(name.splitlines())
    return str(formataddr((name, address)))


def generate_openid_identifier():
    # import here to avoid circular imports
    from django.db import connection
    cursor = connection.cursor()
    cursor.execute("SELECT generate_openid_identifier();")
    row = cursor.fetchone()
    return row[0]


def generate_salt():
    """ From lib/canonical/launchpad/webapp/authentication.py """
    # Salt can be any length, but not more than about 37 characters
    # because of limitations of the binascii module.
    # All 256 characters are available.
    salt = ''
    for n in range(SALT_LENGTH):
        salt += chr(random.randrange(256))
    return salt


def get_person_and_account_by_email(email):
    from identityprovider.models import EmailAddress
    try:
        email = EmailAddress.objects.get(email__iexact=email)
        if email.account is None and email.person is None:
            raise PersonAndAccountNotFoundException()
        elif email.account is None and email.person.is_team():
            raise CannotResetPasswordException(_("The email address "
                "%s can not be used to log in as it belongs "
                "to a team.") % email)
        else:
            return email.person, email.account
    except EmailAddress.DoesNotExist:
        raise PersonAndAccountNotFoundException()


def password_policy_compliant(password):
    if len(password) < 8:
        return False
    has_upper = False
    has_number = False
    has_punctuation = False
    if re.search(r'[A-Z]', password):
        has_upper = True
    if re.search(r'[0-9]', password):
        has_number = True
    search_string = '[\.\?\!\,\-\_\[\]\(\)"#$%&\'\*\+/:;\<\>=\^`\{\}\|~ ]'
    if re.search(search_string, password):
        has_punctuation = True
    return ((int(has_upper) + int(has_number) + int(has_punctuation)) >= 2)


def polite_form_errors(errors):
    if 'email' in errors:
        if errors['email'][0] == _("Enter a valid e-mail address."):
            errors['email'][0] = _("Please enter a valid email address.")


def validate_launchpad_password(plaintext, encrypted):
    encrypted = str(encrypted)
    plaintext = str(plaintext)
    try:
        ref = binascii.a2b_base64(encrypted)
    except binascii.Error:
        # Not valid base64.
        return False
    salt = ref[20:]
    v = binascii.b2a_base64(
        hashlib.sha1(plaintext + salt).digest() + salt)[:-1]
    pw1 = (v or '').strip()
    pw2 = (encrypted or '').strip()
    return pw1 == pw2


def http_request_with_timeout(url, data=None, headers=None, timeout=5):
    if data is not None:
        data = urllib.urlencode(data)
    if headers is None:
        headers = {}
    request = urllib2.Request(url, headers=headers, data=data)
    old_timeout = socket.getdefaulttimeout()
    try:
        socket.setdefaulttimeout(timeout)
        datafile = urllib2.urlopen(request)
        data = datafile.read()
        headers = datafile.info()
    except urllib2.URLError:
        data, headers = None, None
    finally:
        socket.setdefaulttimeout(old_timeout)
    return data, headers


def get_object_or_none(qs_or_model, **kwargs):
    if hasattr(qs_or_model, 'objects'):
        qs_or_model = qs_or_model.objects.all()
    try:
        return qs_or_model.get(**kwargs)
    except ObjectDoesNotExist:
        return None
