# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from openid.yadis.constants import YADIS_HEADER_NAME, YADIS_CONTENT_TYPE

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.http import HttpResponseRedirect
from django.shortcuts import render_to_response, get_object_or_404
from django.template import RequestContext
from django.utils.translation import ugettext as _
from django.views.decorators.vary import vary_on_headers

from identityprovider.cookies import set_test_cookie, test_cookie_worked
from identityprovider.decorators import check_readonly
from identityprovider.forms import EditAccountForm, LoginForm, NewEmailForm
from identityprovider.models import (send_validation_email_request,
    EmailAddress)
from oauth_backend.models import Token
from identityprovider.models.const import EmailStatus
from identityprovider.views.utils import (redirection_url_for_token,
    set_session_token_info)
from identityprovider.signals import account_email_added

from identityprovider.views.server import xrds
from identityprovider.views.ui import display_email_sent

@vary_on_headers('Accept')
def index(request, token=None):
    accept = request.META.get('HTTP_ACCEPT', '')
    if YADIS_CONTENT_TYPE in accept:
        return xrds(request)

    request.token = token
    context = RequestContext(request)
    if request.user.is_authenticated():
        resp = account_index(request, token)
    else:
        context = RequestContext(request, {
            'form': LoginForm(),
        })
        resp = render_to_response('registration/login.html', context)
        # Perhaps the browser sent us the test cookie, or perhaps it
        # didn't.  We'll tell Django that we want one, and it will be
        # intelligent and not send it out again if the browser *did*
        # send it to us.
        set_test_cookie(resp)
    resp[YADIS_HEADER_NAME] = '%s+xrds' % settings.SSO_ROOT_URL
    return resp


def cookie(request):
    next = request.REQUEST.get('next', '/')
    if test_cookie_worked(request):
        return HttpResponseRedirect(next)
    # Send the cookie again, so that the user can hit reload on the
    # page we return to continue on.
    return set_test_cookie(render_to_response('cookies.html'))


@login_required
def account_index(request, token=None):
    request.token = token
    account = request.user
    if request.method == 'POST' and not settings.READ_ONLY_MODE:
        form = EditAccountForm(request.POST, account=account)
        if form.save_account():
            # Translators: Displayed as an in-line message after update (/)
            request.session['message'] = _("Your account details have been "\
                                           "successfully updated")
            return HttpResponseRedirect('.')
    else:
        form = EditAccountForm(account=account)

    context = RequestContext(request, {
        'current_section': 'account',
        'account': account,
        'account_displayname': account.displayname,
        'last_authenticated_sites': account.last_authenticated_sites(limit=10),
        'unverified_emails': account.unverified_emails(),
        'form': form,
        'message': request.session.pop('message', None),
        'message_style': 'informational',
        'token': token,
    })
    return render_to_response('account/edit.html', context)


@login_required
def account_emails(request):
    account = request.user
    form = NewEmailForm()
    context = RequestContext(request, {
        'current_section': 'emails',
        'account': account,
        'account_displayname': account.displayname,
        'verified_emails': account.verified_emails(),
        'unverified_emails': account.unverified_emails(),
        'form': form,
        'message': request.session.pop('message', None),
        'message_style': 'informational'
    })
    return render_to_response('account/emails.html', context)


@login_required
def account_deactivate(request, token=None):
    request.token = token
    import django.contrib.auth as auth
    from identityprovider.models.const import AccountStatus

    account = request.user
    for email in account.emailaddress_set.all():
        email.status = EmailStatus.NEW
        email.save()
    account.status = AccountStatus.DEACTIVATED
    account.save()

    # We don't want to lose session[token] when we log the user out
    raw_orequest = request.session.get(token, None)
    auth.logout(request)
    if token is not None and raw_orequest is not None:
        request.session[token] = raw_orequest
    return HttpResponseRedirect('/+deactivated')


def _send_verification_email(request, email, tokenid=None):
    # TODO: This makes use of `tokenid` if it's passed in, but the
    # call to `send_validation_email_request` always generates a new
    # token.  Is this right?

    # If there are any unverified emails that match they should be deleted.
    EmailAddress.objects.filter(email__iexact=email,
        status=EmailStatus.NEW).delete()

    # Ensure that this account has such email address; return value is not used
    request.user.emailaddress_set.get_or_create(
        email=email, lp_person=request.user.person, status=EmailStatus.NEW)

    redirection_url = redirection_url_for_token(tokenid)

    token = send_validation_email_request(request.user, email, redirection_url)
    set_session_token_info(request.session, token)

    return display_email_sent(request,
        _("Validate your email address"),
        _("We&rsquo;ve just "
           "emailed %(email_to)s (from %(email_from)s) "
           "with instructions on validating your email address.") % {
            'email_to': email,
            'email_from': settings.NOREPLY_FROM_ADDRESS}
    )


@login_required
@check_readonly
def new_email(request, token=None, emailid=None):
    request.token = token
    form = NewEmailForm()

    if request.method == 'POST' and not settings.READ_ONLY_MODE:
        form = NewEmailForm(request.POST)
        if form.is_valid():
            email = form.cleaned_data['newemail']
            account_email_added.send(
                openid_identifier=request.user.openid_identifier,
                sender=None)
            return _send_verification_email(request, email, token)

    context = RequestContext(request, {'form': form})
    return render_to_response('account/new_email.html', context)


@login_required
def verify_email(request, token=None):
    request.token = token
    emailid = request.GET.get('id', 0)
    email = get_object_or_404(request.user.emailaddress_set, pk=emailid,
                              status=EmailStatus.NEW)
    return _send_verification_email(request, email.email, token)


@login_required
def delete_email(request, token=None):
    request.token = token
    emailid = request.GET.get('id', 0)
    email = get_object_or_404(
        request.user.emailaddress_set,
        pk=emailid,
        status__in=[EmailStatus.NEW, EmailStatus.VALIDATED])
    if request.method == 'POST':
        redirection_url = token and redirection_url_for_token(token) or '/+emails'
        request.session['message'] = _("The unverified email address %s was "
            "removed successfully") % email.email
        email.delete()
        return HttpResponseRedirect(redirection_url)
    else:
        context = RequestContext(request, {
            'email': email.email,
        })
        return render_to_response('account/delete_email.html', context)


@login_required
@check_readonly
def applications(request):
    if request.method == 'POST':
        token_id = request.POST.get('token_id')
        if token_id:
            try:
                token = request.user.oauth_tokens().get(pk=token_id)
                message = _("'%s' token was revoked successfully") % token.name
                token.delete()
                request.session['message'] = message
            except (AttributeError, Token.DoesNotExist):
                pass
        return HttpResponseRedirect('/+applications')
    else:
        tokens = request.user.oauth_tokens()
        context = RequestContext(request, {
            'current_section': 'applications',
            'tokens': tokens,
            'message': request.session.pop('message', None),
            'message_style': 'informational',
            })
        return render_to_response('account/applications.html', context)



