# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from functools import wraps

from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseNotAllowed, HttpResponse
from django.shortcuts import render_to_response
from django.template import RequestContext

from gargoyle.decorators import switch_is_active

from .utils import HttpResponseSeeOther, allow_only


def context(request, **kwargs):
    kwargs['current_section'] = 'devices'
    return RequestContext(request, kwargs)


# A decorator
require_twofactor = switch_is_active('TWOFACTOR')


@login_required
@require_twofactor
@allow_only('GET')
def device_list(request):
    return render_to_response('device/list.html', context(request))


@login_required
@require_twofactor
@allow_only('GET', 'POST')
def device_addition(request):

    # I dislike having a single view handle more than two standard
    # (display form, respond to form) steps.  This one is driven by
    # two requiredments: 1) adding *any* device starts at the same
    # point on the device list screen; and, 2) creating a one-time pad
    # might require no confirmation from the user, and so require a
    # POST as the first request from selecting the device-type.
    #
    # In other words, the device-type selection form must POST the
    # user's selection (even though most the the time no modifications
    # will be made as a result), and the handler cannot redirect in
    # the case of "paper", but must handle it directly (at least when
    # no other one-time pad exists).

    if request.method == 'GET':
        return render_to_response('device/types.html', context(request))

    device_type = request.POST.get('type')

    # For devices that generate a key:
    # Display key form
    # [post] add to db, redir to verify

    # For devices that accept a key:
    # Generate a key, add to db, display key page
    # [get, verify] (already redir'd to verify)

    if device_type == 'paper':
        already_has_pad = True
        verified_creation = (request.POST.get('verified', '0') == '1')

        if already_has_pad and not verified_creation:
            return render_to_response('device/pad-verification.html',
                                      context(request))

        return render_to_response('device/pad-printout.html', context(request))

    if device_type not in ('yubi', 'google', 'generic'):
        return render_to_response('device/types.html', context(request))

    generated_key = '1234'

    # Google would base32-encode, yubi would hex-encode, etc.:
    formatted_key = '1234'

    return render_to_response('device/addition-%s.html' % device_type,
        context(request, formatted_key=formatted_key))


@login_required
@require_twofactor
@allow_only('GET', 'POST')
def device_verification(request, device_id):
    # TODO: Verify that the logged-in user actually owns the device
    # with this ID.

    if request.method == 'GET':
        return render_to_response('device/verification.html', context(request))

    # TODO: Finish the UI mock here...
    pass


@login_required
@require_twofactor
@allow_only('GET', 'POST')
def device_removal(request, device_id):
    # TODO: Verify that the logged-in user actually owns the device
    # with this ID.

    if request.method == 'GET':
        return render_to_response('device/removal.html', context(request))

    # We should probably send an e-mail to the user stating which
    # device was removed.  As a security measure, this would be much
    # stronger if bugs #784813, #784817, and #784818 were done.
    return HttpResponseSeeOther('/device-list')


@login_required
@require_twofactor
@allow_only('GET', 'POST')
def pad_removal(request):
    if request.method == 'GET':
        return render_to_response('device/pad-removal.html', context(request))

    return HttpResponseSeeOther('/device-list')
