# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import socket
import urllib
import urllib2
from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponse, HttpResponseRedirect, Http404
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.utils import simplejson
from django.views.decorators.csrf import csrf_exempt
from identityprovider.readonly import ReadOnlyManager


def lowercase_keys(dicts):
    """ Convert all keys in a list of dicts to lowercase. """
    return [dict((k.lower(), v) for (k, v) in d.items()) for d in dicts]

def _remote_req(host, port=None, scheme=None, server_id=None,
                virtual_host=None, post=None):
    """Makes a request to a specific appserver.

    The first five arguments are the same as the keys for each dictionary in
    the APP_SERVERS setting:

    * host: The host at which we can reach the app server (can be an IP)
    * port: The port that should be used (optional)
    * scheme: The scheme to use to contact this app server (http/https)
              (defaults to http)
    * server_id: Some canonical name for this app server (optional)
    * virtual_host: The virtual host that should be used, for app servers that
                    serve multiple sites (optional).

    If post is provided, it should be a sequence of 2-tuples that will be
    encoded in to POST data.
    """
    if post is None:
        post = []
    post.append(('secret', settings.READONLY_SECRET))
    post = urllib.urlencode(post)

    if scheme is None:
        scheme = 'http'
    if port is None:
        portstr = ''
    else:
        portstr = ':' + port
    url = '%s://%s%s/readonlydata' % (scheme, host, portstr)
    headers = {}
    if virtual_host is not None:
        headers['Host'] = virtual_host
    req = urllib2.Request(url, headers=headers, data=post)
    try:
        oldtimeout = socket.getdefaulttimeout()
        socket.setdefaulttimeout(5)
        datafile = urllib2.urlopen(req)
        data = datafile.read()
        socket.setdefaulttimeout(oldtimeout)
    except urllib2.URLError:
        data = None
    return data

@staff_member_required
def readonly_admin(request):
    atts = get_server_atts(settings.APP_SERVERS)
    return render_to_response('admin/readonly.html', atts)

def get_server_atts(servers):
    """Provides a report about readonly status of all app servers."""
    appservers = []
    set_all_readonly = False
    clear_all_readonly = False
    for server in lowercase_keys(servers):
        datastr = _remote_req(**server)
        if datastr is None:
            data = {'reachable': False}
        else:
            try:
                data = simplejson.loads(datastr)
            except ValueError:
                # This is probably caused by the remote database being out
                # of sync with our current database
                data = {'reachable': False}
            else:
                data['reachable'] = True
            if data.get('readonly'):
                clear_all_readonly = True
            else:
                set_all_readonly = True
        data['name'] = server['server_id']
        appservers.append(data)
    atts = {'appservers': appservers,
            'admin_media_prefix': settings.ADMIN_MEDIA_PREFIX,
            'clear_all_readonly': clear_all_readonly,
            'set_all_readonly': set_all_readonly,
            }
    return atts

@staff_member_required
def readonly_confirm(request, action, appserver=None, conn=None):
    if request.method == 'POST':
        update_server(action, appserver, conn)
        return HttpResponseRedirect('/readonly')
    context = RequestContext(request, {
        'appserver': appserver,
        'action': action,
        'conn': conn,
    })
    return render_to_response('admin/readonly_confirm.html', context)

def update_server(action, appserver=None, conn=None):
    if appserver is None:
        appservers = settings.APP_SERVERS
    else:
        appservers = [server for server in settings.APP_SERVERS
                             if server['SERVER_ID'] == appserver]
    appservers = lowercase_keys(appservers)
    for server in appservers:
        post = [('action', action),
                ('conn', conn),
               ]
        _remote_req(post=post, **server)

@csrf_exempt
def readonly_data(request):
    """Provides data about the readonly status of this app server."""
    if request.method != 'POST':
        raise Http404()
    secret = request.POST.get('secret')
    if secret != settings.READONLY_SECRET:
        raise Http404()
    action = request.POST.get('action')
    conn = request.POST.get('conn')
    romanager = ReadOnlyManager()
    if action == 'set':
        romanager.set_readonly()
    elif action == 'clear':
        romanager.clear_readonly()
    elif action == 'enable':
        romanager.clear_failed(conn)
    elif action == 'disable':
        romanager.mark_failed(conn)
    result = {'readonly': settings.READ_ONLY_MODE,
              'automatic': romanager.current_readonly_is_automatic(),
              'next_recovery_due': romanager.next_recovery_due(),
             }
    dbs = []
    for conn in romanager.connections:
        dbid = conn['DATABASE_ID']
        dbs.append({'id': dbid,
                    'failed': romanager.is_failed(dbid),
                   })
    result['connections'] = dbs
    return HttpResponse(simplejson.dumps(result))
