# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import logging

from urlparse import urljoin

import django.contrib.auth as auth

from django.conf import settings
from django.core import urlresolvers
from django.core.mail import send_mail
from django.core.urlresolvers import Resolver404, resolve, reverse
from django.http import (Http404, HttpResponseNotAllowed, HttpResponseNotFound,
                         HttpResponseRedirect, urlencode)
from django.shortcuts import (get_object_or_404, get_list_or_404,
                              render_to_response)
from django.template import RequestContext
from django.template.loader import render_to_string
from django.utils.http import urlquote
from django.utils.translation import ugettext as _

from identityprovider.branding import current_brand
from identityprovider.decorators import (check_readonly, dont_cache,
    guest_required, limitlogin, requires_cookies)
from identityprovider.forms import (
    LoginForm, ForgotPasswordForm, ResetPasswordForm, OldNewAccountForm,
    NewAccountForm, ConfirmNewAccountForm, TokenForm)
from identityprovider.models import (Account, AccountPassword, AuthToken,
    AuthTokenFactory, EmailAddress, get_type_of_token, verify_token_string)
from identityprovider.models.authtoken import InvalidEmailException
from identityprovider.models.captcha import Captcha, VerifyCaptchaError
from identityprovider.models.const import (AccountStatus, EmailStatus,
    LoginTokenType)
from identityprovider.models.openidmodels import OpenIDRPConfig
import identityprovider.signed as signed
from identityprovider.utils import (
    CannotResetPasswordException,
    PersonAndAccountNotFoundException,
    encrypt_launchpad_password,
    format_address,
    get_person_and_account_by_email,
    polite_form_errors,
)
from identityprovider.stats import stats
from identityprovider.views.utils import get_rpconfig, set_session_token_info



logger = logging.getLogger('sso')


def _get_rpconfig_from_request(request, token):
    rpconfig = None
    if token: # won't have an rpconfig without a token
        raw_orequest = request.session.get(token, None)
        if raw_orequest: # is this an actual openid request?
            orequest = signed.loads(raw_orequest, settings.SECRET_KEY)
            rpconfig = get_rpconfig(orequest.trust_root)

    return rpconfig


@guest_required
@dont_cache
@limitlogin()
@requires_cookies
def login(request, token=None, rpconfig=None):
    request.token = token
    form = LoginForm()

    # The rpconfig is not passed in when there was a form validation error
    # So if not present we'll fetch it again to display RP information
    if not rpconfig:
        rpconfig = _get_rpconfig_from_request(request, token)
    if request.method == 'POST':
        form = LoginForm(request.POST)
        if form.is_valid():
            account = form.cleaned_data['account']
            auth.login(request, account)

            # track successful logins
            stats.increment('flows.login', key='success', rpconfig=rpconfig)

            next = request.POST.get('next')
            limitlogin().reset_count(request)
            if next and _is_safe_redirect_url(next):
                return HttpResponseRedirect(next)
            elif token:
                return HttpResponseRedirect('/%s/' % token)
            else:
                return HttpResponseRedirect('/')
        else:
            # track login form errors
            stats.increment('flows.login', key='error', rpconfig=rpconfig)

            polite_form_errors(form._errors)
    else:
        # track login attempts
        stats.increment('flows.login', key='requested', rpconfig=rpconfig)

    context = RequestContext(request, {
        'form': form,
        'next': request.GET.get('next'),
        'rpconfig': rpconfig,
        'token': token is not None,
        'message': request.session.get('message', None),
        'message_style': 'informational',
    })
    try:
        del request.session['message']
    except:
        pass
    return render_to_response('registration/login.html', context)


def _is_safe_redirect_url(url):
    """ Check whether 'url' resolves into a valid URL this application can
    serve. """
    try:
        if url.find('?') >= 0:
            url = url[:url.find('?')]
        response = urlresolvers.resolve(url)
    except Resolver404:
        return False
    if response is None:
        return False
    view, args, kwargs = response
    return True


class LogoutView(object):
    def get_language(self, user):
        try:
            return user.preferredlanguage
        except AttributeError:
            return None

    def set_language(self, response, language):
        if language:
            response.set_cookie(settings.LANGUAGE_COOKIE_NAME, language)

    def get_return_to_url(self, return_to, referer):
        rpconfig = OpenIDRPConfig.objects.for_url(return_to)
        if not return_to or rpconfig is None:
            return None
        elif referer is None or referer.startswith(rpconfig.trust_root):
            return return_to
        else:
            return None

    def set_orequest(self, session, token, raw_orequest):
        if token is not None and raw_orequest is not None:
            session[token] = raw_orequest

    def get_other_sites(self, user, trust_root):
        if not hasattr(user, 'sites_with_active_sessions'):
            return []
        return [site for site in user.sites_with_active_sessions()
                if trust_root != site.trust_root]

    def get_site_name(self, trust_root):
        rpconfig = OpenIDRPConfig.objects.for_url(trust_root)
        if not trust_root:
            return None
        elif rpconfig is None:
            return None
        elif rpconfig.displayname:
            return rpconfig.displayname
        else:
            return trust_root

    def compute_context(self, user, return_to, referrer):
        return_to = self.get_return_to_url(return_to, referrer)

        context = {
            'return_to_url': return_to,
            'return_to_site_name': self.get_site_name(return_to),
            'other_sites': self.get_other_sites(user, return_to),
        }
        return context

    def __call__(self, request, token):
        lang = self.get_language(request.user)

        context = self.compute_context(
            request.user,
            request.GET.get('return_to', None),
            request.META.get('HTTP_REFERER', None))

        # We don't want to lose session[token] when we log the user out
        raw_orequest = request.session.get(token, None)
        auth.logout(request)
        self.set_orequest(request.session, token, raw_orequest)

        template_file = ('%s/registration/logout.html' %
                         current_brand.template_dir)
        response = render_to_response(template_file,
                                      RequestContext(request, context))

        self.set_language(response, lang)
        return response


logout = LogoutView()


def enter_token(request, token=None):
    request.token = token
    confirmation_code = request.REQUEST.get('confirmation_code')
    email = request.REQUEST.get('email') or request.session.get('token_email')
    params = {'confirmation_code': confirmation_code, 'email': email}
    if request.method == 'GET':
        form = TokenForm(initial=params)
    elif request.method == 'POST':
        form = TokenForm(params)
        if form.is_valid():
            confirmation = form.cleaned_data['confirmation']
            return _handle_confirmation(
                confirmation.token_type, confirmation_code, email, request,
                token=token)
    else:
        return HttpResponseNotAllowed(['GET', 'POST'])
    context = RequestContext(request, {
        'form': form,
        'token': token
        })
    return render_to_response('enter_token.html', context)


def _redirect_to_enter_token(confirmation_code=None, email=None):
    params = {}
    if confirmation_code is not None:
        params['confirmation_code'] = confirmation_code
    if email is not None:
        params['email'] = email
    return HttpResponseRedirect('/+enter_token?%s' % urlencode(params))


def claim_token(request, authtoken):
    email = request.GET.get('email') or request.session.get('token_email')
    if email is None:
        return _redirect_to_enter_token(confirmation_code=authtoken)
    token_type = get_type_of_token(authtoken)
    return _handle_confirmation(token_type, authtoken, email, request)


# As soon as _finish_account_creation doesn't need `request`, neither
# does this.
def _handle_confirmation(confirmation_type, confirmation_code, email, request,
                         token=None):
    """Handles a confirmation, either by finishing the work, or
    redirecting to a page that can.  Also accepts an optional
    OpenID-transaction token."""

    args = {
        'authtoken': confirmation_code,
        'email_address': email
        }
    if token is not None:
        args['token'] = token

    if confirmation_type == LoginTokenType.PASSWORDRECOVERY:
        view = reset_password
    elif confirmation_type == LoginTokenType.NEWPERSONLESSACCOUNT:
        # This is a slight inefficiency for now, as the caller may
        # already have retrieved the confirmation code from the
        # database.
        #
        # Also, all this code *really* wants to be a method on the
        # confirmation code... the objects are crying right now.
        confirmation = verify_token_string(confirmation_code, email)

        if confirmation.displayname is not None and \
               confirmation.password is not None:
            user, redirection_url = _finish_account_creation(
                confirmation, confirmation.displayname, confirmation.password,
                request)
            auth.login(request, user)
            request.session['message'] = \
                _("Your account was created successfully")
            return HttpResponseRedirect(redirection_url)

        view = old_confirm_account
        if 'token' in args:
            del args['token']

    elif confirmation_type == LoginTokenType.VALIDATEEMAIL:
        view = confirm_email
    else:
        msg = 'Unknown type %s for confirmation code "%s"' % \
                     (confirmation_type, confirmation_code)
        logger.error(msg)
        raise Http404(msg)

    url = reverse(view, kwargs=args)
    return HttpResponseRedirect(url)


def _prepare_account_creation(displayname, email, encrypted_password,
                              redirection_url):
    try:
        person, account = get_person_and_account_by_email(email)
        should_create = False
    except (CannotResetPasswordException,
            PersonAndAccountNotFoundException):
        should_create = True

    if should_create:
        token = AuthTokenFactory().new(
            requester=None, requester_email=None, email=email,
            displayname=displayname, password=encrypted_password,
            token_type=LoginTokenType.NEWPERSONLESSACCOUNT,
            redirection_url=redirection_url)
        return token

    def log_no_email(condition):
        logger.debug("In view 'new_account' email was not sent "
                     "out because %s" % condition)

    # Only send email if the account is active; otherwise, a disabled
    # account can be spammed.
    if account is None:
        log_no_email("account for person '%s' does not exist" %
                     person.displayname)
    elif not account.is_active:
        log_no_email("account '%s' is not active" % account.displayname)
    else:
        replacements = {
            'forgotten_password_url':
            urljoin(settings.SSO_ROOT_URL, '+forgot_password'),
        }
        message = render_to_string(
            '%s/email/impersonate-warning.txt' %
            current_brand.template_dir, replacements)
        subject = current_brand.name + ": " + _("Warning")
        from_address = format_address(
            settings.NOREPLY_FROM_ADDRESS, current_brand.name)
        send_mail(subject, message, from_address, [email])


# This accepts displayname and password as separate parameters for
# now.  After removing the old confirmation method, this function can
# pull those params directly from atrequest.
#
# This function is currently called from two places, and at each place
# is followed by the exact same code to log the user in and setup some
# session stuff.  I've decided to not include that code in this
# function, to keep this function isolated from the view layer and
# easy to test.  ALTHOUGH, it does currently have other view
# dependencies (for determining the creation rationale), which
# *should* also be moved out.  Maybe rationale should be determined
# when the token is created, and stored?
def _finish_account_creation(atrequest, displayname, password, request):
    """Creates the account according to the given token.  Returns the
    new user, already authorized, and the URL to redirect to."""
    #displayname = atrequest.displayname
    username = atrequest.email
    #password = atrequest.encrypted_password
    creation_rationale = None
    if atrequest.redirection_url is None:
        atrequest.redirection_url = '/'
    if atrequest.redirection_url.endswith('+decide'):
        # Try to determine creation rationale from RP config
        view, args, kwargs = resolve(atrequest.redirection_url)
        if 'token' in kwargs:
            try:
                raw_orequest = request.session.get(
                    kwargs['token'], None)
                openid_request = signed.loads(
                    raw_orequest, settings.SECRET_KEY)
                rpconfig = get_rpconfig(openid_request.trust_root)
                if rpconfig is not None:
                    creation_rationale = rpconfig.creation_rationale
            except Exception:
                pass
    # test if account has already been created
    account = Account.objects.get_by_email(username)
    if account is not None:
        # account already created, just needs confirmation
        email = EmailAddress.objects.get(email=username, account=account)
        email.status = EmailStatus.PREFERRED
        email.save()
    else:
        # account has not yet been created. Let's create it
        account = Account.objects.create_account(
            displayname, username, password, creation_rationale,
            password_encrypted=True)
    atrequest.consume()
    user = auth.authenticate(username=username, encrypted_password=password)
    return user, atrequest.redirection_url


@guest_required
@check_readonly
@requires_cookies
def new_account(request, token=None):
    request.token = token

    # Need this to display RP info when coming from a consumer
    rpconfig = _get_rpconfig_from_request(request, token)

    # {{workflow}}
    old = 'old' in request.REQUEST
    form_class = OldNewAccountForm if old else NewAccountForm

    if request.method == 'GET':
        # track number of registration requests by consumer
        stats.increment('flows.new_account', key='requested',
            rpconfig=rpconfig)

        if 'email' in request.GET:
            form = form_class(initial={'email': request.GET['email']})
        else:
            form = form_class()
    elif request.method == 'POST':
        form = form_class(request.POST)
        if form.is_valid():

            response = _verify_captcha_response(
                'registration/new_account.html', request, form)
            if response:
                # track number of captcha verification failures
                stats.increment('flows.new_account', key='error.captcha',
                    rpconfig=rpconfig)
                return response

            # {{workflow}} All these fields will always be available
            # by this point.
            displayname = form.cleaned_data.get('displayname')
            email = form.cleaned_data['email']
            encrypted_password = encrypt_launchpad_password(
                form.cleaned_data['password']) \
                if 'password' in form.cleaned_data else None

            if token is not None:
                # Redirected from an OpenID request, must get back
                # after confirmation
                redirection_url = '/%s/+decide' % token
            else:
                redirection_url = '/'

            atoken = None
            try:
                atoken = _prepare_account_creation(
                    displayname, email, encrypted_password, redirection_url)
            except InvalidEmailException:
                form.errors['email'] = 'Invalid email'
                polite_form_errors(form._errors)
                # track number of invalid email errors
                stats.increment('flows.new_account', key='error.email',
                    rpconfig=rpconfig)

            if atoken is not None:
                atoken.sendNewUserEmail(old=old)
                set_session_token_info(request.session, atoken)

            return display_email_sent(request,
                _("Registration mail sent"),
                _("We&rsquo;ve just emailed "
                "%(email_to)s (from %(email_from)s) to confirm "
                "your address.") % {
                    'email_to': email,
                    'email_from': settings.NOREPLY_FROM_ADDRESS },
                token = token
                )
        else:
            polite_form_errors(form._errors)
            # track number of form errors
            stats.increment('flows.new_account', key='error.form',
                rpconfig=rpconfig)

    context = RequestContext(request, {
        'form': form,
        'CAPTCHA_PUBLIC_KEY': settings.CAPTCHA_PUBLIC_KEY,
        'CAPTCHA_API_URL_SECURE': settings.CAPTCHA_API_URL_SECURE,
        # {{workflow}}
        'old': old,
        'rpconfig': rpconfig,
        })
    return render_to_response('registration/new_account.html', context)


def _verify_captcha_response(template, request, form):
    captcha = Captcha(request.POST.get('recaptcha_challenge_field'))
    captcha_solution = request.POST.get('recaptcha_response_field')
    email = request.POST.get('email', '')
    ip_addr = request.environ["REMOTE_ADDR"]
    try:
        verified = captcha.verify(captcha_solution, ip_addr, email)
        if verified:
            return None
    except VerifyCaptchaError:
        logging.exception("reCaptcha connection error")

    # not verified
    return render_to_response(
        template,
        RequestContext(request, {
            'form': form,
            'CAPTCHA_PUBLIC_KEY': settings.CAPTCHA_PUBLIC_KEY,
            'captcha_error': ('&error=%s' % captcha.message)}))


# {{workflow}}
@check_readonly
def old_confirm_account(request, authtoken, email_address, token=None):
    request.token = token
    if request.user.is_authenticated():
        return HttpResponseRedirect('/+logout-to-confirm')

    atrequest = verify_token_string(authtoken, email_address)
    if (atrequest is None or
        atrequest.token_type != LoginTokenType.NEWPERSONLESSACCOUNT):
        return HttpResponseRedirect('/+bad-token')

    if request.method == 'GET':
        form = ConfirmNewAccountForm()
    elif request.method == 'POST':
        form = ConfirmNewAccountForm(request.POST)
        if form.is_valid():
            displayname = form.cleaned_data['displayname']
            password = encrypt_launchpad_password(form.cleaned_data['password'])
            user, redirection_url = _finish_account_creation(
                atrequest, displayname, password, request)
            auth.login(request, user)
            request.session['message'] = _("Your account was created "\
                                           "successfully")
            return HttpResponseRedirect(redirection_url)
    return render_to_response('registration/old_confirm_new_account.html',
        RequestContext(request, {'form': form, 'token': token}))


@check_readonly
def confirm_account(request, authtoken, email_address, token=None):
    request.token = token

    if request.user.is_authenticated():
        return HttpResponseRedirect('/+logout-to-confirm')

    # Need this for displaying metrics per RP
    rpconfig = _get_rpconfig_from_request(request, token)

    atrequest = verify_token_string(authtoken, email_address)
    if (atrequest is None or
        atrequest.token_type != LoginTokenType.NEWPERSONLESSACCOUNT):
        # track wrong tokens being used
        stats.increment('flows.new_account', key='error.token',
            rpconfig=rpconfig)
        return HttpResponseRedirect('/+bad-token')

    if request.method == 'GET':
        return render_to_response('registration/confirm_new_account.html',
            RequestContext(request, {'email': email_address, 'token': token}))

    if request.method != 'POST':
        return HttpResponseNotAllowed(['GET', 'POST'])

    user, redirection_url = _finish_account_creation(
        atrequest, atrequest.displayname, atrequest.password, request)
    auth.login(request, user)
    request.session['message'] = _("Your account was created successfully")

    # track number of completed registrations by consumer
    stats.increment('flows.new_account', key='success', rpconfig=rpconfig)

    return HttpResponseRedirect(redirection_url)


def confirm_email(request, authtoken, email_address, token=None):
    request.token = token
    request.session.pop('message', None)
    if not request.user.is_authenticated():
        request.session['message'] = _('Please log in to use this '\
                                       'confirmation code')
        next = urlquote(request.get_full_path())
        return HttpResponseRedirect('%s?next=%s' % (settings.LOGIN_URL, next))
    atrequest = verify_token_string(authtoken, email_address)
    if (atrequest is None or
        atrequest.token_type != LoginTokenType.VALIDATEEMAIL):
        return HttpResponseRedirect('/+bad-token')

    email = get_object_or_404(EmailAddress, email__iexact=atrequest.email)
    if request.user.id != email.account.id:
        # The user is authenticated to a different account.
        # Potentially, the token was leaked or intercepted.  Let's
        # delete it just in case.  The real user can generate another
        # token easily enough.
        atrequest.delete()
        raise Http404

    if request.method == 'POST':
        # only confirm the email address if the form was submitted
        email.status = EmailStatus.VALIDATED
        email.save()
        atrequest.consume()
        request.session['message'] = _("The email address %s has been "
            "validated" % email)
        return HttpResponseRedirect(atrequest.redirection_url)

    # form was not submitted
    return render_to_response('account/confirm_new_email.html',
        RequestContext(request, {'email': email.email, 'token': token}))


def token_from_email(request, email):
    if getattr(settings, 'ENABLE_TOKEN_DEBUG', False):
        tokens = get_list_or_404(AuthToken, email=email)
        return render_to_response('registration/token.html',
            RequestContext(request, {'token': tokens[0].token}))
    else:
        return HttpResponseNotFound()


def display_email_sent(request, heading, reason, extra=None, token=None):
    context = RequestContext(request, {
        'email_feedback': settings.FEEDBACK_TO_ADDRESS,
        'email_heading': heading,
        'email_reason': reason,
        'email_notreceived_extra': extra,
        'token': token,
    })
    return render_to_response('registration/email_sent.html', context)


def bad_token(request):
    context = RequestContext(request, {
        'email_feedback': settings.FEEDBACK_TO_ADDRESS,
    })
    return render_to_response('registration/bad_token.html', context)


def logout_to_confirm(request):
    return render_to_response('registration/logout_to_confirm.html',
        RequestContext(request))


def deactivated(request):
    context = RequestContext(request, {
        'message': request.session.pop('message', None),
        'message_style': 'informational',
    })
    return render_to_response('account/deactivated.html', context)


def suspended(request):
    context = RequestContext(request, {
        'message': request.session.pop('message', None),
        'message_style': 'informational',
    })
    return render_to_response('account/suspended.html', context)


def _handle_reset_request(request, email, token):
    # options for actions to be taken
    NONE, RESET, CREATE = range(3)
    # assume invalid
    send_email = NONE
    # looks like we need this complexity below because it seems the correct
    # action can by indicated by both exceptions AND flags :/
    try:
        person, account = get_person_and_account_by_email(email)
        if account:
            if account.can_reset_password:
                send_email = RESET
            else:
                send_email = NONE
                # log why email was not sent
                condition = ("account '%s' is not active" %
                            account.displayname)
                logger.debug("In view 'forgot_password' email was not "
                            "sent out because %s" % condition)
        else:
            send_email = CREATE
    except CannotResetPasswordException:
        send_email = NONE
    except PersonAndAccountNotFoundException:
        send_email = CREATE

    # we should know now the action we need to take
    if send_email == RESET:
        # send reset email
        if token is not None:
            # Redirected from an OpenID request, must get back after
            # confirmation
            redirection_url = reverse(
                'identityprovider.views.server.decide',
                kwargs={'token': token})
        else:
            redirection_url = '/'
        authtoken = AuthTokenFactory().new(
            account, email, email, LoginTokenType.PASSWORDRECOVERY,
            redirection_url=redirection_url)
        authtoken.sendPasswordResetEmail()
        # TODO: remove this if pos?
        set_session_token_info(request.session, authtoken)
    elif send_email == CREATE:
        # they've tried to reset with an invalid email, so send them an email
        # on how to create an account
        brand = unicode(current_brand.name)
        subject = u"%s: %s" % (brand, u"Password reset request")
        url = urljoin(settings.SSO_ROOT_URL, reverse(new_account))
        context = { 'email': email, 'signup': url }
        tmpl = '%s/email/invitation.txt' % current_brand.template_dir
        body = render_to_string(tmpl, context)
        send_mail(subject, body, settings.NOREPLY_FROM_ADDRESS, [email])


@guest_required
@check_readonly
@requires_cookies
def forgot_password(request, token=None):
    # Need this for displaying metrics per RP
    rpconfig = _get_rpconfig_from_request(request, token)

    request.token = token
    if request.method == 'GET':
        # track forgot password requests
        stats.increment('flows.forgot_password', key='requested',
            rpconfig=rpconfig)

        if 'email' in request.GET:
            form = ForgotPasswordForm(initial={'email': request.GET['email']})
        else:
            form = ForgotPasswordForm()
    elif request.method == 'POST':
        form = ForgotPasswordForm(request.POST)
        if form.is_valid():
            response = _verify_captcha_response(
                'registration/forgot_password.html', request, form)
            if response:
                # track captcha errors
                stats.increment('flows.forgot_password', key='error.captcha',
                    rpconfig=rpconfig)
                return response

            email = form.cleaned_data['email']

            # handle the reset request
            _handle_reset_request(request, email, token)

            #regardless of result, we always display the same information back
            # to the user
            return display_email_sent(request,
                _("Forgotten your password?"),
                _("We&rsquo;ve just emailed "
                  "%(email_to)s (from %(email_from)s) with "
                  "instructions on resetting your password.") % {
                    'email_to': email,
                    'email_from': settings.NOREPLY_FROM_ADDRESS,
                    },
                _("Check that you&rsquo;ve actually "
                   "entered a subscribed email address."),
                token=token
                )
        else:
            # track form errors
            stats.increment('flows.forgot_password', key='error.form',
                rpconfig=rpconfig)
            polite_form_errors(form._errors)

    return render_to_response('registration/forgot_password.html',
            RequestContext(request, {
                'form': form,
                'CAPTCHA_PUBLIC_KEY': settings.CAPTCHA_PUBLIC_KEY,
                'CAPTCHA_API_URL_SECURE': settings.CAPTCHA_API_URL_SECURE,
                }))


@guest_required
def reset_password(request, authtoken, email_address, token=None):
    # Need this for displaying metrics per RP
    rpconfig = _get_rpconfig_from_request(request, token)

    request.token = token
    atrequest = verify_token_string(authtoken, email_address)
    account = atrequest and atrequest.requester
    if (atrequest is None or
        atrequest.token_type != LoginTokenType.PASSWORDRECOVERY or
        not account.can_reset_password):
        # track bad token error
        stats.increment('flows.forgot_password', key='error.token',
            rpconfig=rpconfig)
        # we hide the fact the the account is inactive, to avoid
        # exposing valid accounts. however, deactivated accounts can be
        # reactivated.
        return HttpResponseRedirect('/+bad-token')

    if request.method == 'POST':
        form = ResetPasswordForm(request.POST)
        if form.is_valid():
            password = form.cleaned_data['password']
            if not account.is_active and account.can_reactivate:
                # if account can be reactivated, set the preferred email
                # address
                email_obj, created = EmailAddress.objects.get_or_create(
                    account=account,
                    email=atrequest.requester_email,
                    defaults={'status': EmailStatus.PREFERRED})
                if not created:
                    account.preferredemail = email_obj
                # reactivate the account
                account.status = AccountStatus.ACTIVE
                account.save()
                request.session['message'] = _(
                    "Your account was reactivated, and the preferred email "
                    "address was set to %s." % email_obj.email)
            elif account.preferredemail is None:
                # handle this case gracefully according to #524582
                email_obj, created = EmailAddress.objects.get_or_create(
                    account=account,
                    email=atrequest.requester_email,
                    defaults={'status': EmailStatus.PREFERRED})
                if not created:
                    account.preferredemail = email_obj
            email = account.preferredemail.email
            try:
                password_obj = account.accountpassword
            except AccountPassword.DoesNotExist:
                password_obj = AccountPassword.objects.create(
                    account=account,
                    password='invalid',
                )
            password_obj.password = encrypt_launchpad_password(password)
            password_obj.save()
            user = auth.authenticate(username=email, password=password)
            auth.login(request, user)
            atrequest.consume()
            # track password recovery success
            stats.increment('flows.forgot_password', key='success',
                rpconfig=rpconfig)
            return HttpResponseRedirect(atrequest.redirection_url)

        # track form errors
        stats.increment('flows.forgot_password', key='error.form',
            rpconfig=rpconfig)
    else:
        form = ResetPasswordForm()
    return render_to_response('registration/reset_password.html',
        RequestContext(request, {'form': form, 'token': token}))


def static_page(request, page_name):
    return render_to_response('static/%s.html' % page_name,
        RequestContext(request))
